unit DocFile;

interface
  uses PalmDatabase, classes, sysutils;

const
  MAX_BOOKMARK_NAME_LENGTH = 16;

type
    TProgressCallback = procedure(Percent: Integer);

    //////////////////////////////////////////////////////////////
    //Record type for the Doc header
    TDocHeaderNative = packed record
        Version, Spare: word;
        Length: Longint;
        Records, RecSize: word;
        Position: Longint;
    end;

    TPeanutHeaderNative = packed record
        Version: word;
        Junk1: array[0..5] of char;
        Records: word;
        Junk2: array[0..105] of char;
    end;

    //////////////////////////////////////////////////////////////
    //Bookmark support
    TBookmarkNative = packed record
        Name: array[0..MAX_BOOKMARK_NAME_LENGTH - 1] of char;
        Offset: Longint;
    end;

    TBookmark = class(TObject)
    public
        Name: String;
        Offset: Integer;  //This is the offset in the RichEdit text, with CRLF
        PalmOffset: Integer;  //This is the offset as stored on the Palm, with LF only
        constructor Create(Name: PChar; Offset: Integer);
    end;

    //////////////////////////////////////////////////////////////
    //TDocFile
    //
    //A way to read a DOC file stored as a Palm PDB file
    //Inherits from TPalmPDB
    TDocFile = class(TPalmPDB)
    public
        Compressed: Boolean;
        DocNumRecords: Integer;
        DocRecSize: Integer;
        CurrPos: Integer;
        BookmarkList: TList;
        //Note: Peanut documents are not fully supported
        PeanutFile: Boolean;
        ShowPeanutCodes: Boolean;

        constructor Create(FileName: String);
        constructor CreateNewDoc(FileName, Title: String; Compressed: Boolean);
        destructor Destroy; Override;
        function GetRecord(RecNo: Integer): String;
        function GetDocument: String;
        function GetDocumentHTML: String;
        procedure WriteDocument(Buffer: String; ProgressCallback: TProgressCallback);
        function CompressRecord(Buffer: String): PChar;
        procedure AddBookmark(Name: String; Offset: Integer);
        procedure UpdateBookmark(Index: Integer; Offset: Integer);
        procedure DeleteBookmark(Index: Integer);
        procedure RenameBookmark(Index: Integer; Name: String);
    private
//        function FindBookmark(Name: String): Integer;
        function RemovePeanutCodes(const S: String): String;
        function LFOnly(const S: String): String;
        function ValidCRLF(const S: String): String;
        procedure IncBookmarks(FirstOffset: Integer);
        procedure DecBookmarks(FirstOffset: Integer);
        procedure ResetBookmarkPalmOffsets;
    end;

    EInvalidDocFile = class(Exception);

implementation

const
    DISP_BITS = 11;
    COUNT_BITS = 3;
    DOC_UNCOMPRESSED = 1;
    DOC_COMPRESSED = 2;

constructor TBookmark.Create(Name: PChar; Offset: Integer);
begin
    Self.Name := Name;
    Self.Offset := Offset;
end;


//////////////////////////////////////////////////////////////
//TDocFile definitions
//
//Read an existing Doc
constructor TDocFile.Create(FileName: String);
var
    HeaderBuffer: ^TDocHeaderNative;
    PeanutHeaderBuffer: ^TPeanutHeaderNative;
    BookmarkBuffer: ^TBookmarkNative;
    I: Integer;
begin
    //Call the parent's constructor
    try
        inherited Create(FileName, TDatabaseOpenMode(doExisting));
    except
        on E: Exception do
            raise EInvalidPalmDatabase.Create('Not a Palm database or database corrupt');
    end;

    //Make sure it's really a Doc
    PeanutFile := False;
    if (UpperCase(DBHeader.DBType) <> 'TEXT') or
       (UpperCase(DBHeader.CreatorID) <> 'READ') then
        if (UpperCase(DBHeader.DBType) <> 'PNRD') or
           (UpperCase(DBHeader.CreatorID) <> 'PPRS') then
        raise EInvalidDocFile.Create('Not a valid Doc file')
      else
          PeanutFile := True;

    if not PeanutFile then
    begin
        try
           //Read the Doc header (the first record on the database) and initialize
           HeaderBuffer := Pointer(inherited GetRecord(1));
           Compressed := (HeaderBuffer.Version and $0200) <> 0;
           DocNumRecords := Swap(HeaderBuffer.Records);
           DocRecSize := Swap(HeaderBuffer.RecSize);
           CurrPos := SwapLongInt(HeaderBuffer.Position);
        finally
           //Get rid of the native header
           StrDispose(PChar(HeaderBuffer));
        end;

        //Read the bookmarks (if any)
        BookmarkList := TList.Create;
        if (DBHeader.NumRecords - 1) > DocNumRecords then
            for I := (DocNumRecords + 2) to DBHeader.NumRecords do
            try
                BookmarkBuffer := Pointer(inherited GetRecord(I));
                BookmarkList.Add(TBookmark.Create(BookmarkBuffer.Name, SwapLongInt(BookmarkBuffer.Offset)));
            finally
                StrDispose(PChar(BookmarkBuffer));
            end;
       end
    else
    begin  //Peanut doc
        try
           //Read the Doc header (the first record on the database) and initialize
           PeanutHeaderBuffer := Pointer(inherited GetRecord(1));
           Compressed := (PeanutHeaderBuffer.Version and $0200) <> 0;
           //It seems you have to take 1 off this value...  not sure why
           DocNumRecords := Swap(PeanutHeaderBuffer.Records) - 1;
           DocRecSize := 4096;
           CurrPos := 0;
        finally
           //Get rid of the native header
           StrDispose(PChar(PeanutHeaderBuffer));
        end;

        BookmarkList := TList.Create;
    end;
end;


//Create a new Doc (we don't necessarily need the filename at this point)
constructor TDocFile.CreateNewDoc(FileName, Title: String; Compressed: Boolean);
begin
    inherited Create(FileName, doNew);
    //Set the Database header information
    if (Length(Title) > MAX_DB_NAME_LENGTH) then
    begin
        Title := Copy(Title, 1, MAX_DB_NAME_LENGTH);
    end;

    DBHeader.Name := Title;
    DBHeader.DBType := 'TEXt';
    DBHeader.CreatorID := 'REAd';
    DBHeader.Backup := True;
    //Set the Doc header information
    Self.Compressed := Compressed;
    //Empty bookmark list
    BookmarkList := TList.Create;
end;


destructor TDocFile.Destroy;
var
    I: Integer;
begin
    //Free the bookmark list
    if BookmarkList <> nil then
    begin
        BookmarkList.Pack;
        for I := 0 to (BookmarkList.Count - 1) do
            TBookmark(BookmarkList.Items[I]).Free;
        BookmarkList.Free;
    end;
    inherited Destroy;
end;


//This function does several things (in this order):
// It gets the physical record off the database (inherited GetRecord)
// It uncompresses the record, if it is compressed
function TDocFile.GetRecord(RecNo: Integer): String;
var
    BufferIn: PChar;
    BufferOut: String;
    I, HoldInt, CopyOffset, CopyLength: Integer;
    IndexIn, IndexOut: Integer;
begin
    //Let the parent class do the I/O (read the physical record)
    BufferIn := inherited GetRecord(RecNo);

    //Allocate a buffer to store the text when we uncompress it
    //The DocRecSize is the maximum that the uncompressed record will be
    //BufferOut := StrAlloc(DocRecSize);
    SetLength(BufferOut, DocRecSize);

    //This works for unencrypted Peanut files
    if PeanutFile then
        for IndexIn := 0 to (StrBufSize(BufferIn) - 1) do
            BufferIn[IndexIn] := Char((Byte(BufferIn[IndexIn])) xor $A5);

    if Compressed then
    begin   //Decompress routine
        IndexIn := 0;
        IndexOut := 1;
        while IndexIn < StrBufSize(BufferIn) do
            begin
                //This shouldn't really occur, but it sometimes does...
                if IndexOut > DocRecSize then
                    SetLength(BufferOut, Length(BufferOut) + 256);

                case Byte(BufferIn[IndexIn]) of
                $00:
                begin
                    //Just in case there are nulls, substitute spaces
                    BufferOut[IndexOut] := ' ';
                    Inc(IndexOut);
                    Inc(IndexIn);
                end;

                $01..$08:
                begin
                    //Copy this many bytes following this character
                    //First, skip this character
                    Inc(IndexIn);
                    //But make sure we look at the proper count byte
                    for I := 1 to Byte(BufferIn[IndexIn - 1]) do
                    begin
                        BufferOut[IndexOut] := BufferIn[IndexIn];
                        Inc(IndexOut);
                        Inc(IndexIn);
                    end;
                end;

                $09..$7F:
                begin
                    //Copy this character directly
                    //We can't reliably count the CR/LF characters here, since they
                    //only find the ones not in a compression sequence!
                    BufferOut[IndexOut] := BufferIn[IndexIn];
                    Inc(IndexOut);
                    Inc(IndexIn);
                end;

                $80..$BF:
                begin
                    //Copy a sequence
                    //HoldInt is the 2 bytes of the compression sequence
                    //(Add the high byte, increment, add the low byte)
                    HoldInt := Byte(BufferIn[IndexIn]) shl 8;
                    Inc(IndexIn);
                    Inc(HoldInt, Byte(BufferIn[IndexIn]));

                    CopyOffset := (HoldInt and $3FFF) shr COUNT_BITS;
                    //1 shl COUNT_BITS is 1000, minus 1 is 111 (this returns the lower 3 bits)
                    CopyLength := (HoldInt and ((1 shl COUNT_BITS) - 1)) + 3;

                    //Using the offset and length, copy the stuff (it's
                    //already in the destination buffer; just go back and
                    //grab it)
                    for I := 1 to CopyLength do
                    begin
                        BufferOut[IndexOut] := BufferOut[IndexOut - CopyOffset];
                        Inc(IndexOut);
                    end;
                    Inc(IndexIn);
                end;

                $C0..$FF:
                begin
                    //Space and this character xor $80
                    //First the space
                    BufferOut[IndexOut] := ' ';
                    Inc(IndexOut);
                    //Now the character
                    BufferOut[IndexOut] := Char(Byte(BufferIn[IndexIn]) xor $80);
                    Inc(IndexOut);
                    Inc(IndexIn);
                end;
            end; //case Byte(BufferIn[IndexIn]) of
        end; //while IndexIn < StrBufSize(BufferIn)
        //In case it didn't uncompress to the specified number of bytes
        SetLength(BufferOut, IndexOut - 1);
    end  //if Compressed
    else
    begin   //Uncompressed: copy the in buffer to the out buffer
        for I := 0 to (StrBufSize(BufferIn) - 1) do
            BufferOut[I + 1] := BufferIn[I];
        SetLength(BufferOut, StrBufSize(BufferIn));
    end;

    //Get rid of the input buffer
    StrDispose(BufferIn);

    Result := BufferOut;
end;


//This gets the entire document (all records)
function TDocFile.GetDocument: String;
var
    I: Integer;
begin
    Result := '';

    try
        for I := 1 to DocNumRecords do
            Result := Result + GetRecord(I + 1);
    except
        Result := '';
        raise EInvalidDocFile.Create('Error reading Doc file');
    end;
    if PeanutFile and not ShowPeanutCodes then
        Result := RemovePeanutCodes(Result);

    Result := ValidCRLF(Result);
end;


//This gets the entire document, but returns <BR> in place of CRLF
//Seemingly this code could all be replaced by a call to StringReplace,
//but StringReplace is MUCH slower.
function TDocFile.GetDocumentHTML: String;
var
    DocOutPosition, DocInPosition: Integer;
    DocumentBuffer: String;
    CRLFCount, I: Integer;
begin
    Result := '';

    //Get the document
    DocumentBuffer := GetDocument;

    //Count all the CRLF pairs
    CRLFCount := 0;
    for I := 1 to length(DocumentBuffer) do
        if Byte(DocumentBuffer[I]) = $0D then
            Inc(CRLFCount);

    //Give our result string some additional space (since CRLF,
    //2 bytes, will become <BR>, 4 bytes
    SetLength(Result, Length(DocumentBuffer) + (CRLFCount * 2));

    //Loop through the document
    DocInPosition := 0;
    DocOutPosition := 0;
    while DocInPosition < Length(DocumentBuffer) do
    begin
        //If CRLF encountered, replace with <BR>
        case Byte(DocumentBuffer[DocInPosition]) of
        $0D:
        begin
            if DocumentBuffer[DocInPosition + 1] = Char($0A) then
                begin
                    Move('<BR>', Result[DocOutPosition], 4);
                    Inc(DocOutPosition, 4);
                    Inc(DocInPosition, 2);
                end;
            end;
            else
            begin
                //Copy what was there
                Result[DocOutPosition] := DocumentBuffer[DocInPosition];

                Inc(DocOutPosition);
                Inc(DocInPosition);
            end;
        end; //case
    end;
end;


//Get rid of Peanut tags; note, there is still work to be done
function TDocFile.RemovePeanutCodes(const S: String): String;
var
    Source, SourceEnd, Dest: PChar;
    CharCode: String;
begin
    Source := Pointer(S);
    SourceEnd := Source + Length(S);

    SetString(Result, nil, Length(S));
    Dest := Pointer(Result);
    while Source < SourceEnd do
    begin
        while Source^ = '\' do
        begin
            Inc(Source);
            case Source^ of
            'p', 'x', 'c', 'r', 'i', 'u', 'o', 'v', 'n', 's', 'b', 'l', 't':
                //Ignore almost everything
                Inc(Source);
            'a':
            begin
                //Look at next 3 digits
                Inc(Source);
                CharCode := '' + Source^;
                Inc(Source);
                CharCode := CharCode + Source^;
                Inc(Source);
                CharCode := CharCode + Source^;
                Inc(Source);
                try
                    Dest^ := Char(StrToInt(CharCode));
                except
                    Dest^ := '?';
                end;
                Inc(Dest);
            end
            else
            end;
        end;
        Dest^ := Source^;
        Inc(Source);
        Inc(Dest);
    end;

    SetLength(Result, Dest - Pointer(Result));
end;

//Compress routine
function TDocFile.CompressRecord(Buffer: String): PChar;
var
    //Used by CompressDocument
    Hit, PrevHit, TestHead, TestTail, BufEnd: PChar;
    Dist, Compound: Cardinal;
    I, K: Cardinal;
    //Shared with Issue
    Space: Boolean;
    DestBuf: PChar;
    DestIndex: Cardinal;

    //Handle writing a single byte to the compressed output
    procedure Issue(Src: Char);
    begin
        if Space then  //Space flag (preceding character is space)
        begin
            Space := False;   //This is the $C0..$FF condition in the decompress
            if (Byte(Src) >= $40) and (Byte(Src) <= $7F) then
            begin
                DestBuf[DestIndex] := Char(Byte(Src) xor $80);
                Inc(DestIndex);
            end
            else
            begin
                DestBuf[DestIndex] := ' '; //Out of range, can't use space compression code
                Inc(DestIndex);            //Must insert an actual space, then the byte
                if ((Byte(Src) < $80) and ((Byte(Src) = 0) or (Byte(Src) > 8))) then
                begin
                    DestBuf[DestIndex] := Src;
                    Inc(DestIndex);
                end
                else
                begin
                    DestBuf[DestIndex] := Char($01);
                    Inc(DestIndex);
                    DestBuf[DestIndex] := Src;
                    Inc(DestIndex);
                end;
            end;
        end
        else //Space flag clear
            if (Src = ' ') then
            begin
                Space := True;
            end
            else
            begin
                if ((Byte(Src) < $80) and ((Byte(Src) = 0) or (Byte(Src) > 8))) then
                begin
                    DestBuf[DestIndex] := Src;
                    Inc(DestIndex);
                end
                else
                begin
                    DestBuf[DestIndex] := Char($01);
                    Inc(DestIndex);
                    DestBuf[DestIndex] := Src;
                    Inc(DestIndex);
                end;
            end;
    end;

    //Called in CompressRecord
    function MemFind(T: PChar; TLen: Integer; M: PChar; MLen: Integer): PChar;
    var
        I: Integer;
    begin
        I := TLen - MLen + 1;
        while I > 0 do
        begin
            if (T[0] = M[0]) and (CompareMem(Pointer(T), Pointer(M), MLen)) then
            begin
                Result := T;
                exit;
            end;
            Inc(T);
            Dec(I);
        end;
        Result := nil;
    end;

begin //CompressRecord
    //Initialization
    Hit := @Buffer[1];
    PrevHit := @Buffer[1];
    TestHead := @Buffer[1];
    TestTail := TestHead + 1;
    DestIndex := 0;
    Space := False;

    BufEnd := StrEnd(PChar(Buffer));

    //Allocate a destination buffer
    DestBuf := StrAlloc(Round(Length(Buffer) * 3 / 2));

    //Loop, absorbing one more char from the input buffer on each pass
    while TestHead <> BufEnd do
    begin
        //Establish where the scan can begin
        //If the previous hit is too far back (over 2047 bytes, the max 11 bits can point)
        if (TestHead - PrevHit) > ((1 shl DISP_BITS) - 1) then
            PrevHit := TestHead - ((1 shl DISP_BITS) - 1);  //set it back to where it is still in range

        //Scan in the previous data for a match
        Hit := MemFind(PrevHit, TestTail - PrevHit, TestHead, TestTail - TestHead);

        //On a mismatch or end of buffer, issue codes
        if (Hit = nil) or
           (Hit = TestHead) or
           (TestTail - TestHead > (1 shl COUNT_BITS) + 2) or  //there's a match, and the repeated bytes are past the limit
           (TestTail = BufEnd) then
            begin
                //Write the codes
                if ((TestTail - TestHead) < 4) then  //Not worth compressing less than 4 bytes...
                begin
                    Issue(TestHead[0]);              //just write this out
                    Inc(TestHead);
                end
                else
                begin  //for longer runs, issue a run-code
                    //issue space char if required
                    if Space then
                    begin
                        DestBuf[DestIndex] := ' ';
                        Inc(DestIndex);
                        Space := False;
                    end;

                    Dist := TestHead - PrevHit;  //Distance back from where we are
                    Compound := (Dist shl COUNT_BITS) + TestTail - TestHead - 4;

                    //The most significant bits of Compound + $80
                    DestBuf[DestIndex] := Char(Byte($80 + (Compound shr 8)));
                    Inc(DestIndex);
                    //The least significant bits of Compound
                    DestBuf[DestIndex] := Char(Byte(Compound and $FF));
                    Inc(DestIndex);
                    TestHead := TestTail - 1;
                end;
                //Start the search again
                PrevHit := @Buffer[1];
            end
            else  //got a match - keep going backwards
            begin
                PrevHit := Hit;
            end;
        //When we get to the end of the buffer, don't Inc past the end
        //this forces the residue chars out one at a time
        if TestTail = BufEnd then
            Dec(TestTail);

        Inc(TestTail);
    end;

    //Clean up any dangling spaces
    if Space then
    begin
        DestBuf[DestIndex] := ' ';
        Inc(DestIndex);
    end;

    //Final scan to merge consecutive high chars together
    I := 0;
    K := 0;
    while (I < DestIndex) do
    begin
        DestBuf[K] := DestBuf[I];

        //If we just copied the first byte of a compression sequence,
        //copy the second one too
        if ((DestBuf[K] >= Char($80)) and (DestBuf[K] < Char($C0))) then
        begin
            Inc(I);
            Inc(K);
            DestBuf[K] := DestBuf[I];
        end
        else
            //If we hit a high char marker, look ahead for another
            if (DestBuf[K] = Char($01)) then
            begin
                DestBuf[K + 1] := DestBuf[I + 1];
                while ((I + 2 < DestIndex)
                  and (DestBuf[I + 2] = Char($01))
                  and (DestBuf[K] < Char($08))) do
                    begin
                        DestBuf[K] := Char(Byte(DestBuf[K]) + 1);
                        DestBuf[K + Byte(DestBuf[K])] := DestBuf[I + 3];
                        Inc(I, 2);
                    end;
                Inc(K, Byte(DestBuf[K]));
                Inc(I);
            end;

        Inc(I);
        Inc(K);
    end;

    //K is the final length of the buffer
    //Allocate the result to K + 1 and return it
    Result := StrAlloc(K);
    Move(DestBuf^, Result^, K);

    StrDispose(DestBuf);
end;


//Accept a large amount of text (like the contents of the RichText)
//Update Doc header at end with size information
//Compress 1 chunk at a time (4096 bytes)
//Write all the chunks to the output file
procedure TDocFile.WriteDocument(Buffer: String; ProgressCallback: TProgressCallback);
var
    ChunkCounter: Cardinal;
    RecNo: Integer;
    Header: ^TDocHeaderNative;
    BookmarkBuffer: ^TBookmarkNative;
    Records: Integer;
    UncompressedBuffer: PChar;
    WorkingBuffer: String;
    UniqueID: array[0..2] of Char;
    I: Integer;
const
    CHUNKSIZE = 4096;
    UNIQUEIDBASE = $6FA000;

    //Probably a better way to do this... but there's no such thing as a 3-byte
    //integer type
    procedure SetUniqueID(UniqueIDInt: Integer);
    begin
        UniqueID[0] := char((UniqueIDInt and $FF0000) shr 16);
        UniqueID[1] := char((UniqueIDInt and $00FF00) shr 8);
        UniqueID[2] := char((UniqueIDInt and $0000FF));
    end;

begin
    //Start by clearing the record list... otherwise it may be incorrect
    //after we rebuild it
    ClearRecordList;

    WorkingBuffer := LFOnly(Buffer);

    //Can't save an empty document
    if WorkingBuffer = '' then
        WorkingBuffer := ' ';

    Header := Pointer(StrAlloc(SizeOf(TDocHeaderNative)));

    ChunkCounter := 0;
    RecNo := 2;

    Records := Length(WorkingBuffer) div CHUNKSIZE;
    if (Length(WorkingBuffer) mod CHUNKSIZE) > 0 then
        Inc(Records);

    //Write the Doc header (record 1)
    if Compressed then
        Header.Version := Swap(DOC_COMPRESSED)
    else
        Header.Version := Swap(DOC_UNCOMPRESSED);
    Header.Spare := 0;
    Header.Length := SwapLongInt(Length(WorkingBuffer));
    Header.Records := Swap(Records);
    Header.RecSize := Swap(CHUNKSIZE);
    Header.Position := 0;

    PutRecord(1, PChar(Header));
    SetUniqueID(UNIQUEIDBASE + 1);
    SetRecordHeaderInfo(1, [rfDirty], UniqueID);

    //Write the Doc records
    while ChunkCounter <= Length(WorkingBuffer) do
    begin
        if Compressed then
            PutRecord(RecNo, CompressRecord(copy(WorkingBuffer, ChunkCounter + 1, CHUNKSIZE)))
        else
        begin
            //This is freed in FinalizeWrite (CompressRecord allocates the buffer within the function)
            //Don't cast... allocate a PChar and copy it
            if Length(WorkingBuffer) < (ChunkCounter + 1 + CHUNKSIZE) then
                UncompressedBuffer := StrAlloc(Length(WorkingBuffer) - (ChunkCounter + 1) + 1)
            else
                UncompressedBuffer := StrAlloc(CHUNKSIZE);
            Move(WorkingBuffer[ChunkCounter + 1], UncompressedBuffer^, StrBufSize(UncompressedBuffer));
            PutRecord(RecNo, UncompressedBuffer);
        end;

        SetUniqueID(UNIQUEIDBASE + RecNo);
        SetRecordHeaderInfo(RecNo, [rfDirty], UniqueID);

        Inc(ChunkCounter, CHUNKSIZE);
        Inc(RecNo);
        if Assigned(ProgressCallback) then
            ProgressCallback(Round(RecNo/Records * 100));
    end;

    //Write the Bookmark records (if any)
    for I := 0 to (BookmarkList.Count - 1) do
    begin
        //Allocate a buffer and set the values
        BookmarkBuffer := Pointer(StrAlloc(SizeOf(TBookmarkNative)));
        StrPLCopy(BookmarkBuffer.Name, TBookmark(BookmarkList.Items[I]).Name, SizeOf(BookmarkBuffer.Name));
        BookmarkBuffer.Offset := SwapLongInt(TBookmark(BookmarkList.Items[I]).PalmOffset);
        //Write to the database
        PutRecord(0, PChar(BookmarkBuffer));
    end;

    //FinalizeWrite frees all the strings we allocated above
    FinalizeWrite;
end;


procedure TDocFile.AddBookmark(Name: String; Offset: Integer);
begin
    //Add a bookmark to the end of the list
    BookmarkList.Add(TBookmark.Create(PChar(Name), Offset));
end;

{
//Find the bookmark by name and return the index or -1 if not found
//(I thought this would be useful, but it turns out I'm not using it)
function TDocFile.FindBookmark(Name: String): Integer;
var I: Integer;
begin
    Result := -1;
    for I := 0 to (BookmarkList.Count - 1) do
    begin
        if TBookmark(BookmarkList.Items[I]).Name = Name then
        begin
            Result := I;
            exit;
        end;
    end;
end;
}

procedure TDocFile.UpdateBookmark(Index: Integer; Offset: Integer);
begin
    //Update an existing bookmark with a new offset
    TBookmark(BookmarkList.Items[Index]).Offset := Offset;
end;


procedure TDocFile.DeleteBookmark(Index: Integer);
begin
    //Delete the given bookmark
    TBookmark(BookmarkList.Items[Index]).Free;
    BookmarkList.Delete(Index);
    BookmarkList.Pack;
end;


procedure TDocFile.RenameBookmark(Index: Integer; Name: String);
begin
    TBookmark(BookmarkList.Items[Index]).Name := Name;
end;


//Replace stray CR or LF with CRLF pairs (clone of AdjustLineBreaks)
//Also, since this will affect offsets (and therefore bookmarks), adjust
//bookmark records to compensate for added characters
//This is why we can't simply call AdjustLineBreaks
function TDocFile.ValidCRLF(const S: String): String;
var
    Source, SourceEnd, Dest: PChar;
    AddBytes: Integer;
begin
    ResetBookmarkPalmOffsets;

    Source := Pointer(S);
    SourceEnd := Source + Length(S);
    AddBytes := 0;
    //Find out how many bytes will be dropped: loop through the source
    //Also, adjust the bookmark records where appropriate
    while Source < SourceEnd do
    begin
        case Source^ of
        #10:
        begin
            Inc(AddBytes);
            IncBookmarks(Source - Pointer(S));
        end;

        #13:
            if Source[1] = #10 then
                Inc(Source)
            else
            begin
                Inc(AddBytes);
                IncBookmarks(Source - Pointer(S));
            end;

        else
            if Source^ in LeadBytes then
                Inc(Source);
        end;
        Inc(Source);
    end;

    //Make the actual changes
    if AddBytes = 0 then
        Result := S
    else
    begin
        Source := Pointer(S);
        SetString(Result, nil, SourceEnd - Source + AddBytes);
        Dest := Pointer(Result);
        while Source < SourceEnd do
            case Source^ of
            #10:
            begin
                Dest^ := #13;
                Inc(Dest);
                Dest^ := #10;
                Inc(Dest);
                Inc(Source);
            end;

            #13:
            begin
                Dest^ := #13;
                Inc(Dest);
                Dest^ := #10;
                Inc(Dest);
                Inc(Source);
                if Source^ = #10 then
                    Inc(Source);
            end;

            else
                if Source^ in LeadBytes then
                begin
                    Dest^ := Source^;
                    Inc(Dest);
                    Inc(Source);
                end;

                Dest^ := Source^;
                Inc(Dest);
                Inc(Source);
            end;
    end;
end;

//Replace CRLF pairs with CR only (TealDoc puts in 2 breaks if we leave CRLF),
//plus 1 byte takes up less space than 2 bytes :)
//Used AdjustLineBreaks as a guide, but we know that all CR will have a
//corresponding LF, so not as much checking in the first loop
//
//Also, since this will affect offsets (and therefore bookmarks), adjust
//bookmark records to compensate for deleted characters
function TDocFile.LFOnly(const S: String): String;
var
    Source, SourceEnd, Dest: PChar;
    DropBytes: Integer;
begin
    ResetBookmarkPalmOffsets;

    Source := Pointer(S);
    SourceEnd := Source + Length(S);
    DropBytes := 0;
    //Find out how many bytes will be dropped: loop through the source
    while Source < SourceEnd do
    begin
        if Source^ = #13 then
        begin
            Inc(DropBytes);
            DecBookmarks(Source - Pointer(S));
            Inc(Source);
        end;
        Inc(Source);
    end;

    //Copy the string, dropping the CR
    if DropBytes = 0 then
        Result := S
    else
    begin
        Source := Pointer(S);
        SetString(Result, nil, SourceEnd - Source - DropBytes);
        Dest := Pointer(Result);
        while Source < SourceEnd do
        begin
            if Source^ = #13 then
                Inc(Source);
            Dest^ := Source^;
            Inc(Dest);
            Inc(Source);
        end;
    end;
end;


//Loop through the bookmark list, and increment any bookmarks with an
//offset >= the FirstOffset they pass in
procedure TDocFile.IncBookmarks(FirstOffset: Integer);
var
    I: Integer;
begin
    for I := 0 to (BookmarkList.Count - 1) do
        if TBookmark(BookmarkList.Items[I]).PalmOffset > FirstOffset then
            Inc(TBookmark(BookmarkList.Items[I]).Offset);
end;

//Loop through the bookmark list, and decrement any bookmarks with an
//offset >= the FirstOffset they pass in
procedure TDocFile.DecBookmarks(FirstOffset: Integer);
var
    I: Integer;
begin
    for I := 0 to (BookmarkList.Count - 1) do
        if TBookmark(BookmarkList.Items[I]).Offset > FirstOffset then
            Dec(TBookmark(BookmarkList.Items[I]).PalmOffset);
end;


//This copies the bookmarks from their screen offsets (which include CRLF)
//into the PalmOffset field, which will be decremented when LFOnly is called
procedure TDocFile.ResetBookmarkPalmOffsets;
var
    I: Integer;
begin
    for I := 0 to (BookmarkList.Count - 1) do
        TBookmark(BookmarkList.Items[I]).PalmOffset := TBookmark(BookmarkList.Items[I]).Offset;
end;


end.
