unit PalmDatabase;

interface
uses classes, sysutils;

const
    MAX_DB_NAME_LENGTH = 31;

type
    TDatabaseOpenMode = (doNew, doExisting);

    //////////////////////////////////////////////////////////////
    //Record type for the database header (common to PDB and PRC)
    TPDBHeaderNative = packed record
        Name: array[0..MAX_DB_NAME_LENGTH] of char;
        Flags, Version: Word;
        Created, Modified, BackedUp: Longint;
        Modification, AppInfoArea, SortInfoArea: Longint;
        DBType, CreatorID: array[0..3] of char;
        IDSeed, NextRecordList: Longint;
        NumRecords: Word;
    end;

    //////////////////////////////////////////////////////////////
    //Record type for the record list (PDB)
    TPDBRecordHeaderNative = packed record
        DataOffset: Longint;
        Flags: char;
        UniqueID: array[0..2] of char;
    end;

    //try this eventually
    //TDBFlag = (dbReadOnly, dbDirty, dbBackup,
    //           dbOverwrite, dbForceReset, dbNoBeam);
    //TDBFlags = set of TDBFlag;

    //////////////////////////////////////////////////////////////
    //TPDBHeader
    //
    //The database header (for PDB and PRC)
    TPDBHeader = record
        Name: String;
        ReadOnly, Dirty, Backup, Overwrite, ForceReset, NoBeam: Boolean;
        //DBFlags: TDBFlags;
        Version: Integer;
        Created, Modified, BackedUp: TDateTime;
        Modification, AppInfoArea, SortInfoArea: Integer;
        DBType, CreatorID: String;
        IDSeed, NextRecordList: Integer;
        NumRecords: Integer;
    end;

    TRecFlag = (rfSecret, rfBusy, rfDirty, rfDelete);
    TRecFlags = set of TRecFlag;

    //////////////////////////////////////////////////////////////
    //TPDBRecordHeader
    //
    //Record header class (A PDB maintains a TList of these)
    //Physically, this is right after the database header in the file
    TPDBRecordHeader = class
    public
        Offset: Integer;
        RecFlags: TRecFlags;
        UniqueID: array[0..2] of char;
        Size: Integer;
        RecordData: PChar;
        constructor Create(Offset, Size: Integer);
        destructor Destroy; Override;
    end;

    //////////////////////////////////////////////////////////////
    //TPalmDatabase
    //
    //Parent class for each item that gets printed on the check
    //Generic parent class for both PDB and PRC
    TPalmDatabase = class
    public
        DBHeader: TPDBHeader;
        DBFileSize: Integer;
        PDBFile: String;
        constructor Create(FileName: String; OpenMode: TDatabaseOpenMode);
        procedure SetFile(FileName: String);
        procedure GetHeader;
    protected
        OpenMode: TDatabaseOpenMode;
        function SwapLongInt(L: Longint): longint;
        function PalmToTDateTime(Seconds: Cardinal): TDateTime;
        function TDateTimeToPalm(DateTime: TDateTime): Cardinal;
    end;

    //////////////////////////////////////////////////////////////
    //TPalmPDB
    //
    //PDB database
    TPalmPDB = class(TPalmDatabase)
        PDBRecordList: TList;
        constructor Create(FileName: String; OpenMode: TDatabaseOpenMode);
        destructor Destroy; Override;
        procedure GetRecordList;
        procedure ClearRecordList;
        function GetRecord(RecNo: Integer): PChar;
        procedure PutRecord(RecNo: Integer; RecordData: PChar);
        procedure SetRecordHeaderInfo(RecNo: Integer; RecFlags: TRecFlags; UniqueID: PChar);
        procedure FinalizeWrite;
    end;

    //Implement this someday
    //TPalmPRC = class(TPalmDatabase)
    //end;

    EInvalidPalmDatabase = class(Exception);

//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// Implementation
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////

implementation

//////////////////////////////////////////////////////////////
//TPalmDatabase definitions
//
constructor TPalmDatabase.Create(FileName: String; OpenMode: TDatabaseOpenMode);
begin
    Self.OpenMode := OpenMode;
    //Initialize the header class so we have access to it later
    PDBFile := FileName;
    if OpenMode = doExisting then
        GetHeader;
end;


procedure TPalmDatabase.SetFile(FileName: String);
begin
    PDBFile := FileName;
end;


procedure TPalmDatabase.GetHeader;
var
    DBFile: file;
    DBHeaderNative: TPDBHeaderNative;
begin
    //Open the file and read off the header
    try
      try
        AssignFile(DBFile, PDBFile);
        FileMode := 0;  // Set file access to read only
        Reset(DBFile, 1);
        DBFileSize := FileSize(DBFile);
        BlockRead(DBFile, DBHeaderNative, Sizeof(TPDBHeaderNative));
      except
        raise EInvalidPalmDatabase.Create('Error reading file');
      end;
    finally
      try
        CloseFile(DBFile);
      except
      end;
    end;

    try
        //Assign the native types to Delphi types
        DBHeader.Name := DBHeaderNative.Name;
        //Remember, the bytes are swapped on Intel
        DBHeaderNative.Flags := Swap(DBHeaderNative.Flags);
        DBHeader.ReadOnly := Boolean(DBHeaderNative.Flags and $0002);
        DBHeader.Dirty := Boolean(DBHeaderNative.Flags and $0004);
        DBHeader.Backup := Boolean(DBHeaderNative.Flags and $0008);
        DBHeader.Overwrite := Boolean(DBHeaderNative.Flags and $0010);
        DBHeader.ForceReset := Boolean(DBHeaderNative.Flags and $0020);
        DBHeader.NoBeam := Boolean(DBHeaderNative.Flags and $0040);
        DBHeader.Version := Swap(DBHeaderNative.Version);
        DBHeader.Created := PalmToTDateTime(SwapLongInt(DBHeaderNative.Created));
        DBHeader.Modified := PalmToTDateTime(SwapLongInt(DBHeaderNative.Modified));
        DBHeader.BackedUp := PalmToTDateTime(SwapLongInt(DBHeaderNative.BackedUp));
        DBHeader.Modification := SwapLongInt(DBHeaderNative.Modification);
        DBHeader.AppInfoArea := SwapLongInt(DBHeaderNative.AppInfoArea);
        DBHeader.SortInfoArea := SwapLongInt(DBHeaderNative.SortInfoArea);
        DBHeader.DBType := DBHeaderNative.DBType;
        DBHeader.CreatorID := DBHeaderNative.CreatorID;
        DBHeader.IDSeed := SwapLongInt(DBHeaderNative.SortInfoArea);
        DBHeader.NextRecordList := SwapLongInt(DBHeaderNative.SortInfoArea);
        DBHeader.NumRecords := Swap(DBHeaderNative.NumRecords);
    except
        raise EInvalidPalmDatabase.Create('Not a Palm database, or header corrupt');
    end;
end;


//Convert the Motorola format Lontint to the Intel format
function TPalmDatabase.SwapLongInt(L: Longint): longint;
begin
    result := swap(L shr 16) or (longint(swap(L and $ffff)) shl 16);
end;


//Delphi determines the date/time based on days since 12/30/1899
//Palm determines the date/time based on seconds since 1/1/1904
function TPalmDatabase.PalmToTDateTime(Seconds: Cardinal): TDateTime;
var
    Days, Remainder : Integer;
//This function doesn't work right
//It was translating 1999 dates to 1933, not sure why, this ErrorOffset
//kludge fixes it pretty well... ErrorOffset gets us to 1/1/1970
//(Unix time).  I don't question it, I just use it...
const
    ErrorOffset = 2082844800;  //seconds
begin
    Inc(Seconds, ErrorOffset);
    //1462 is the number of days between 12/30/1899 and 1/1/1904
    Inc(Seconds, (1462 * 86400));
    Days := Seconds div 86400;
    Remainder := Seconds mod 86400;
    Result := Days + (Remainder / 86400);
end;


//This probably doesn't work right either...
function TPalmDatabase.TDateTimeToPalm(DateTime: TDateTime): Cardinal;
const
    ErrorOffset = 2082844800;  //Seconds
var
    Days: Integer;
begin
    Days := Round(Int(DateTime)) - 1462;
    Result := Days * 86400 + Round((frac(DateTime) / 86400)) - ErrorOffset;
end;

//////////////////////////////////////////////////////////////
//TPDBRecordHeader definitions
//
constructor TPDBRecordHeader.Create(Offset, Size: Integer);
begin
    if (Offset < 0) or (Size < 0) then
        raise EInvalidPalmDatabase.Create('Invalid Palm database: record headers corrupt');
    Self.Offset := Offset;
    Self.Size := Size;
end;


//Testing
destructor TPDBRecordHeader.Destroy;
begin
    inherited Destroy;
end;


//////////////////////////////////////////////////////////////
//TPalmPDB definitions
//
constructor TPalmPDB.Create(FileName: String; OpenMode: TDatabaseOpenMode);
begin
    inherited Create(FileName, OpenMode);
    PDBRecordList := TList.Create;
    if OpenMode = doExisting then
        GetRecordList;
end;


destructor TPalmPDB.Destroy;
begin
    //Free the data pointed to by the record list
    ClearRecordList;

   	//Free the list
    PDBRecordList.Free;

    //Free everything else
    inherited Destroy;
end;


procedure TPalmPDB.GetRecordList;
var
    I: Integer;
    DBFile: file;
    PDBRecordHeaderNative: TPDBRecordHeaderNative;
begin
  try
    //Open the file
    AssignFile(DBFile, PDBFile);
    FileMode := 0;  // Set file access to read only
    Reset(DBFile, 1);
    //Skip past the header
    Seek(DBFile, Sizeof(TPDBHeaderNative));

    for I := 1 to DBHeader.NumRecords do
    begin
        //Read off a record header
        BlockRead(DBFile, PDBRecordHeaderNative, Sizeof(TPDBRecordHeaderNative));

        //Add it to the list
        PDBRecordList.Add(TPDBRecordHeader.Create(SwapLongInt(PDBRecordHeaderNative.DataOffset), 0));
        TPDBRecordHeader(PDBRecordList.Items[I - 1]).RecordData := nil;

        //Determine the size of the previous record (now that we know this one's offset)
        if PDBRecordList.Count > 1 then
            TPDBRecordHeader(PDBRecordList.Items[PDBRecordList.Count - 2]).Size := TPDBRecordHeader(PDBRecordList.Last).Offset - TPDBRecordHeader(PDBRecordList.Items[PDBRecordList.Count - 2]).Offset;
        end;

    //Set the size of the last one
    TPDBRecordHeader(PDBRecordList.Last).Size := FileSize(DBFile) - TPDBRecordHeader(PDBRecordList.Last).Offset;
  finally
    CloseFile(DBFile);
  end;
end;


procedure TPalmPDB.ClearRecordList;
var
    I: Integer;
begin
    //Free all the header items
    PDBRecordList.Pack;
    for I := 0 to (PDBRecordList.Count - 1) do
    begin
        //Free any data still associated with the record
        if TPDBRecordHeader(PDBRecordList.Items[I]).RecordData <> nil then
            StrDispose(TPDBRecordHeader(PDBRecordList.Items[I]).RecordData);
        //Free the header item
        TPDBRecordHeader(PDBRecordList.Items[I]).Free;
    end;
    PDBRecordList.Clear;
end;


//Get the record specified, allocate a buffer for it, and return a pointer
function TPalmPDB.GetRecord(RecNo: Integer): PChar;
var
    DBFile: File;
    TextBuffer: PChar;
begin
    //Allocate memory for the buffer
    TextBuffer := StrAlloc(TPDBRecordHeader(PDBRecordList.Items[RecNo - 1]).Size);

    //Open the file
    AssignFile(DBFile, PDBFile);
    Reset(DBFile, 1);
    //Skip to the requested record's offset
    Seek(DBFile, TPDBRecordHeader(PDBRecordList.Items[RecNo - 1]).Offset);

    //Read the record
    BlockRead(DBFile, TextBuffer^, TPDBRecordHeader(PDBRecordList.Items[RecNo - 1]).Size);

    //Close the file and return a pointer
    CloseFile(DBFile);
    Result := TextBuffer;
end;


procedure TPalmPDB.PutRecord(RecNo: Integer; RecordData: PChar);
var
    I: Integer;
begin
    //Add to the PDBRecordList if it's new
    if (RecNo = 0) then
    begin
        PDBRecordList.Add(TPDBRecordHeader.Create(0, StrBufSize(RecordData)));
        RecNo := PDBRecordList.Count;
    end
    else
    if (RecNo > PDBRecordList.Count) then  //Create empty filler records
    begin
        for I := 1 to (RecNo - PDBRecordList.Count) do
        begin
            PDBRecordList.Add(TPDBRecordHeader.Create(0, 0));
        end;
    end;

    //Keep a pointer to the data
    TPDBRecordHeader(PDBRecordList.Items[RecNo - 1]).RecordData := RecordData;
    TPDBRecordHeader(PDBRecordList.Items[RecNo - 1]).Size := StrBufSize(RecordData);

    //Update all the offsets
    //First record comes after database header and record headers (omit the 2 bytes optional filler)
    TPDBRecordHeader(PDBRecordList.Items[0]).Offset := Sizeof(TPDBHeaderNative) + (PDBRecordList.Count * SizeOf(TPDBRecordHeaderNative));
    for I := 1 to (PDBRecordList.Count - 1) do
    begin
        TPDBRecordHeader(PDBRecordList.Items[I]).Offset := TPDBRecordHeader(PDBRecordList.Items[I - 1]).Offset + TPDBRecordHeader(PDBRecordList.Items[I - 1]).Size;
    end;
end;


procedure TPalmPDB.SetRecordHeaderInfo(RecNo: Integer; RecFlags: TRecFlags; UniqueID: PChar);
begin
    if (RecNo > PDBRecordList.Count) then
        exit;

    TPDBRecordHeader(PDBRecordList.Items[RecNo - 1]).RecFlags := RecFlags;
    Move(UniqueID[0], TPDBRecordHeader(PDBRecordList.Items[RecNo - 1]).UniqueID[0], SizeOf(TPDBRecordHeader(PDBRecordList.Items[RecNo - 1]).UniqueID));
end;


procedure TPalmPDB.FinalizeWrite;
var
    PDBHeaderNative: TPDBHeaderNative;
    PDBRecordHeaderNative: TPDBRecordHeaderNative;
    I: Integer;
    OutStream: TFileStream;
begin
    //Open the file (overwrite if necessary)
    OutStream := TFileStream.Create(PDBFile, fmCreate or fmShareDenyWrite);

    //Prepare the database header
    FillChar(PDBHeaderNative, SizeOf(TPDBHeaderNative), 0);
    StrPLCopy(PDBHeaderNative.Name, DBHeader.Name, Length(DBHeader.Name));
    //Remember, the bytes are swapped on Intel
    PDBHeaderNative.Flags := Swap(Word((Ord(DBHeader.ReadOnly) shl 1) +
                             (Ord(DBHeader.Dirty) shl 2) +
                             (Ord(DBHeader.Backup) shl 3) +
                             (Ord(DBHeader.OverWrite) shl 4) +
                             (Ord(DBHeader.ForceReset) shl 5) +
                             (Ord(DBHeader.NoBeam) shl 6)));
    PDBHeaderNative.Version := $0;
    PDBHeaderNative.Created := SwapLongInt(TDateTimeToPalm(Now));
    PDBHeaderNative.Modified := SwapLongInt(TDateTimeToPalm(Now));
    PDBHeaderNative.BackedUp := 0;
    PDBHeaderNative.Modification := 0;
    PDBHeaderNative.AppInfoArea := 0;
    PDBHeaderNative.SortInfoArea := 0;
    PDBHeaderNative.DBType := 'TEXt';
    PDBHeaderNative.CreatorID := 'REAd';
    PDBHeaderNative.IDSeed := 0;
    PDBHeaderNative.NextRecordList := 0;
    PDBHeaderNative.NumRecords := Swap(PDBRecordList.Count);

    //Write the database header
    OutStream.WriteBuffer(PDBHeaderNative, SizeOf(PDBHeaderNative));

    //Loop through the PDBRecordList and write each list entry
    for I := 0 to (PDBRecordList.Count - 1) do
    begin
        PDBRecordHeaderNative.DataOffset := SwapLongInt(TPDBRecordHeader(PDBRecordList.Items[I]).Offset);
        PDBRecordHeaderNative.Flags := char((Ord(rfSecret in TPDBRecordHeader(PDBRecordList.Items[I]).RecFlags) shl 4) +
                                        (Ord(rfBusy in TPDBRecordHeader(PDBRecordList.Items[I]).RecFlags) shl 5) +
                                        (Ord(rfDirty in TPDBRecordHeader(PDBRecordList.Items[I]).RecFlags) shl 6) +
                                        (Ord(rfDelete in TPDBRecordHeader(PDBRecordList.Items[I]).RecFlags) shl 7));
        Move(TPDBRecordHeader(PDBRecordList.Items[I]).UniqueID[0], PDBRecordHeaderNative.UniqueID[0], SizeOf(PDBRecordHeaderNative.UniqueID));
        OutStream.WriteBuffer(PDBRecordHeaderNative, SizeOf(PDBRecordHeaderNative));
    end;

    //Loop through the PDBRecordList and write all the record data
    for I := 0 to (PDBRecordList.Count - 1) do
    begin
        OutStream.WriteBuffer(TPDBRecordHeader(PDBRecordList.Items[I]).RecordData^, TPDBRecordHeader(PDBRecordList.Items[I]).Size);
    end;

    //Set the file size
    DBFileSize := OutStream.Position;

    //Finished with output
    OutStream.Free;

    //Deallocate all the record data
    for I := 0 to (PDBRecordList.Count - 1) do
    begin
        StrDispose(TPDBRecordHeader(PDBRecordList.Items[I]).RecordData);
        TPDBRecordHeader(PDBRecordList.Items[I]).RecordData := nil;
    end;
end;

end.
