unit Settings;

//This unit manages loading/saving settings in the .INI file and
//provides a way for the main form (DocReaderUnit.pas) and the
//settings form (frmSettingsUnit.pas) to share the settings data
interface
uses filectrl, inifiles, sysutils, graphics;

type
    //Stores info for each MRU entry
    MRUEntry = record
      FileName: String;
      Position: Integer;  //Position in the document
      MenuItem: Pointer;  //Pointer to a TMenuItem
    end;

    TSettings = class(TObject)
        IniFileName: String;
        //
        DocDirectory: String;
        Color: Integer;  //background
        Font: TFont;
        //Window size
        WindowTop: Integer;
        WindowLeft: Integer;
        WindowHeight: Integer;
        WindowWidth: Integer;
        //Printing
        TopMargin: Real;
        BottomMargin: Real;
        LeftMargin: Real;
        RightMargin: Real;

        //Other
        EditMode: Boolean;
        BookmarkScan: Boolean;
        DesktopInstalled: Boolean;
        IntegrateWithDesktop: Boolean;
        ShowPeanutCodes: Boolean;
        //
        MRUList: array[1..4] of MRUEntry;

        constructor Create(IniFileName: String);
        destructor Destroy; override;
        procedure ReadSettings;
        procedure WriteSettings;
    end;

function PilotDesktopInstalled: Boolean;

implementation

uses DocReaderUnit, AutoInst;


function PilotDesktopInstalled: Boolean;
var
    TestUserData: TUserData;
begin
    //Tell whether or not the Desktop software is installed
    TestUserData := nil;
    Result := True;
    try
      TestUserData := TUserData.Create;
    except
      on EPalmDesktopError do
      begin
          Result := False;
      end;
    end;
    if TestUserData <> nil then
        TestUserData.Free;
end;


constructor TSettings.Create(IniFileName: String);
begin
    Self.IniFileName := IniFileName;
    ReadSettings;
end;


procedure TSettings.ReadSettings;
var
    IniFile: TIniFile;
    I: Integer;
begin
    //Read the .INI file and set up values
    IniFile := TIniFile.Create(IniFileName);

    //Get the Doc directory
    DocDirectory := IniFile.ReadString('General', 'DefaultDirectory', ExtractFileDir(IniFileName));

    //Read the MRU list
    for I := low(MRUList) to high(MRUList) do
    begin
       	MRUList[I].FileName := IniFile.ReadString('LastOpenFiles', Format('File%2.2d', [I]), '');
       	MRUList[I].Position := IniFile.ReadInteger('LastOpenPositions', Format('File%2.2d', [I]), 0);
    end;

    Font := TFont.Create;

    //Set the font and background color
    Font.Name := IniFile.ReadString('Font', 'Name', frmDocReader.reText.DefAttributes.Name);
    Font.Size := IniFile.ReadInteger('Font', 'Size', frmDocReader.reText.DefAttributes.Size);
    Font.Color := IniFile.ReadInteger('Font', 'Color', frmDocReader.reText.DefAttributes.Color);
    Color := IniFile.ReadInteger('General', 'Color', frmDocReader.reText.Color);

    //Window size and position
    WindowTop := IniFile.ReadInteger('General', 'Top', frmDocReader.Top);
    WindowLeft := IniFile.ReadInteger('General', 'Left', frmDocReader.Left);
    WindowHeight := IniFile.ReadInteger('General', 'Height', frmDocReader.Height);
    WindowWidth := IniFile.ReadInteger('General', 'Width', frmDocReader.Width);

    //Print margins
    TopMargin := IniFile.ReadFloat('Margins', 'Top', 0);
    BottomMargin := IniFile.ReadFloat('Margins', 'Bottom', 0);
    LeftMargin := IniFile.ReadFloat('Margins', 'Left', 0);
    RightMargin := IniFile.ReadFloat('Margins', 'Right', 0);

    //Editing mode, Bookmark scan, integration, etc.
    EditMode := IniFile.ReadBool('General', 'EditMode', False);
    BookmarkScan := IniFile.ReadBool('General', 'BookmarkScan', True);
    IntegrateWithDesktop := IniFile.ReadBool('General', 'IntegrateWithDesktop', False);
    ShowPeanutCodes := IniFile.ReadBool('General', 'ShowPeanutCodes', False);

    //Desktop installed
    DesktopInstalled := PilotDesktopInstalled;

    IniFile.Free;
end;

procedure TSettings.WriteSettings;
var IniFile: TIniFile;
    I: Integer;
begin
    IniFile := TIniFile.Create(IniFileName);
    //Update the INI file (try..except in case they're running from CD-ROM, etc.)
    //MRU entries
    for I := low(MRUList) to high(MRUList) do
    try
        IniFile.WriteString('LastOpenFiles', Format('File%2.2d', [I]), MRUList[I].FileName);
        IniFile.WriteInteger('LastOpenPositions', Format('File%2.2d', [I]), MRUList[I].Position);
    except
    end;

    try
        //Font name and size, edit mode
        IniFile.WriteString('Font', 'Name', Font.Name);
        IniFile.WriteInteger('Font', 'Size', Font.Size);
        IniFile.WriteInteger('Font', 'Color', Font.Color);
        IniFile.WriteInteger('General', 'Color', Color);
        IniFile.WriteString('General', 'DefaultDirectory', DocDirectory);
        IniFile.WriteBool('General', 'EditMode', EditMode);
        IniFile.WriteBool('General', 'BookmarkScan', BookmarkScan);
        IniFile.WriteBool('General', 'IntegrateWithDesktop', IntegrateWithDesktop);
        //Window size and position
        IniFile.WriteInteger('General', 'Top', WindowTop);
        IniFile.WriteInteger('General', 'Left', WindowLeft);
        IniFile.WriteInteger('General', 'Height', WindowHeight);
        IniFile.WriteInteger('General', 'Width', WindowWidth);
        //Margins
        IniFile.WriteFloat('Margins', 'Top', TopMargin);
        IniFile.WriteFloat('Margins', 'Bottom', BottomMargin);
        IniFile.WriteFloat('Margins', 'Left', LeftMargin);
        IniFile.WriteFloat('Margins', 'Right', RightMargin);
    except
    end;

    //Release the TIniFile object
    IniFile.Free;
end;

destructor TSettings.Destroy;
begin
    Font.Free;
end;

end.
