/*********************************************************************
 * Filename:      doorkeeper.c
 * Description:   Un*x server component 
 * Author:        Per Harald Myrvang <perm@stud.cs.uit.no>
 * 	
 * 	Copyright University of Troms (Norway)
 * 	See the file COPYING for details
 ********************************************************************/


/* This is the UN*X component, i.e. the server.
*/

/* TODO:
 *	- When doorkeeper detects that binary data is comming over the port, it
 *	  should send hotsync'ish data back, telling the device to shut up and
 *	  go away.
 *	- Add ACL for control-channel, or at least log who does what!
 *	- Make some control-channel clients which uses the authentication
 *	  scheme.
 *	- Add support for multiple users with individual shared secrets. This can
 *	  simply be implemented by having the username specified at request
 *	  initiation time (from the pilot), then this username is used to look up
 *	  the secret, and the rest proceeds as usual. The Pilot DoorKnob
 *	  application already prepends the username to the initial request. All
 *	  doorkeeper now has to do, is to keep track of which user has which
 *	  secret. Configfile? Lex? Eh? Eh? Eh? ***DONE***
 *	- S/KEY support?
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <time.h>
#include <strings.h>
#include <assert.h>
#include <sys/errno.h>
#include <signal.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <ctype.h>



#include <md5.h>		/* SSLeay 0.8.1 at least */

#include "common.h"

#define VERSION "0.9.3b"

#define DOOR_MODE_GET	0
#define DOOR_MODE_SET	1

#define DEFAULT_CONTROLPORT	4242

/* The maximum allowed number of seconds to delay before automagically locking.
 */
#define MAX_AUTOTIMEOUT (60 * 2)

#ifndef MAX
#define MAX(x,y) ((x)>(y)?(x):(y))
#endif

static int	verbose_mode = 0;
static FILE	*LOG = NULL;
static int	autotimeout = 5;
static char *pilot_device = "/dev/pilot";
static char *door_device = "/dev/lpt0";
static int	initial_state = DOOR_STATE_LOCK;
static int	door_state = DOOR_STATE_OPEN;
static int	autolock_state = 1;		/* 0=off, 1=on */

typedef struct {
	int		socket;
	int		port;
	int		current_state;
} ControlChannelType;

static ControlChannelType ctrl_channel = {-1, -1};

typedef struct {
	char name[USERNAME_MAXLEN + 1];
	char secret[SECRET_MAXLEN + 1];
	int flags;
} UserTableEntryType;

static UserTableEntryType user_table[101]; /* Ok, so this is nasty, but... */
static int user_table_entries = 0;
static char *user_table_filename = "./user_secrets";


static const char hs[] = "0123456789abcdef";	/* used by hexify below */

/* Turn a sequence of bytes into a hexadecimal ascii-notation.  
 * Programmer beware: We will always require double the size 
 * for the hex-encoded data!
 *  This function will terminate the to-string. Thus the 
 * required size of the to-buffer is (strlen(from)*2)+1 
 */
static void
hexify(unsigned char *from, char *to, int len)
{
    int i,l;
    for (l = i = 0; l < len; l++) {
	to[i++] = hs[((from[l] >> 4) & 0x0F)];
	to[i++] = hs[(from[l] & 0x0F)];
    }
    to[i] = '\0';
}

/* 
 * The reverse of hexify
 */
static void
unhexify(unsigned char *from, char *to, int len)
{
    int i,l;
    for (l = i = 0; l < len; l++) {
	/* error-checking is for wimps! */
	to[l] = (((from[i++] - '0') << 4) | (from[i++] - '0'));
    }
}



/* 
 * Removes any trailing linefeeds and carriage returns 
 */
char *
chop(char *s)
{
    char *q;
    if (NULL != s) {
	q = strchr(s, '\n');
	if ( q != NULL )
	    *q = '\0';
	q = strchr(s, '\r');
	if ( q != NULL ) 
	    *q = '\0';
    }
    return s;
}


/* From authtime.c
 * Convert a timeval (as returned by gettimeofday(2)) 
 * to a SDSI-conformant time/date format. <outbuf> should 
 * be no less than 30 chars.  
 * Returns 0 on success, -1 on failure.
 */
int
tv2sdsi(struct timeval *tv, char *outbuf)
{
    struct tm	*tm;
    time_t		t = tv->tv_sec;
	
    tm = localtime(&t);
    sprintf(outbuf, "%04d-%02d-%02dT%02d.%02d.%02d.%03d%+05d",
	    tm->tm_year + 1900,
	    tm->tm_mon + 1,
	    tm->tm_mday,
	    tm->tm_hour,
	    tm->tm_min,
	    tm->tm_sec,
	    (int)(tv->tv_usec/1000),
#if (defined(__NetBSD__) || defined(__FreeBSD__)) /* ...and other stupid non-POSIX systems */
	    (int)(-(tm->tm_gmtoff/36))
#else
	    (int)(-(timezone/36))
#endif
	);
    return 0;
}

/* Generates a SDSI-timestamp, and returns a pointer to a static const buffer
 */
static const char *
get_sdsi_timestamp(void)
{
    struct timeval	tv;
    static char	sdsi_date[30];
    
    /* Get timestamp for user-lookup and request-log */
    assert(0 == gettimeofday(&tv, NULL));	/* Get timestamp */
    tv2sdsi(&tv, sdsi_date);
    
    return sdsi_date;
}


/* Attempts to read a maximum of <count> bytes into the buffer <data> from the
 * serial device <fd>. Returns the number of bytes read, or -1 if an error
 * occured.   Since we have an alarm-signal pending at all (most) times, we'll
 * ignore any EINTR error codes from select(2) 
 */
static int
s_read(int fd, char *data, int count)
{
    int		nread;
    fd_set	ready;
    int		r;

    r = 0;
    while (0 >= r) {			/* Set up to ignore any EINTR errors */
	nread = 0;
	FD_ZERO(&ready);
	FD_SET(fd, &ready);
	r = select(fd+1, &ready, 0, 0, NULL);
	if (0 >= r) {
	    if (EINTR == errno)
		continue;
	    perror("s_read: select(2):");
	}
    }
    /* if data is available in time, read it */
    if(FD_ISSET(fd, &ready)) {		
	nread = read(fd, data, count);
    } else {
	fprintf(LOG, "s_read: Serial RX timeout (Should NOT happen)!\n");
	nread = -1;
    }
    return nread;
}

/* Get a newline (\n) terminated string, up to <len> bytes. 
 * The string will be chopped (i.e. the newline removed). 
 * Return the number of characters in the string, or -1 if 
 * some failure occured. 
 */
static int
s_gets(int fd, char *data, int len)
{
    int num, nread;
    char *s;
    int r;
	
    num = nread = 0;
    memset(data, 0, len);		/* Pre-zero-terminate string */
    while(num < len) {
	nread = s_read(fd, data + num, len - num);
	if (nread <= 0) {
	    perror("s_gets");
	    return -1;
	}
	s = strchr(data, '\n');
	if (NULL != s ) {
	    *s = '\0';
	    break;
	}

	/* If we start reading binary data, where null-chars will occure (as in
	 * the case where the user hits the hotsync-button, we just drop
	 * everything, and return with an error! */
	for(r = num; r < (num + nread); r++) {
	    if (!isprint(data[r]))
		return -1;
	}

	num += nread;
    }
    return strlen(data);
}

/* Attempts to write <count> bytes from the buffer <data> to 
 * the serial device <fd>. Returns the number of bytes written, 
 * or the negated number of bytes written, if this was less 
 * than <count>.
 */
static int
s_write(int fd, char *data, int count)
{
    int len, nwrote;

    len = 0;
    while (len < count) {
	nwrote = 0;
	nwrote = write(fd, data + len, count);
	if (nwrote <= 0) {
	    len = -len;
	    break;			/* transmission failure */
	}	
	len += nwrote;
    }
    return len;
}


/* This is the place where the door-device is opened/locked/
 * read/written to. It should be fairly portable, at least it 
 * doesn't do peek/poke to the parallel port. Now, if you use 
 * some other device/setup than we're using at the PASTA
 * project, change this to your needs!
*/
static int
door_handle(char *door_device, int mode, int state)
{
    int device;
    int ret;
    char tinybuf;
    int r;

    ret = -1;
    tinybuf = 0;
    /* Open door-device */
#ifdef linux
    device = open(door_device, O_RDWR|O_SYNC);
#else
    device = open(door_device, O_RDWR|O_NONBLOCK);
#endif
    if (device < 0) {
	perror("door_handle() - open door-device failed!");
	return -1;
    }

    if (DOOR_MODE_SET == mode) {
	if (DOOR_STATE_OPEN == state)
	    tinybuf = '\x3';
	else
	    tinybuf = '\0';
	r = s_write(device, &tinybuf, 1);
	if (verbose_mode)
	    fprintf(LOG, "wrote 0x%x to door-device\n", (int)tinybuf);
	if (r != 1)
	    fprintf(LOG, "s_write to door-device %s returned %d\n", 
		    door_device, r);
	else
	    ret = 0;

    } else {
	fprintf(LOG, "**ERROR: cannot get state of door_device (yet)!\n");
    }
	
    close(device);
    return ret;
}


/* Lookup a user's secret, given his/hers name. A username is 
 * defined to be up to 10 characters or a space. It is thus not 
 * neccessary to do any massaging before calling this function. 
 * Returns either a pointer to the user's secret, or NULL if 
 * user wasn't found. 
 * Note that empty secrets are possible (yuck!) 
 */
static char *
lookup_usersecret(char *username)
{
    int i, l, r;

    for(l = 0; l <= 10; l++)
	if ((' ' == username[l]) || ('\0' == username[l]))
	    break;
    if (l > 10)
	l = 10;
    for (i = 0; i < user_table_entries; i++) {
	r = strncmp(user_table[i].name, username, l);
	if (0 == r) {
	    if (verbose_mode)
		fprintf(stderr,
			"Lookup_usersecret: user '%s' has secret '%s'\n", 
			username, user_table[i].secret);
	    return &user_table[i].secret[0];
	}
    }
    return NULL;
}


/* Setup socket, i.e. do all boring stuff like bind and setsockopt.
 * Input: <port> - portnumber to listen at
 *        <sin>  - a pointer to a sockaddr_in struct
 * Returns: -1 on failure, or a valid socket-descriptor.
 */
int
socket_setup(	int port,
		struct sockaddr_in *sin)
{
    int sock, optval, i;
    struct protoent *pe;

    assert((pe = getprotobyname("tcp")));
    sock = socket(AF_INET, SOCK_STREAM, pe->p_proto);
    if (-1 != sock) {
	optval = 1;
	i = setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, 
		       &optval, sizeof(optval));
	if (-1 != i) {
	    memset(sin, 0, sizeof(struct sockaddr_in));
	    sin->sin_family = AF_INET;
	    sin->sin_port = htons(port);
	    sin->sin_addr.s_addr = htonl(INADDR_ANY);
	    i = bind(sock, (struct sockaddr *)sin, sizeof(struct sockaddr_in));
	    if (0 <= i) {
		i = listen(sock, 5);
		if (-1 != i) {
		    return sock;
		}
		else
		    perror("listen");
	    }
	    else
		perror("bind");
	}
	else
	    perror("setsockopt");
	shutdown(sock, 2);
	close(sock);
    }
    else
	perror("socket");

    return -1;
}




/* Handle any nagging from the control channel */
int
handle_control_channel(	ControlChannelType *control_channel,
			char 	*door_device) 

{
    char	inbuf[256];
    char	buf[256];
    const char	*challenge;
    char	*result_txt[] = {"OK", "FAILED", NULL};
    char	*result = result_txt[1]; /* Failure is default */
    char	mdbuf[16];	/* MD5 buffer */
    char	mdtxt[40];	/* MD5 buffer (hexified) */
    char	*token;

    struct sockaddr	sa;
    int		sl;
    int		sock;
    FILE	*SOCK;
    char	*adm_secret;
    fd_set	ready;
    struct timeval tv;

    char	*s;
    int		r;

    sock = accept(control_channel->socket, &sa, &sl);
			
    if (sock < 0) {
	fprintf(LOG, "%s: accept failed!\n", get_sdsi_timestamp());
	return -1;
    }

    SOCK = fdopen(sock, "r+");
    if (NULL == sock) {
	fprintf(LOG, "%s: fdopen failed!\n", get_sdsi_timestamp());
	return -1;
    }

    /* XXX find out who's calling, i.e. ACL! */

    /* HACK-WARNING! To prevent someone from keeping the command-channel locked
     * up forever, we let select wait for 2 seconds.  And bacon. */

    FD_ZERO(&ready);
    FD_SET(sock, &ready);
    tv.tv_sec = 2;
    tv.tv_usec = 0;

    r = select(sock + 1, &ready, 0, 0, &tv);
    if (r < 1) {				/* If no ready data, die. */
	fprintf(LOG, "%s: control-channel was too late giving commands!\n", 
		get_sdsi_timestamp());
	fclose(SOCK);
	return -1;
    }
     
    s = fgets(inbuf, sizeof(inbuf), SOCK);	/* See whats on control channel */
    if (NULL == s) {				/* If nothing on channel, just exit. */
	fprintf(LOG, "%s: Couldn't read a line from control-channel!\n", 
		get_sdsi_timestamp());
	fclose(SOCK);
	return -1;
    }

    chop(inbuf);		/* We don't want content-free commands. */
    if (inbuf[0] == '\0') {
	fprintf(LOG, "%s: Read empty string from control-channel!\n", 
		get_sdsi_timestamp());
	fclose(SOCK);
	return -1;
    }

    /* But before we to anything, do some authentication */
    adm_secret= lookup_usersecret("Administrator");
    if (NULL == adm_secret) {
	fprintf(LOG, "%s: Cannot find administrator secret - assuming there's none (DANGEROUS)!\n", 
		get_sdsi_timestamp());
	adm_secret = "";
	/* "response-state" has no errors. *ahem* */
	r = 0;				
    } else {
	/* Log attempt */
	fprintf(LOG, "%s: received '%s' request from control-channel\n", 
		get_sdsi_timestamp(), buf);
	/* Get timestamp for use as challenge. Do NOT call
	 * get_sdsi_timestamp() before _after_ we've computed the
	 * response! Why? challenge is a pointer to a static buffer. */
	challenge = get_sdsi_timestamp();

	if (verbose_mode)
	    fprintf(LOG, "Sending challenge \"%s\" to control-channel\n", 
		    challenge);

	/* Build and send challenge */
	fflush(SOCK);
	fprintf(SOCK, "%s\n", challenge);
	fflush(SOCK);

	/* Compute correct response to our challenge */
	strcpy(buf, challenge);
	strcat(buf, adm_secret);
	MD5(buf, strlen(buf), mdbuf);	/* Compute MD5 */
	hexify(mdbuf, mdtxt, 16);

	/* Read response */
	fgets(buf, sizeof(buf), SOCK);
	chop(buf);

	r = strcmp(buf, mdtxt);		/* Determine response-status */
    }
    if (0 == r ) {			/* Correct response? */
	fprintf(LOG, "%s: control channel authenticated!\n", get_sdsi_timestamp());

	/* Now see what it wanted */
	token = strtok(inbuf, " \t");		/* Get first token */
	r = strcasecmp(token, "command");	/* Is it a command instruction? */
	if (0 == r) {
	    token = strtok(NULL, " \t");		/* Get next token */
	    r = strcasecmp(token, "lock");		/* Is it a lock-action? */
	    if (0 == r) {
		token = strtok(NULL, " \t");		/* Get next token */
		if (NULL != token) {
		    r = strcasecmp(token, "auto");		/* Is it a auto-qualifier? */
		    if (0 == r) {
			/* Lock door, and enable auto-lock feature */
			autolock_state = 1;
			door_state = DOOR_STATE_LOCK;
			r = door_handle(door_device, DOOR_MODE_SET, door_state);
			result = result_txt[0];	/* OK result */
		    } else {
			r = strcasecmp(token, "set");		/* Is it a set-qualifier? */
			if (0 == r) {
			    /* Set auto-lock feature timeout in seconds */
			    token = strtok(NULL, " \t");
			    if (NULL != token) {
				autotimeout = atoi(token);
				if ((autotimeout < 1) || 
				    (autotimeout > MAX_AUTOTIMEOUT)) {
				    fprintf(LOG, "%s: Illegal set lock command option \"%s\"! Set back to 5\n", 
					    get_sdsi_timestamp(), token);
				    autotimeout = 5;
				    result = result_txt[0];	/* OK result */
				}
			    } else {
				fprintf(LOG, "%s: Unknown set lock command option \"%s\"!\n", 
					get_sdsi_timestamp(), token);
			    }

			} else {
			    fprintf(LOG, "%s: Unknown lock command qualifier \"%s\"!\n", 
				    get_sdsi_timestamp(), token);
			}
		    }
		} else {		
		    /* i.e. no qualifier 
		     * Lock door, but do not touch the 
		     * auto-lock feature 
		     */
		    door_state = DOOR_STATE_LOCK;
		    r = door_handle(door_device, DOOR_MODE_SET, door_state);
		    result = result_txt[0];	/* OK result */
		}
	    } else {
		r = strcasecmp(token, "open");		/* Is it an open-action? */
		if (0 == r) {
		    token = strtok(NULL, " \t");		/* Get next token */
		    if (NULL != token) {
			r = strcasecmp(token, "keep");		/* Is it a keep-qualifier? */
			if (0 == r) {
			    /* Open door, and disable auto-lock feature */
			    autolock_state = 0;
			    door_state = DOOR_STATE_OPEN;
			    r = door_handle(door_device, DOOR_MODE_SET, door_state);
			    result = result_txt[0];	/* OK result */
			} else {
			    fprintf(LOG, "%s: Unknown open command qualifier \"%s\"!\n", get_sdsi_timestamp(), token);
			}
		    } else {		
			/* i.e. no qualifier 
			 * Open door, but do not touch the 
			 *auto-lock feature 
			 */
			door_state = DOOR_STATE_OPEN;
			r = door_handle(door_device, DOOR_MODE_SET, 
					door_state);
			result = result_txt[0];	/* OK result */

			/* See if we aught to set the alarm or not. */
			if (autolock_state) {
			    /* Start auto-timeout timer. It will automagically lock the
			     * door after <autotimeout> seconds. */
			    alarm(autotimeout);
			} else {
			    alarm(0);		/* Cancel any pending alarms */
			}
		    }
		} else {
		    fprintf(LOG, "%s: Unknown command action \"%s\"!\n", 
			    get_sdsi_timestamp(), token);
		}
	    }

	} else {	/* if "command" */
	    r = strcasecmp(token, "query"); /* Is it a query instruction? */
	    if (0 == r) {
		token = strtok(NULL, " \t"); /* Get next token */
	    } else {			/* Ah, it's an unknown command! */
		fprintf(LOG, "%s: Unknown command \"%s\"!\n", 
			get_sdsi_timestamp(), token);
	    }
	}

    } else {			/* Failure */
	fprintf(LOG, "%s: control channel authentication FAILED!\n", 
		get_sdsi_timestamp());
    }

    /* Send a result back */
    fflush(SOCK);
    fprintf(SOCK, "%s\n", result);
    fflush(SOCK);

    fclose(SOCK);
    return 0;
}


/* This is where the action happens. */
static int
do_server(char 	*pilot_device, 
	  char 	*door_device, 
	  ControlChannelType *control_channel,
	  int 	initial_state)
{
    struct termios	tios;
    int			device;
    char		*secret; 
    struct timeval	tv;
    char		sdsi_date[30];
    char		mdbuf[16];	/* MD5 buffer */
    char		mdtxt[40];	/* MD5 buffer (hexified) */
    char		buf[256];
    char		*s;
    int			request_state; /* Requested door state: -1 - unknown */
    fd_set		ready;
    /* scratch */
    int			r;
    char		*p;

    secret = "";
    request_state = -1;

    /* Set door to initial state */
    r = door_handle(door_device, DOOR_MODE_SET, initial_state);
    if (r != 0) {
	fprintf(LOG, "**ERROR: couldn't set door to specified initial state!");
	return -1;
    }
    door_state = initial_state;


    /* Open pilot link device */
#ifdef linux
    device = open(pilot_device, O_RDWR|O_SYNC);
#else
    device = open(pilot_device, O_RDWR|O_NONBLOCK);
#endif

    if (device < 0) {
	perror("open pilot_device failed!");
	exit(1);
    }

    /* Setup IO-flags */
    assert(!tcgetattr(device, &tios));
    tios.c_iflag |= IGNBRK | IGNPAR;
    tios.c_oflag |= 0;
    tios.c_cflag |= CREAD | CLOCAL | CS8;
    tios.c_lflag |= NOFLSH;
    cfmakeraw(&tios);
    cfsetospeed(&tios, B9600);
    cfsetispeed(&tios, B9600);
    tcsetattr(device, TCSANOW, &tios);


    /* Main loop */
    while(1) {
	if (verbose_mode) {
	    fprintf(LOG, "\n\nWaiting for pilot-device to make contact...\n");
	    fprintf(LOG, "Door state: %s\n", door_state ? "locked" : "open");
	}
	tcflush(device, TCIOFLUSH);	/* Flush any old junk */
	tcdrain(device);

	FD_ZERO(&ready);
	FD_SET(device, &ready);
	FD_SET(control_channel->socket, &ready);

	/* Wait for either pilot or control-channel to come nagging */
	r = select(MAX(device, control_channel->socket) + 1, &ready, 0, 0, NULL);
	if (r < 1)
	    continue;

	if (FD_ISSET(control_channel->socket, &ready)) { /* control channel */
	    handle_control_channel(control_channel, door_device);
	} else if (FD_ISSET(device, &ready)) {			/* pilot device */
	    /* See what the pilot wants */
	    r = s_gets(device, buf, sizeof(buf)); 
	    if (-1 == r) {
		if (verbose_mode)
		    fprintf(LOG, "Reading from pilot-device failed!\n");
		continue;
	    }

	    if (verbose_mode)
		fprintf(LOG, "Read %d bytes from pilot-device: '%s'\n", r, buf);

	    /* Have a look at what we've read, and make some decisions */
	    p = strchr(buf, ' ');	/* Skip first token (i.e. name) */
	    if (NULL == p)		/* If nothing, drop all and redo from start */
		continue;
	    else
		p++;
	    r = strcmp(p, REQUEST_OPEN);	/* Is it a open request? */
	    if (0 == r)
		request_state = 0;
	    else {
		r = strcmp(p, REQUEST_LOCK);	/* Is it a lock request? */
		if (0 == r)
		    request_state = 1;
		else
		    request_state = -1;		/* It is a bogus request. */
	    }
	    /* Deal with bogus requests, e.g. if anyone tries to hotsync. */
	    if (-1 == request_state) {
		fprintf(LOG, "%s: Bogus request '%s' - Ignored!\n", 
			get_sdsi_timestamp(), buf);
#if 0
		tcsetattr(device, TCSANOW, &tios);	/* We even reset the attributes to be sure */
#endif
		tcflush(device, TCIOFLUSH);		/* Flush any old junk */
		
		continue;		/* continue from top of main loop! */
	    }

	    if (verbose_mode)
		fprintf(LOG, "Requested state: %s (%d)\n", 
			request_state ? "locked" : "open", request_state);


	    /* Try to find secret */
	    s = strchr(buf, ' ');
	    if (NULL != s) {
		*s = '\0';
		s = buf;
		if (*s == '\x8')	/* Sometime we receive a backspace before anything else. */
		    s++;		/* If this happens, look past it. */
		secret = lookup_usersecret(s);
		if (NULL == secret) {
		    request_state = -1;
		    if (verbose_mode)
			fprintf(LOG, "Couldn't lookup user \"%s\" pilot-device\n", s);
		}
	    } else
		request_state = -1;

	    if (-1 == request_state) {
		fprintf(LOG, "%s: %s request from user '%s' NOT properly authenticated\n",
			get_sdsi_timestamp(), request_state ? "locked" : "open", buf);
		r = s_write(device, ACK_FAILED, strlen(ACK_FAILED));
		if (verbose_mode)
		    fprintf(LOG, "Wrote %d bytes to pilot-device\n", r);
		continue;			/* continue from top of main loop! */
	    }

	    /* Log attempt */
	    fprintf(LOG, "%s: received %s request from user '%s'\n",
		    get_sdsi_timestamp(), request_state ? "locked" : "open", buf);

	    assert(0 == gettimeofday(&tv, NULL));	/* Get timestamp */
	    tv2sdsi(&tv, sdsi_date);			/* Build challenge */

	    if (verbose_mode)
		fprintf(LOG, "Sending challenge \"%s\" to pilot-device\n", 
			sdsi_date);

	    r = s_write(device, sdsi_date, strlen(sdsi_date));
	    if (verbose_mode)
		fprintf(LOG, "Wrote %d bytes to pilot-device\n", r);

	    strcpy(buf, sdsi_date);
	    strcat(buf, secret);
	    MD5(buf, strlen(buf), mdbuf);	/* Compute MD5 */
	    hexify(mdbuf, mdtxt, 16);

	    if (verbose_mode)
		fprintf(LOG, "Waiting for pilot-device to respond: '%s'\n", 
			mdtxt);

	    s_gets(device, buf, sizeof(buf));
	    if (verbose_mode)
		fprintf(LOG, "Read %d bytes from pilot-device:     '%s'\n", 
			r, buf);

	    /* Log response to challenge. */
	    r = strcmp(buf, mdtxt);
	    if (0 == r ) {			/* Correct response? */
		if (verbose_mode)
		    fprintf(LOG, "Got correct response from pilot-device\n");
		fprintf(LOG, "%s: %s request authenticated\n", 
			get_sdsi_timestamp(), request_state ? "locked" : "open");

		/* Decide whether the response should be ACK_OK 
		 * or ACK_NOEFFECT 
		 */
		if ((door_state == request_state) || (request_state == -1))
		    r = s_write(device, ACK_NOEFFECT, strlen(ACK_OK));
		else {
		    door_state = request_state;
		    request_state = (door_state == DOOR_STATE_OPEN) 
			?  DOOR_STATE_OPEN : DOOR_STATE_LOCK;
		    r = door_handle(door_device, DOOR_MODE_SET, request_state);
		    if (verbose_mode)
			fprintf(LOG, "door_handle() returned %d\n", r);
		    r = s_write(device, ACK_OK, strlen(ACK_OK));

		    /* See if we aught to set the alarm or not. */
		    if (autolock_state) {
			/* Start auto-timeout timer. It will automagically lock the
			 * door after <autotimeout> seconds. */
			alarm(autotimeout);
		    } else {
			alarm(0);		/* Cancel any pending alarms */
		    }
		}

		if (verbose_mode)
		    fprintf(LOG, "Wrote %d bytes to pilot-device\n", r);
	    } else {			/* Failure */
		if (verbose_mode)
		    fprintf(LOG, "Got incorrect response from pilot-device\n");
		fprintf(LOG, "%s: %s request NOT properly authenticated\n", 
			get_sdsi_timestamp(), request_state ? "locked" : "open");
		r = s_write(device, ACK_FAILED, strlen(ACK_FAILED));
		if (verbose_mode)
		    fprintf(LOG, "Wrote %d bytes to pilot-device\n", r);
	    }

	    tcdrain(device);		/* Wait for completion of write. */
		
	}	/* if (FD_ISSET(device, &ready) */
    } /* while(1) */
}

void 
alarm_handler()
{
    int i;

    i = (int)signal(SIGALRM, alarm_handler);
    if (-1 == i) {
	perror("alarm_handler failed setting alarm-signal");
	exit(1);	/* Eh... Is this a Good Thing?? */
    }
    if (autolock_state) {
	/* lock door */
	i = door_handle(door_device, DOOR_MODE_SET, DOOR_STATE_LOCK);
	if (i != 0)
	    fprintf(stderr, "**ERROR: alarm_handler() couldn't lock door!\n");
	else
	    door_state = DOOR_STATE_LOCK;
    }
}

static void
usage(char *progname)
{
    fprintf(stderr, "Usage:\t%s [-d pilot-device] [-o door-device] [-s initial-state]\n", progname);
    fprintf(stderr, "\t\t[-t timeout] [-L logfile] [-f user-secrets] [-v]\n");
    fprintf(stderr, "\t-d <pilot-device> defaults to \"%s\"\n", pilot_device);
    fprintf(stderr, "\t-o <door-device> defaults to \"%s\"\n", door_device);
    fprintf(stderr, "\t-s <initial-state> defaults to \"lock\" (\"open\")\n");
    fprintf(stderr, "\t-t <autolock-timeout> defaults to 5 (seconds)\n");
    fprintf(stderr, "\t-L <logfile> defaults to stdout. Output is appended.\n");
    fprintf(stderr, "\t-f <user-secrets> defaults to \"./user_secrets\".\n");
    fprintf(stderr, "\t-c <controll-channel port> defaults to %d.\n", DEFAULT_CONTROLPORT);
    fprintf(stderr, "\t-v for verbose output\n");
    fprintf(stderr, "Version " VERSION " BETA " __DATE__ " " __TIME__ "\n");
    fprintf(stderr, "Brought to you by the PASTA-project at the University of Troms, Norway\n");
    fprintf(stderr, "Code by Per Harald Myrvang <perm@stud.cs.uit.no>\n");
    exit(0);
}

void
main(int argc, char **argv)
{
    int ch;
    char *progname;
    int tfd, old_lflag;
    struct termios tios;
    FILE *FP;
    struct sockaddr_in	sin;
    /* scratch */
    int i;

    progname = argv[0];
    /* Logfile defaults to standard-output */
    LOG = stdout;			
    ctrl_channel.port = DEFAULT_CONTROLPORT;

    /* Process command-line arguments */
    while((ch = getopt(argc, argv, "d:o:s:t:L:vh")) != -1) {
	switch(ch) {
	case 'd':			/* set pilot-device */
	    pilot_device = optarg;
	    break;
	case 'o':			/* set door/output-device */
	    door_device = optarg;
	    break;
	case 's':			/* set initial-state */
	    if (0 == strcasecmp(optarg, "open"))
		initial_state = DOOR_STATE_OPEN;
	    else {
		i = strcasecmp(optarg, "lock");
		if (0 == i )
		    initial_state = DOOR_STATE_LOCK;
		else {
		    fprintf(stderr, "Initial-state can ONLY be \"open\" or \"lock\"!\n");
		    exit(0);
		}
	    }
	    break;
	case 't':
	    autotimeout = atoi(optarg);
	    if ((autotimeout < 1) || (autotimeout > MAX_AUTOTIMEOUT)) {
		fprintf(stderr, "**WARNING: Autotimeout value unrealistic! Must be between 1 and %d!\n",
			MAX_AUTOTIMEOUT);
		fprintf(stderr, "Autotimeout value set to default!\n");
		autotimeout = 5;
	    }
	    break;
	case 'L':
	    LOG = fopen(optarg, "a+");
	    if (NULL == LOG) {
		fprintf(stderr, "**WARNING: Couldn't use '%s' as logfile.  (%s)\n",
			optarg, strerror(errno));
		fprintf(stderr, "Logfile set to standard output!\n");
		LOG = stdout;
	    }
	    break;
	case 'f':
	    user_table_filename = optarg;
	    break;
	case 'v':
	    verbose_mode = 1;
	    break;
	case 'c':
	    ctrl_channel.port = atoi(optarg);
	    if ((ctrl_channel.port< 1) || (ctrl_channel.port > 0xffff)) {
		ctrl_channel.port = DEFAULT_CONTROLPORT;
		fprintf(stderr, "control port is out of range! Reset to default port #%d!\n",
			ctrl_channel.port);
	    }
	    break;
	case 'h':
	default:
	    usage(progname);
	    break;
	}
    }

    /* Set up command_channel socket and stuff */
    ctrl_channel.socket = socket_setup(ctrl_channel.port, &sin);
    if (-1 == ctrl_channel.socket) {
	fprintf(stderr, "socket_setup() failed!\n");
	exit(1);
    }

    
    /* Open usertable file */
    FP = fopen(user_table_filename, "r");
    if (NULL != FP) {
	char inbuf[256], *s, *t;
	int line = 0, i;
	for (user_table_entries = 0; user_table_entries < 100;) {
	    s = fgets(inbuf, sizeof(inbuf), FP);
	    if (NULL != s) {
		line++;
		chop(s);
		t = strchr(s, '#');	/* Find and kill comments */
		if (NULL != t)
		    *t = '\0';
		i = strlen(s);		/* Check length of line */
		if (i < 1)		/* ignore empty lines */
		    continue;
		t = strtok(s, " \t");	/* Get first token (username) */
		if (NULL != t) {
		    strncpy(user_table[user_table_entries].name, 
			    t, USERNAME_MAXLEN);
		    user_table[user_table_entries].name[USERNAME_MAXLEN] = '\0';
		    t = strtok(NULL, " \t"); /* Get next token (secret) */
		    if (NULL != t) {
			strncpy(user_table[user_table_entries].secret, 
				t, SECRET_MAXLEN);
			user_table[user_table_entries].secret[SECRET_MAXLEN] = '\0';
			/* In case a LF has sneaked in */
			chop(user_table[user_table_entries++].secret); 
		    } else {
			fprintf(stderr, "**WARNING: user %s has no secret (line %d)!\n", 
				user_table[user_table_entries].name, line);
			user_table[user_table_entries].secret[0] = '\0';
		    }
		} else {
		    if (verbose_mode)
			fprintf(stderr, "**INFO: cannot get token! (line %d)\n", line);
		}
	    } else {
		/* Assume we've reached EOF. Terminate table. */
		user_table[user_table_entries].name[0] = '\0';
		user_table[user_table_entries].secret[0] = '\0';
		break;
	    }
	}
    } else {
	fprintf(stderr, "**WARNING: Couldn't open user-secrets file \"%s\"\n", 
		user_table_filename);
	fprintf(stderr, "           Reverting to nameless single-secret system.\n");
	/* Get secret from user */
	tfd = fileno(stdin);		/* Prepare to turn off echoing */
	memset(&tios, '0', sizeof(&tios));
	tcgetattr(tfd, &tios);
	old_lflag = tios.c_lflag;
	tios.c_lflag &= ~(ECHO | ECHONL);
	tcsetattr(tfd, TCSANOW, &tios);	/* Turn off echoing */

	fprintf(stdout, "Enter shared secret: "); /* jalla, jalla */
	fflush(stdout);
	fgets(user_table[0].secret, SECRET_MAXLEN, stdin); /* Read secret */
	chop(user_table[0].secret);
	fflush(stdin);
	fprintf(stdout, "Ok\n");
	fflush(stdout);

	user_table_entries = 1;
	user_table[user_table_entries].name[0] = '\0';
	user_table[user_table_entries].name[1] = '\0';
	user_table[user_table_entries].secret[1] = '\0';

	tios.c_lflag = old_lflag;
	tcsetattr(tfd, TCSANOW, &tios);	/* Turn on echoing */
    }


    /* Set alarm-handler (and other signal-handlers) */
    i = (int)signal(SIGALRM, alarm_handler);
    if (-1 == i) {
	perror("signal(SIGALRM, alarm_handler)");
	exit(1);
    }
				
    do_server(pilot_device, door_device, &ctrl_channel, initial_state);
}
/*
vim:ts=8
*/
