/*********************************************************************
 * Filename:      doorknob.c
 * Description:   PalmPilot (PalmOS2.0) application component of doorknob
 * Author:        Per Harald Myrvang <perm@stud.cs.uit.no>
 * 	
 * 	Copyright University of Troms (Norway)
 * 	See the file COPYING for details
 ********************************************************************/

/* Main code for Doorknob application */

/* TODO:
 *	- Consider "encrypting" the username too.
 *	- Make a proper icon!
 *	- Decide when this inane piece of wossname is stable enough to define
 *	  ERROR_CHECK_LEVEL to ERROR_CHECK_NONE (0).
*/

#include <Pilot.h>
#include <SerialMgr.h>

#include <md5.h>		/* pilotSSLeay2.01 preferably */

#include "callback.h"

#include "doorknobRsc.h"

#include "common.h"

#define VERSION "0.9.8"

/* Macro for converting number of micro-seconds to number of system ticks */
#define prvMicroSecondsToTicks(microSeconds)	\
	( ((ULong)((ULong)sysTicksPerSecond * (microSeconds)) / 1000000L) + 1 )

#define IOBUFFERSIZE		60
#define MIN_CHALLENGE_LEN	SDSI_DATE_LEN

#define MAX_TIMEOUT			prvMicroSecondsToTicks(3000000)

/* Yes, these two is supposed to have a space declared infront of REQUEST_*. */
static char request_open[] = " " REQUEST_OPEN "\n";
static char request_lock[] = " " REQUEST_LOCK "\n";
static char *initiate_txt = NULL;	/* Points to one of the two above. */

static VoidHand challenge_handle = NULL;
static VoidHand response_handle = NULL;
static char *challenge_buffer;
static char *response_buffer;

static char *shared_secret_ptr;		/* Points into the preference-struct */
static char *username_ptr;		/* Ditto */

typedef struct {
    UChar	shared_secret[SECRET_MAXLEN];
    Word	silent;
    Char	username[USERNAME_MAXLEN];
} DoorKnobPreferenceType;

static DoorKnobPreferenceType doorKnobPrefs;

/* 
 * Version number of preference-structure. Change when changing
 * DoorKnobPreferenceType. 
 */ 
#define doorKnobPrefVersionNum	2

static const char hs[] = "0123456789abcdef"; /* used by hexify below */

/* Turn a sequence of bytes into a hexadecimal ascii-notation.  
 * Programmer beware: We will always require double the size for 
 * the hex-encoded data!
 * This function will terminate the to-string. Thus the required 
 * size of the to-buffer is (strlen(from)*2)+1 
 * */
static void
hexify(unsigned char *from, char *to, int len)
{
    int i,l;
    for (l = i = 0; l < len; l++) {
	to[i++] = hs[((from[l] >> 4) & 0x0F)];
	to[i++] = hs[(from[l] & 0x0F)];
    }
    to[i] = '\0';
}

#if 0 /* Not needed for now. */
/* The reverse of hexify */
static void
unhexify(unsigned char *from, char *to, int len)
{
	int i,l;
	for (l = i = 0; l < len; l++) {
		/* error-checking is for wimps! */
		to[l] = (((from[i++] - '0') << 4) | (from[i++] - '0'));
	}
}
#endif


/* Play a sound, given frequency in Hertz and the delay in
 *  milliseconds.
 * Meep-meep! 
 */
static void
play_sound(int Hz, int ms)
{
    static SndCommandType snd;

    if (0 == doorKnobPrefs.silent) {
	snd.cmd = sndCmdFreqDurationAmp;
	snd.param1 = (ULong)Hz;		/* frequency in Hz */
	snd.param2 = (UInt)ms;		/* duration in milliseconds */
	snd.param3 = sndMaxAmp;
	SndDoCmd(NULL, &snd, 0);
    }
}

/* Set a text-field to the given text. Is now safe on 
 * editable objects. 
 */
static void
set_text_field(Int id, CharPtr new_text)
{
    FormPtr  frm;
    FieldPtr fld;
    VoidHand hnd;
    CharPtr  p;
	
    frm = FrmGetActiveForm();
    fld = FrmGetObjectPtr(frm,FrmGetObjectIndex(frm, id));
    hnd = MemHandleNew(StrLen(new_text)+1);
    ErrFatalDisplayIf(hnd==NULL, "MemHandleNew");
    p = MemHandleLock(hnd);
    ErrFatalDisplayIf(p==NULL, "MemHandleLock");
    StrCopy(p, new_text);
    MemHandleUnlock(hnd);
	
    FldSetTextHandle(fld, (Handle)hnd);
    FldSetUsable(fld, true);
    FldDrawField(fld);
}

/* Set an object's visibility (i.e. hide/reveal it) */
static void
set_object_visibility(Int id, Boolean usable)
{
    FormPtr  frm;
    Word	 i;
	
    frm = FrmGetActiveForm();
    i = FrmGetObjectIndex(frm, id);
    if (usable)
	FrmShowObject(frm, i);
    else
	FrmHideObject(frm, i);
}


/* Set serial options */
static void
set_serial(UInt serialref)
{
    SerSettingsType serialsettings;
    int err;

    MemSet(&serialsettings, sizeof(serialsettings), 0);
    serialsettings.baudRate = 9600;
    serialsettings.ctsTimeout = prvMicroSecondsToTicks(2000000); /* 2 secs */
    serialsettings.flags = serSettingsFlagStopBits1 |
	serSettingsFlagBitsPerChar8 |
	serSettingsFlagRTSAutoM |
	serSettingsFlagCTSAutoM;
    err = SerSetSettings(serialref, &serialsettings);
    ErrNonFatalDisplayIf(err, "SerSetSettings");
}


/* Send a piece of data over the serial port. */
static void
serial_send(UInt serRef, VoidPtr buf, ULong count, Long timeout)
{
    Err error;

    SerSendFlush(serRef);		/* Flush any old junk. */
    SerClearErr(serRef);
    SerSend(serRef, buf, count, &error);
    if (error)
	ErrThrow(error);
    SerClearErr(serRef);
    error = SerSendWait(serRef, timeout);
    if (error)
	ErrThrow(error);
    SerClearErr(serRef);
}

/* Receive some data over the serial port. */
static void
serial_receive(UInt serRef, VoidPtr buf, ULong count, Long timeout)
{
    Err error;
    ULong numBytes;

    SerClearErr(serRef);
    MemSet(buf, count, 0);
    error = SerReceiveWait(serRef, count, timeout);
    if (error)
	SerClearErr(serRef);
    SerReceiveCheck(serRef, &numBytes);
    SerReceive(serRef, buf, count, 0, &error);
    if (error)
	ErrThrow(error);
    SerClearErr(serRef);
}
/* Main communication-function. Here's all the fun 
 * authentication-protocolish things live.
 * Note that we do all communication in half-duplex mode, 
 * to easily allow for any IrDA-compliant add-on. Sounds great, 
 * doesn't it. Or something. 
 * Return values:
 *	-1				= generated by ErrCatch.
 *	-2				= madness afot
 *	ack_OK			= ok
 *	ack_NOEFFECT	= no effect, i.e. trying to open an open door
 *	ack_FAILED		= failure - i.e. not authenticated
 */

static Int
do_communication(UInt serRef)
{
    unsigned char mdout[16];
    Int ret;
    Int i;

    /* Wondering about this? Read about the Error Manager. */
    ErrTry{			
	/* 1) Initiate contact */
	set_text_field(StatusField, "Initiating");
	/* Build request in the challenge_buffer to save on space 
	 * (we'll clear it after use) */
	StrCopy(challenge_buffer, username_ptr);
	/* A bit magic here - initiate_txt has a space prepended, 
	 * see the global declaration of request_open/lock above. 
	 * It must have a space prepended, so that the server can 
	 * tokenize. Why add the username?
	 * So that each user can have his/hers own secret! 
	 * Why? *SOCK* */
	StrCat(challenge_buffer, initiate_txt);
	serial_send(serRef, challenge_buffer, StrLen(challenge_buffer), 
		    MAX_TIMEOUT);
	set_text_field(StatusField, "Awaiting challenge");

	/* 1.1) clear buffers */
	MemSet(challenge_buffer, IOBUFFERSIZE, 0);
	MemSet(response_buffer, IOBUFFERSIZE, 0);

	/* 2) Wait for challenge */
	serial_receive(serRef, challenge_buffer, MIN_CHALLENGE_LEN, 
		       MAX_TIMEOUT);
	set_text_field(StatusField, "Challenged");
	set_object_visibility(ChallengeLabel, true);
	set_text_field(ChallengeField, challenge_buffer);
#ifdef DEBUG
	SysTaskDelay(SysTicksPerSecond());
#endif

	/* 3) Compute and send response */
#ifdef DEBUG
	challenge_buffer[SDSI_DATE_LEN]='\0';
	set_text_field(ChallengeField, challenge_buffer);
	SysTaskDelay(SysTicksPerSecond());
#endif
	StrCat(challenge_buffer, shared_secret_ptr);

#ifdef DEBUG
	set_text_field(ChallengeField, challenge_buffer);
	SysTaskDelay(SysTicksPerSecond());
#endif
	MD5(challenge_buffer, StrLen(challenge_buffer), mdout);
	/* MD5 hash is 16 bytes (128 bits). */
	hexify(mdout, response_buffer, 16); 
	StrCat(response_buffer, "\n");
	serial_send(serRef, response_buffer, StrLen(response_buffer), 
		    MAX_TIMEOUT);
	set_text_field(StatusField, "Responded");

	/* 4) Wait for acknowledgement */
	serial_receive(serRef, challenge_buffer, ACK_LEN, MAX_TIMEOUT);
	challenge_buffer[ACK_LEN]='\0';
	i = StrCompare(challenge_buffer, ACK_OK);
	if (0 == i )
	    ret = ack_OK;
	else {
	    i = StrCompare(challenge_buffer, ACK_NOEFFECT);
	    if (0 == i )
		ret = ack_NOEFFECT;
	    else
		ret = ack_FAILED;
	}

	/* 5) Finished */
	return ret;
    }

    /* Wondering about this? Read about the Error Manager. */
    ErrCatch(err) {	
	SerClose(serRef);
	switch(err) {
	case serErrTimeOut:
	    FrmCustomAlert(ErrorAlert, "Handshake or Interbyte timeout!", "", "");
	    break;
	case serErrLineErr:
	    FrmCustomAlert(ErrorAlert, "RS232 line error!", "", "");
	    break;
	default:
	    StrPrintF(challenge_buffer, "Eeeek:\nError 0x%lx", (Long)err);
	    FrmCustomAlert(ErrorAlert, challenge_buffer, "", "");
	    break;
	}
	return -1;
    } ErrEndCatch;

    /*NOTREACHED*/
    return -2;	
}

/* Initialize serial communications. The serial port is kept open 
 * for as short a time as possible. When it is open, a check-symbol 
 * appears in the top right corner. The serial port is never open 
 * for longer than the check-symbol is present. However, it might 
 * close before the check-symbol is removed. Ah, the pains one take 
 * to preserve battery-life!
 */
static void
init_serial(void)
{
    Err err;
    UInt serRefNum;
    Int ret = -1;
    int i;

    err = SysLibFind("Serial Library", &serRefNum);
    if (!err) {
	err = SerOpen( serRefNum, 0, 9600L );
	if (!err) {
	    FntSetFont(symbolFont);
	    WinDrawChars("\x16", 1,  148, 2);
	    set_serial(serRefNum);
	    ret = do_communication(serRefNum);
	    SerClose( serRefNum );	/* Close the Serial Manager */
	    FntSetFont(symbolFont);
	    WinEraseChars("\x16", 1,  148, 2);
	    switch(ret) {
	    case ack_OK:
		set_text_field(StatusField, "Authenticated");
		play_sound(880, 40);
		play_sound(1760, 40);
		play_sound(880, 40);
		SysTaskDelay(SysTicksPerSecond());
		break;
	    case ack_NOEFFECT:
		set_text_field(StatusField, "No Effect!");
		for(i = 440; i < 783; i += 23)
		    play_sound(i, 5);
		SysTaskDelay(SysTicksPerSecond());
		break;
	    case ack_FAILED:
		play_sound(220, 100);
		play_sound(150, 200);
	    default:
		/* Is it neccessary to display a whole alert for this? */
		FrmCustomAlert(InfoAlert, "Authentication failed!", "", "");
		break;
	    }
	} else {
	    if (err == serErrAlreadyOpen) {
		SerClose( serRefNum );
	    }
	}
    } else {
	ErrDisplay("No serial library");
    }
}

/* Handle events for the main form, although all other forms 
 * are handled here also, by means of some tricks, in order 
 * to save a few bytes. 
 */
static Boolean
MainFormHandleEvent (EventPtr e)
{
    Boolean	handled;
    FormPtr	frm;
    FieldPtr	fld;
    CharPtr	txt;

    CALLBACK_PROLOGUE;

    handled = false;
    txt = NULL;
    
    if (MainForm == FrmGetActiveFormID()) {
	set_object_visibility(ChallengeLabel, false);
	set_object_visibility(ChallengeField, false);
	set_text_field(StatusField, "Idle");
    }

    switch (e->eType) {
    case frmOpenEvent:
	frm = FrmGetActiveForm();
	FrmDrawForm(frm);
	if (PwdForm == FrmGetActiveFormID()) {
	    frm = FrmGetActiveForm();
	    CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, SndChk)), 
			doorKnobPrefs.silent); 
	    set_text_field(NameField, doorKnobPrefs.username);
	    set_text_field(KeyField, doorKnobPrefs.shared_secret);
	}
	handled = true;
	break;

    case menuEvent:
	MenuEraseStatus(NULL);

	switch(e->data.menu.itemID) {
	case Menu_Help:
	    FrmHelp(MenuHelp);
	    break;
	case Menu_About:
	    FrmCustomAlert(InfoAlert, "Version " VERSION "\nCompiled " 
			   __DATE__ " " __TIME__ "\n" \
			   "Brought to you by the PASTA-project at the University of Troms, Norway.\n" \
			   "Programming by\nPer Harald Myrvang\n<perm@stud.cs.uit.no>", "", "");
	    AbtShowAbout((ULong)'DKNB');
	    break;
	case Menu_Pref:
	    FrmPopupForm(PwdForm);
	    break;
	}

	handled = true;
	break;

    case ctlSelectEvent:
	switch(e->data.ctlSelect.controlID) {
	case openBtn:
	    initiate_txt = request_open;
	    init_serial();
	    break;
	case closeBtn:
	    initiate_txt = request_lock;
	    init_serial();
	    break;
	case okBtn:			/* Only on PwdForm */
	    frm = FrmGetActiveForm();
	    fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, KeyField)); 
	    txt = FldGetTextPtr(fld);
	    if (NULL != txt) {
		StrNCopy(shared_secret_ptr, txt, SECRET_MAXLEN);
#ifdef DEBUG
		FrmCustomAlert(InfoAlert, "Secret updated!", "", "");
#endif
	    }
	    fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, NameField)); 
	    txt = FldGetTextPtr(fld);
	    if (NULL != txt) {
		StrNCopy(username_ptr, txt, USERNAME_MAXLEN);
#ifdef DEBUG
		FrmCustomAlert(InfoAlert, "Username updated!", "", "");
#endif
	    }
	    doorKnobPrefs.silent = CtlGetValue(
		FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, SndChk))); 
	    /* FALLTROUGH to noBtn! */
	case noBtn:			/* Only on PwdForm */
	    FrmReturnToForm(MainForm);
	    break;
	}
	break;

    default:
	break;
    }

    CALLBACK_EPILOGUE;

    return handled;
}

/* Handle events for the application. */
static Boolean
ApplicationHandleEvent(EventPtr e)
{
    FormPtr frm;
    Word    formId;
    Boolean handled = false;

    if (e->eType == frmLoadEvent) {
	formId = e->data.frmLoad.formID;
	frm = FrmInitForm(formId);
	FrmSetActiveForm(frm);

	switch(formId) {
	case PwdForm:
	    /* I use some nasty re-entry tricks here; normally one 
	     * would use a separate HandeEvent-function for each 
	     * form. This is not exactly true, but you'll have 
	     * to juggle your buttons carefully. 
	     * Why do it this way? 
	     * It trades clarity for a smaller size. 
	     */
	    /* FALLTHROUGH */
	case MainForm:
	    FrmSetEventHandler(frm, MainFormHandleEvent);
	    break;
	}
	handled = true;
    }
    return handled;
}

/* The key used to "encrypt" the shared secret, just in case. 
 * Why 0x4e? Why not. 
 */
#define SECRET_KEY 0x4e /**/

/* Get preferences, open (or create) app database */
static Word
StartApplication(void)
{
    Word prefsize;
    int i;

    /* Fetch application preferences */
    i = PrefGetAppPreferences('DKNB', 0, &doorKnobPrefs, &prefsize, true);

    /* See if there were some prefs, if not, force user to 
     * enter secret 
     */
    if (noPreferenceFound == i) {
	FrmCustomAlert(InfoAlert,
		       "No prefs found!\n"\
		       "Name now defaults to \"test\". "\
		       "Secret now defaults to \"secret\". "\
		       "Set your own via the Options menu.",
		       "", "");
	StrCopy(doorKnobPrefs.shared_secret, DEBUG_SECRET);
	StrCopy(doorKnobPrefs.username, "test");
	doorKnobPrefs.silent = 0;
    } else {
	/* "decrypt" the secret before using it. */
	for(i = 0; i < SECRET_MAXLEN; i++)
	    doorKnobPrefs.shared_secret[i] ^= 0x4e; /* Why 0x4e? Why not. */
    }
    shared_secret_ptr = &doorKnobPrefs.shared_secret[0];
    username_ptr = &doorKnobPrefs.username[0];

    /* Allocate and lock temporal challenge/reponse buffers */
    challenge_handle = MemHandleNew(IOBUFFERSIZE + 4);
    ErrFatalDisplayIf(challenge_handle == NULL, "MemHandleNew failed!");
    challenge_buffer = MemHandleLock(challenge_handle);
    ErrFatalDisplayIf(challenge_buffer == NULL, "MemHandleLock failed!");

    response_handle = MemHandleNew(IOBUFFERSIZE + 4);
    ErrFatalDisplayIf(response_handle == NULL, "MemHandleNew failed!");
    response_buffer = MemHandleLock(response_handle);
    ErrFatalDisplayIf(response_buffer == NULL, "MemHandleLock failed!");

    FrmGotoForm(MainForm);
    return 0;
}

/* Save preferences, close forms, close app database */
static void
StopApplication(void)
{
    int i;
    FrmSaveAllForms();
    FrmCloseAllForms();

    /* Free stuff we allocated at startup */
    MemHandleUnlock(challenge_handle);
    MemHandleFree(challenge_handle);
    MemHandleUnlock(response_handle);
    MemHandleFree(response_handle);
    challenge_handle = response_handle = NULL;

    /* "encrypt" the secret before storing it in the preferences. */
    for(i = 0; i < SECRET_MAXLEN; i++)
	doorKnobPrefs.shared_secret[i] ^= 0x4e;	/* Why 0x4e? Why not. */

    /* Set preferences */
    PrefSetAppPreferences('DKNB', 0, doorKnobPrefVersionNum, &doorKnobPrefs,
			  sizeof(DoorKnobPreferenceType), true);

    /* Clear secret */
    MemSet(doorKnobPrefs.shared_secret, 0, SECRET_MAXLEN);
    shared_secret_ptr = NULL;
}

/* The main event loop */
static void
EventLoop(void)
{
    Word err;
    EventType e;

    do {
	EvtGetEvent(&e, evtWaitForever);
	if (! SysHandleEvent (&e))
	    if (! MenuHandleEvent (NULL, &e, &err))
		if (! ApplicationHandleEvent (&e))
		    FrmDispatchEvent (&e);
    } while (e.eType != appStopEvent);
}

/* 
 * Main entry point; it is unlikely you will need to 
 * change this except to handle other launch command codes 
 */
DWord
PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
    Word err;

    if (cmd == sysAppLaunchCmdNormalLaunch) {

	err = StartApplication();
	if (err) return err;

	EventLoop();
	StopApplication();

    } else {
	return sysErrParamErr;
    }

    return 0;
}
