/**
* MakeDocJ
* version 0.7
*
* Compresses text files into a format that is ready to export to a Pilot
* and work with Rick Bram's PilotDOC reader.
*
* Copyright (C) Pat Beirne, 2000.
* Distributable under the GNU General Public License Version 2 or later.
*
*/

import java.io.*;

class tDocHeader {
	String sName;
	int dwUnknown1;
	int dwTimeCreate;
	int dwTimeModded;
	int dwTimeBackup;
	int dwLastSync;
	int ofsSort;
	int ofsCatagories;
	int dwCreator;
	int dwType;
	int dwUnknown2;
	int dwUnknown3;
	short  wNumRecs;

	public void read(DataInput in) throws IOException {
		byte[] b = new byte[32];
		in.readFully(b);
		sName = new String(b,0);
		int i = sName.indexOf('\0');
		if (i>=0)
			sName = sName.substring(0,i);
		dwUnknown1 = in.readInt();
		dwTimeCreate = in.readInt();
		dwTimeModded = in.readInt();
		dwTimeBackup = in.readInt();
		dwLastSync = in.readInt();
		ofsSort = in.readInt();
		ofsCatagories = in.readInt();
		dwCreator = in.readInt();
		dwType = in.readInt();
		dwUnknown2 = in.readInt();
		dwUnknown3 = in.readInt();
		wNumRecs = in.readShort();
	}

	public void write(DataOutput out) throws IOException {
		byte[] b = new byte[32];
		if (sName.length()<32) {
			sName.getBytes(0,sName.length(),b,0);
			b[sName.length()] = 0;
		}
		else {
			sName.getBytes(0,32,b,0);
			b[31] = 0;
		}
		out.write(b);
		out.writeInt(dwUnknown1);
		out.writeInt(dwTimeCreate);
		out.writeInt(dwTimeModded);
		out.writeInt(dwTimeBackup);
		out.writeInt(dwLastSync);
		out.writeInt(ofsSort);
		out.writeInt(ofsCatagories);
		out.writeInt(dwCreator);
		out.writeInt(dwType);
		out.writeInt(dwUnknown2);
		out.writeInt(dwUnknown3);
		out.writeShort(wNumRecs);
	}
}

class tRecIndex {
	public int fp;		// file pointer
	public int id;		// pilot id
	public byte attr;	// attributes

	// bury the read in the constructor
	public tRecIndex(DataInput di) throws IOException {
		fp = di.readInt();
		attr = di.readByte();
		id = ((0xff & di.readByte())<<16) + ((0xff & di.readByte())<<8) + (di.readByte() & 0xff);
	}
	public void write(DataOutputStream dos) throws IOException {
		dos.writeInt(fp);
		dos.write(attr);
		dos.write((byte)(id>>16));
		dos.write((byte)(id>>8));
		dos.write((byte) id);
	}
}

class tDocRecord0 {
	static public short UNCOMPRESSED = 1;
	static public short COMPRESSED = 2;
	public short wVersion;	// 1=plain text, 2=compressed
	public short wSpare;
	public int dwStoryLen;   // in bytes, when decompressed
	public short wNumRecs; 			// text records only; equals tDocHeader.wNumRecs-1
	public short wRecSize;			// usually 0x1000
	public int dwSpare2;

	public void read(DataInput di) throws IOException {
	    wVersion = di.readShort();
	    wSpare = di.readShort();
	    dwStoryLen = di.readInt();
	    wNumRecs = di.readShort();
	    wRecSize = di.readShort();
	    dwSpare2 = di.readInt();
	}

	public void write(DataOutput out) throws IOException {
		out.writeShort(wVersion);
		out.writeShort(wSpare);
		out.writeInt(dwStoryLen);
		out.writeShort(wNumRecs);
		out.writeShort(wRecSize);
		out.writeInt(dwSpare2);
	}
}

class MakeDocJ {

	static void Decomp(String src, String dest, boolean bBinary) throws IOException
	{
		RandomAccessFile fin;
		FileOutputStream fout;
		try {
			fin = new RandomAccessFile(src,"r");
			fout = new FileOutputStream(dest);
		} catch (FileNotFoundException e) {
			System.out.println("the file "+src+" could not be found");
			return;
		} catch (IOException e) {
			System.out.println("the file "+dest+" could not be opened");
			return;
		}

		tDocHeader head = new tDocHeader();
		head.read(fin);

		if (head.dwType != 0x52454164 //'REAd'
			|| head.dwCreator !=  0x54455874) //'TEXt'L)
		{
			System.out.println(".prc file is not the correct format");
			System.exit(3);
		}

		tRecIndex[] ri = new tRecIndex[head.wNumRecs];
		for (int i=0; i<head.wNumRecs; i++) {
			ri[i] = new tRecIndex(fin);
		}

		int dwPos;

		tDocRecord0 t0 = new tDocRecord0();

		fin.seek(ri[0].fp);
		t0.read(fin);

		if (t0.wVersion!=tDocRecord0.UNCOMPRESSED && t0.wVersion!=tDocRecord0.COMPRESSED)
			System.out.println("\nWARNING: unknown file compression type: "+t0.wVersion);

		boolean bCompressed = (t0.wVersion==tDocRecord0.COMPRESSED);

		short nRecs;
		nRecs = head.wNumRecs;

		// this is the main record buffer
		// it knows how to stretch to accomodate the decompress
		tBuf t;

		int dwRecLen;
		for (int i=1; i<nRecs; i++)
		{
			// for the last, use the file len
			if (i==nRecs-1)
				dwRecLen = (int) fin.length() - ri[i].fp;
			else
				dwRecLen = ri[i+1].fp - ri[i].fp;

			fin.seek(ri[i].fp);

			t = new tBuf(dwRecLen);
			fin.read(t.buf);
			if(bCompressed)
				t.Decompress();

			// check for CR insert
			if (!bBinary)
				t.DuplicateCR();
			System.out.println("\rreconverting "+head.sName+": record "+i+" of "+nRecs);

			fout.write(t.buf);
		}

	}


	public static void
	main(String args[])
	{
		System.out.println("MakeDocJ ver 1");
		if (args.length<3)
		{
			System.out.println("syntax makedoc [-n] [-b] <text-file> <prc-file> <story-name>");
			System.out.println("                 convert text files to .PRC format");
			System.out.println("       makedoc -d [-b] <prc-file> <text-file>");
			System.out.println("                 decodes the PRC back into the txt file");
			System.out.println("       -n builds the .prc file without compression");
			System.out.println("       -b option compresses/decompresses binary");
			System.exit(1);
		}

		int iArg = 0;
		boolean bDecomp = false;
		boolean bBinary = false;
		boolean bReport = false;
		boolean bCompress = true;

		while (args[iArg].charAt(0)=='-' || args[iArg].charAt(0)=='\\')
		{
			if (args[iArg].charAt(1)=='d')
				bDecomp = true;
			if (args[iArg].charAt(1)=='b')
				bBinary = true;
			if (args[iArg].charAt(1)=='r')
				bReport = true;
			if (args[iArg].charAt(1)=='n')
				bCompress = false;
			iArg++;
		}

        try {
    		if (bDecomp)
    			Decomp(args[iArg], args[iArg+1], bBinary);
			else
				Compress(args[iArg], args[iArg+1], args[iArg+2], bCompress, bBinary, bReport);
    	} catch (Exception e) {
    	    System.out.println(e.toString());
    	}
	}

	public static void Compress(String src, String dest, String title, boolean bCompress, boolean bBinary, boolean bReport) {

		int storySize = 0;
		FileInputStream fin;
		RandomAccessFile fout;
		try {
			fin = new FileInputStream(src);
			storySize = fin.available();
			(new File(dest)).delete();
			fout = new RandomAccessFile(dest,"rw");
		} catch (FileNotFoundException e) {
			System.out.println("the file "+src+" could not be found");
			return;
		} catch (IOException e) {
			System.out.println("the file "+dest+" could not be opened");
			return;
		}

		tDocHeader head1 = new tDocHeader();

		int	x;
		short w;
		int	recSize = 4096;
		short		z,numRecs;

		head1.sName = title;
		System.out.println("saving to "+dest+" as <"+title+">,"+
				(bBinary ? " binary mode," : "")  +
				(bCompress ? "" : " not") +
				" compressed");

		head1.dwUnknown1 = 0;
		head1.dwTimeCreate = (int) (System.currentTimeMillis() / 1000);
		head1.dwTimeModded = (int) (System.currentTimeMillis() / 1000);
		head1.dwTimeBackup = 0;
		head1.dwLastSync = 0;
		head1.ofsSort = 0;
		head1.ofsCatagories = 0;
		head1.dwCreator = 0x54455874; //'TEXt';				// database type
		head1.dwType = 0x52454164;    //'REAd';	 			// database creator
		head1.dwUnknown2 = 0;
		head1.dwUnknown3 = 0;


		try {
			z = (short) (storySize/recSize);
			if (z * recSize < storySize)
				z ++;

			numRecs = z;
			z ++;

			head1.wNumRecs = z;		//  the number of records to follow
			head1.write(fout);


			int index;
			index = 0x406f8000;		// the pattern for attributes=dirty + unique_id=0x6f8000
			x = 0x50 + (int) z * 8;

			fout.writeInt(x);		// start writing the record offsets
			fout.writeInt(index);

			x += 0x0010;		// length of record 0

			while(--z > 0) {
				fout.writeInt(x);			//more record offsets
				fout.writeInt(++index);	// the attributes + ID's
				x += 0x1000L;				// typical record length
			}
			// one more short.....don't know why
			fout.writeShort(0);

			tDocRecord0 rec0 = new tDocRecord0();
			rec0.wVersion = (bCompress ? tDocRecord0.COMPRESSED : tDocRecord0.UNCOMPRESSED);
			rec0.wSpare = 0;
			rec0.dwStoryLen = storySize;
			rec0.wNumRecs = (short) (head1.wNumRecs - 1);
			rec0.wRecSize = (short) recSize;
			rec0.dwSpare2 = 0;

			rec0.write(fout);

			int n = recSize;
			// dump the whole story into the new file
			int recNum = 0;
			System.out.println();

			tBuf buf = new tBuf(4096);

			while(recNum < numRecs)
				{
				int pos;
				pos = (int) fout.getFilePointer();
				fout.seek(0x56 + 8*recNum);
				fout.writeInt(pos);
				fout.seek(pos);

				int nOrg;

				byte[] bl = new byte[4096];
				n = nOrg = fin.read(bl);
				buf.assign(bl,n);

//System.out.println("file read returned " + buf.buf.length);
				if (nOrg==-1) break;

				if (!bBinary)
					n = buf.RemoveBinary();
				if (bCompress)
					n = buf.Compress();
				fout.write(buf.buf);
//System.out.println("file wrote used " + buf.buf.length);

				System.out.println("converting record "+(recNum+1)+" of "+numRecs);
				if (bReport && n>0 && bCompress)
					System.out.println("\noriginal "+nOrg+" bytes, compressed to "+n+" bytes, ratio: "+
						(100. * n / nOrg));
				recNum++;
				}
		} catch (Exception e) {
			System.out.println(e);
			return;
		}

	}
}

/**
* This app reads an arbitrary file and converts it to a PilotDoc
* .pdb file. It can also decode a .pdb back into text.
*
* It also can compress the original source text into a proprietary
* compression scheme.
*/



/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
/////////////////////                                  //////////////////////
/////////////////////      tBuf class                  //////////////////////
/////////////////////                                  //////////////////////
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/**
* This little class handles in-memory 4096 byte buffers. It can
* compress, decompress, remove binary codes, duplicate \n codes
*
*/
class tBuf {
	static final int DISP_BITS = 11;
	static final int COUNT_BITS = 3;
	static final int DEF_LEN = 6000;

	// the main storage buf
	// during translations, this is the source buffer
	byte[] buf;
	// during translations, this it the destination buffer
	private byte[] dest;
	private int iDest;
	private boolean bSpace;

	tBuf() { this(DEF_LEN); }
	tBuf(int len) {
		buf = new byte[len];};

	public int length()
		{return buf.length;}

	/**
	* Remove the binary codes (0 and 0x0D) from the buffer
	* this nasty little beast removes really low ASCII and 0's
	* and handles the CR problem
	*
	* if a cr appears before a lf, then remove the cr
	* if a cr appears in isolation, change to a lf
	* @return the resulting buffer length
	*/
	public int RemoveBinary() {
		dest = new byte[(buf.length*3)/2];
		int k,j;
		for (j=k=0; j<buf.length; j++,k++)
		{
			// copy each byte
			dest[k] = buf[j];

			// throw away really low ASCII
			if ((dest[k]>=0 && dest[k]<9)) k--;

			// for CR
			if (dest[k]==0x0D)
			{
				// if next is LF, then drop it
				if (j<buf.length-1 && buf[j+1]==0x0A)
					k--;
				else // turn it into a LF
					dest[k] = 0x0A;
			}
		}
		assign(dest,k);
//System.out.println("removecr returns "+k+ " " + buf.length);
		return k;
	}

	/**
	* Decompress the buffer
	*action: make a new buffer
	*				run through the source data
	*				check the 4 cases:
	*					0,9...7F represent self
	*					1...8		escape n chars
	*					80...bf reference earlier run
	*					c0...ff	space+ASCII
	*
	* @return the resulting buffer length
	*/
	public int Decompress() {
		// we "know" that all decompresses fit within 4096, right?
		dest = new byte[(buf.length*10)];
		iDest = 0;

		int i,j;
		for (j=i=0; j<buf.length; )
		{
			int c;

			// take a char from the input buffer
			c = buf[j++] & 0xFF;

			// separate the char into zones: 0, 1...8, 9...0x7F, 0x80...0xBF, 0xC0...0xFF

			// codes 1...8 mean copy that many bytes; for accented chars & binary
			if (c>0 && c<9)
				while(c-- > 0) dest[iDest++] = buf[j++];

			// codes 0, 9...0x7F represent themselves
			else if (c<0x80)
				dest[iDest++] = (byte) c;

			// codes 0xC0...0xFF represent "space + ascii char"
			else if (c>=0xC0) {
					dest[iDest++] = ' ';
					dest[iDest++] = (byte) (c ^ 0x80);
			}

			// codes 0x80...0xBf represent sequences
			else
			{
				int m,n;
				c <<= 8;
				c += buf[j++] & 0xFF;
				m = (c & 0x3FFF) >> COUNT_BITS;
				n = c & ((1<<COUNT_BITS) - 1);
				n += 3;
				while (n-- > 0)
				{
					dest[iDest] = dest[iDest-m];
					iDest++;
				}
			}
		}
		assign(dest, iDest);
		return iDest;
	}

	/**
	* Compress the buffer
	* @return the resulting buffer length
	*/
	public int Compress() {	int i,j;
		dest = new byte[(buf.length* 3)/2];
//System.out.println("got to compress");

		// run through the input buffer
		int pHit;			// points to a walking test hit; works upwards on successive matches
		int pPrevHit;	 	// previous value of pHit
		int pTestHead;		// current test string
		int pTestTail;		// current walking pointer; one past the current test buffer
		int pEnd;			// 1 past the end of the input buffer

		pHit = pPrevHit = pTestHead = 0;
		pTestTail = pTestHead+1;
//		pTestTail = 0;
		pEnd = buf.length;

		// make a dest buffer and reassign the local buffer
		dest = new byte[6000];
		iDest = 0;		// used to walk through the output buffer

		// make a source string, so we can search it quickly
		String src = new String(buf,0);
		String testStr = "";

		// loop, absorbing one more char from the input buffer on each pass
		for (pHit = 0; pTestHead != pEnd; pTestTail++) {

			// scan in the previous data for a match
			// pHit = memfind(pPrevHit, pTestTail - pPrevHit, pTestHead, pTestTail - pTestHead);
			testStr = src.substring(pTestHead, pTestTail);
			pHit = src.lastIndexOf(testStr, pTestHead-1);
//System.out.println("pointers pTestHead "+pTestHead+" pTestTail "+pTestTail+" testString=<"+testStr+"> hit="+pHit);

			// on a mismatch or end of buffer, issued codes
			if (pHit==-1
				|| pTestHead-pHit>((1<<DISP_BITS)-1)
				|| pHit==pTestHead
				|| pTestTail-pTestHead>(1<<COUNT_BITS)+2
				|| pTestTail==pEnd)
			{
				// issued the codes
				// first, check for short runs
				if (pTestTail-pTestHead < 4) {
//System.out.println("issue a char "+buf[pTestHead]);
					Issue(buf[pTestHead]);
					pTestHead++;
				}
				// for inter runs, issue a run-code
				else {
					// issue space char if required
					if (bSpace)  { dest[iDest++] = ' '; bSpace = false; }

					int dist = pTestHead - pPrevHit;
					int compound = (dist << COUNT_BITS) + pTestTail-pTestHead - 4;

//if (dist>=(1<<DISP_BITS)) printf("\n!! error dist overflow");
//if (pTestTail-pTestHead-4>7) printf("\n!! error dist overflow");

					dest[iDest++] = (byte) (0x80 + (compound>>8));
					dest[iDest++] = (byte) (compound & 0xFF);
//if (dist>1500)
//System.out.println("issue a code for sequence len "+(pTestTail-pTestHead-1)+" looking back "+Integer.toString(dist,16)+" bytes (at "+Integer.toString(iDest,16)+")");
//printf("\nissuing code for sequence len %d <%c%c%c>",pTestTail-pTestHead-1,pTestHead[0],pTestHead[1],pTestHead[2]);
//printf("\n          <%x%x>",pOut[-2],pOut[-1]);
					// and start again
					pTestHead = pTestTail-1;
				}
				// start the search again
				pPrevHit = pTestHead;
			}
			// got a match
			else {
				pPrevHit = pHit;
			}
//printf("pointers %x %x %x",pTestHead, pTestTail, pPrevHit);
		// when we get to the end of the buffer, don't inc past the end
		// this forces the residue chars out one at a time
			if (pTestTail==pEnd) pTestTail--;
		}

		// clean up any dangling spaces
		if (bSpace) dest[iDest++] = ' ';


		// final scan to merge consecutive high chars together
		int k;
		byte[] cbuf = new byte[dest.length];
		for (i=k=0; i<iDest; i++,k++) {
			cbuf[k] = dest[i];
			// skip the run-length codes
			if (cbuf[k]>=0x80 && cbuf[k]<0xC0)
				cbuf[++k] = dest[++i];
			// if we hit a high char marker, look ahead for another
			else if (cbuf[k]==1)
			{
				cbuf[k+1] = dest[i+1];
				while (i+2<iDest && dest[i+2]==1 && cbuf[k]<8)
				{
					cbuf[k]++;
					cbuf[k+buf[k]] = dest[i+3];
					i+=2;
				}
				k += cbuf[k]; i++;
			}
		}

		// free up resources
		dest = null; iDest = 0;

		assign(cbuf,k);
		return k;
	}

	/**
	* Handle the details of writing a single
	* character to the compressed stream
	* @return the resulting buffer length
	*/
	private int Issue(byte src) {
		// if there is an outstanding space char, see if
		// we can squeeze it in with an ASCII char
		if (bSpace) {
			if (src>=0x40 && src<=0x7F)
				dest[iDest++] = (byte) (src ^ 0x80);
			else {
				// couldn't squeeze it in, so issue the space char by itself
				// most chars go out simple, except the range 1...8,0x80...0xFF
				dest[iDest++] = ' ';
				if (src>8)
					dest[iDest++] = src;
				else {
					dest[iDest++] = 1; dest[iDest++] = src; }
			}
			// knock down the space flag
			bSpace = false;
		}
		else {
			// check for a space char
			if (src==' ')
				bSpace = true;
			else {
				if (src>8)
					dest[iDest++] = src;
				else {
					dest[iDest++] = 1; dest[iDest++] = src; }
			}
		}
		return iDest;
	}

	/**
	* Change all 0x0A into 0x0D,0x0A pairs
	* @return the resulting buffer length
	*/
	public int DuplicateCR() {
		dest = new byte[(buf.length*3)/2];
		iDest = 0;
		int k,j;
		for (j=k=0; j<buf.length; j++, k++)
		{
			dest[k] = buf[j];
			if (dest[k]==0x0A) { dest[k++] = 0x0D; dest[k] = 0x0A; }
		}
		assign(dest, k);
		return k;
	}
	public void assign(byte[] b, int n) {
		buf = new byte[n];
		System.arraycopy(b, 0, buf, 0, n);
	}
}





