/**
 * Title:        PBM<p>
 * Description:  some methods for working with PBM images<p>
 *
 * Copyright:    Copyright (c) 2000 Andreas Heilwagen<p>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * @author       Andreas Heilwagen
 * @version      1.0
 */


import java.io.*;
import java.util.*;


/**
 * PBM is abled to do some basic operations on PBM images like
 * reading an image from a file and getting pixel information.
 */

public class PBM
{
  static private String MAGIC     = "P4";
  static private int    MAXWIDTH  = 160;   // has to be a multiple of 8
  static private int    MAXHEIGHT = 160;
  static private byte   BITS[]    = { 1, 2, 4, 8, 16, 32, 64, (byte)128 };

  private        int    mWidth    = 0;
  private        int    mHeight   = 0;
  private        int    mLineLen  = 0;
  private        int    mOffset   = 0;
  private        byte   mPixels[];


  /**
   * reads a b/w dithered PBM image
   *
   * @param         aFilename        name of a PBM file
   * @throws        IOException      if anything goes wrong
   */

  public PBM (String aFilename)
  throws IOException
  {
    DataInputStream stream = new DataInputStream (new FileInputStream   (aFilename));
    BufferedReader  reader;
    String          buf;
    StringTokenizer tokens;
    int             lines = 0;

    // read complete file

    mPixels = new byte[(int)new File (aFilename).length ()];
    stream.readFully (mPixels);
    stream.close     ();
    reader = new BufferedReader (new InputStreamReader (new ByteArrayInputStream (mPixels)));

    // read magic

    buf = reader.readLine ();
    lines++;
    if ((buf == null) || !buf.equals (MAGIC))
      throw new IOException ("file is not a b/w PBM image");

    // skip comments
    for (;;)
    {
      buf = reader.readLine ();
      lines++;
      if (!buf.startsWith ("#"))
        break;
    }

    // get size
    if (buf == null)
      throw new IOException ("preliminary end of file");

    tokens = new StringTokenizer (buf);
    if (tokens.countTokens () != 2)
      throw new IOException ("could not parse image size");

    mWidth   = new Integer (tokens.nextToken ()).intValue ();
    mHeight  = new Integer (tokens.nextToken ()).intValue ();
    mLineLen = (mWidth / 8) + (((mWidth % 8) > 0) ? 1 : 0);
    if ((mWidth > MAXWIDTH) || (mHeight > MAXHEIGHT))
      throw new IOException ("image size too big");

    // locate pixel data

    for (mOffset = 0; mOffset < mPixels.length; mOffset++)
      if ((mPixels[mOffset] == 10) || (mPixels[mOffset] == 13))
      {
        lines--;
        if ((mPixels[mOffset + 1] == 10) || (mPixels[mOffset + 1] == 13))
          mOffset++;
        if (lines == 0)
          break;
      }
  }


  /**
   * returns true if the pixel at the given position is set
   *
   * @param           x            horizontal offset
   * @param           y            vertical offset
   *
   * @return          true if the pixel is set
   */

  public boolean isSet (int x, int y)
  {
    return (mPixels[mOffset + 1 + y * mLineLen + (x / 8)] & BITS[x % 8]) != 0;
  }


  /**
   * get width of image
   *
   * @return    width of the image
   */

  public int getWidth ()
  {
    return mWidth;
  }


  /**
   * get height of image
   *
   * @return    height of the image
   */

  public int getHeight ()
  {
    return mHeight;
  }
}
