/**
 * Title:        PBM2Palm<p>
 * Description:  Converts a PBM image to Palm bitmaps for the KVM<p>
 * Copyright:    Copyright (c) 2000 Andreas Heilwagen<p>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * @author       Andreas Heilwagen
 * @version      1.0
 */


import java.io.*;


/**
 * PBM2Palm converts a raw PBM image (b/w dithered) to Palm bitmaps
 * suitable to be displayed by KJava. The output is a Java code
 * fragment.
 */

public class PBM2Palm
{
  static public void main (String[] args)
  {
    PBM image = null;

    // check for commandline args

    if (args.length != 2)
    {
      System.out.println ("PBM2Palm <pbm file> <classname> > <java source file>");
      System.exit        (-1);
    }

    // load pbm file

    try {
      image = new PBM (args[0]);
    } catch (IOException e) {
      System.out.println ("could not read " + args[0] + ": " + e.getMessage ());
      System.exit (-1);
    }

    // generate Java code for KJava

    genCode (image, args[1]);

    System.exit (0);
  }


  /**
   * generate Java code for KJava for the given PBM image
   *
   * @param     aPBM          an instance of PBM
   * @param     aClassname    string name of the generated class
  */

  static private void genCode (PBM aPBM, String aClassname)
  {
    StringBuffer buf   = new StringBuffer ("");
    int          bytes = (aPBM.getWidth () / 8)
                       + (((aPBM.getWidth () % 8) > 0) ? 1 : 0);
    int          bmaps = (bytes / 4) + (((bytes % 4) > 0) ? 1 : 0);
    int          width;
    int          baseX;

    System.out.println ("import com.sun.kjava.Bitmap;");
    System.out.println ("import com.sun.kjava.Graphics;\n\n");

    System.out.println ("public class " + aClassname + "\n{");
    System.out.println ("  final static public int width   = " + aPBM.getWidth ()  + ";");
    System.out.println ("  final static public int height  = " + aPBM.getHeight () + ";");
    System.out.println ("  final static public int bitmaps = " + bmaps             + ";\n\n");

    System.out.println ("  final static public void show (Graphics aContext, int x, int y)");
    System.out.println ("  {");
    for (int i = 0; i < bmaps; i++)
      System.out.println ("    aContext.drawBitmap (x + " + i * 32 + ", y, bitmap" + i + ");");
    System.out.println ("  }\n\n");

    for (int bmap = 0; bmap < bmaps; bmap++)
    {
      width = (bmap < bmaps - 1) ? 4 : (bytes - bmap * 4);

      buf.append ("  final static public Bitmap bitmap");
      buf.append (bmap);
      buf.append (" = new Bitmap ((short)");
      buf.append (width);
      buf.append (", new byte[] {\n    ");

      for (int y = 0; y < aPBM.getHeight (); y++)
      {
        for (int x = 0; x < width; x++)
        {
          baseX = bmap * 32 + x * 8;

          buf.append ("(byte)");
          buf.append ((aPBM.isSet (baseX    , y) ?   1 : 0)
                    + (aPBM.isSet (baseX + 1, y) ?   2 : 0)
                    + (aPBM.isSet (baseX + 2, y) ?   4 : 0)
                    + (aPBM.isSet (baseX + 3, y) ?   8 : 0)
                    + (aPBM.isSet (baseX + 4, y) ?  16 : 0)
                    + (aPBM.isSet (baseX + 5, y) ?  32 : 0)
                    + (aPBM.isSet (baseX + 6, y) ?  64 : 0)
                    + (aPBM.isSet (baseX + 7, y) ? 128 : 0));
          buf.append (", ");
        }
        buf.append ("\n    ");
      }
      buf.setLength (buf.length () - 7);
      buf.append    (" });\n\n");

      System.out.print (buf.toString ());
      buf.setLength (0);
    }
    System.out.println ("}");
  }
}