/*
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License. 
 * 
 * The Original Code is pilot-mailsync.
 * 
 * The Initial Developer of the Original Code is Hollis Blanchard
 * <hollis@andrew.cmu.edu>. Portions created by Hollis Blanchard are Copyright
 * (C) 2000 Hollis Blanchard. All Rights Reserved.
 * 
 * Contributor(s):
 * Frederic Devernay <Frederic.Devernay@sophia.inria.fr>
 * 
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "imap/c-client/c-client.h"

#include "cclientcalls.h"
#include "options.h"
#include "pilot-mailsync.h"
#include "llist.h"
#include "util.h"


#ifdef DEBUG
static void printMail(const struct Mail *mail);
#endif
static char *getHeader(MAILSTREAM *stream, unsigned long msg_no, const char *headerstr);
void markDeleted(MAILSTREAM *stream, unsigned long number);
void clearDeleted(MAILSTREAM *stream);
void getMsg(MAILSTREAM *stream, unsigned long number);

/* everything that we got back from a search */
listItem *gSearchResults;


/* MailToPalm()
 *
 * transfer mail from desktop to Palm
 */
void MailToPalm(char *mailbox)
{
    MAILSTREAM *mstream = NIL;
    SEARCHPGM search;
	long options = 0;
	int msgno;

/********* include a .c file... **********/
#include "imap/c-client/linkage.c"
/*****************************************/

#ifdef DEBUG
	options = OP_DEBUG;
#endif

	if (NULL == (mstream = mail_open(mstream, mailbox, options))) {
		mm_log("fatal error: couldn't open mailbox", ERROR);
		return;
	}

	memset(&search, 0, sizeof(SEARCHPGM));

	/* XXX yeah, I need to make the other sync types work too */
	if (gPrefs.syncType == mailSyncUnread ||
		gPrefs.syncType == mailSyncFilter ||
		gPrefs.syncType == mailSyncAll) {
		/* search for new messages */
		search.unseen = 1;
	}

#if 0
	/* XXX unfinished */
	} else if (gPrefs.syncType == FILTER) {
		/* set up some filters */

		if (gPrefs.getHigh) {
			/* this is regardless of include/exclude */
			search.flagged = 1;
		}

		if (gPrefs.getContaining) {
			/* "include" search */
			makeSearchPgm(&search, &tostr, &fromstr, &subjstr);
		} else {
			/* "exclude" search */
			memset(&not, 0, sizeof(SEARCHPGMLIST));
			memset(&notsearch, 0, sizeof(SEARCHPGM));
			search.not = &not;
			not.pgm = &notsearch;

			makeSearchPgm(&notsearch, &tostr, &fromstr, &subjstr);
		}
#endif

	mail_search_full(mstream, "US-ASCII", &search, NIL); /* XXX "US-ASCII" */
	while ((msgno = (int)pop(&gSearchResults)))
		getMsg(mstream, msgno);
	printf("\n"); /* terminate "...." from getMsg */

	mail_close_full(mstream, CL_EXPUNGE); /* expunge deleted mails */
}

/* getMsg
 *
 * do the actual message processing:
 * - get from c-client (delete if appropriate)
 * - construct new Palm Mail item
 * - transfer to Palm
 */
void getMsg(MAILSTREAM *stream, unsigned long number)
{
	BODY *body;
	MESSAGECACHE *elt;

#ifdef DEBUG
	/* transfer only 1 message */
	static int guard=0;
	if (guard++)
		return;
#endif

	mail_fetchstructure(stream, number, &body);

	if (body) {
		unsigned char record[8192]; /* XXX max size of Mail record? */
		struct Mail pilotMail;
		char *datestr;
		unsigned long len=0;

		/* mark mail to be deleted */
		if (gDelete)
			markDeleted(stream, number);

		/* I don't know what the following do for Palm-bound mail */
		pilotMail.signature = 0;
		pilotMail.confirmRead = 0;
		pilotMail.confirmDelivery = 0;
		pilotMail.addressing = 0;

		elt = mail_elt(stream, number); /* for msg flags */

		/* in the Palm OS Mail struct, 0 is high priority (presumably 2 is low,
		 * but low doesn't seem to be visibly marked in the Mail app)
		 */
		pilotMail.priority = !(elt->flagged);
		pilotMail.read = elt->seen;

		/* convert message's text-date to struct tm */
		datestr = strip_cr(getHeader(stream, number, "Date"));
		pilotMail.date = *DateToTm(datestr);
		pilotMail.dated = 1;

		/* fill in all the text headers */
		pilotMail.subject = strip_cr(getHeader(stream, number, "Subject"));
		pilotMail.from = strip_cr(getHeader(stream, number, "From"));
		pilotMail.to = strip_cr(getHeader(stream, number, "To"));
		pilotMail.cc = strip_cr(getHeader(stream, number, "Cc"));
		pilotMail.bcc = strip_cr(getHeader(stream, number, "Bcc"));
		pilotMail.replyTo = strip_cr(getHeader(stream, number, "Reply-To"));
		pilotMail.sentTo = strip_cr(getHeader(stream, number, "Sent-To"));

		/* get body, truncating if necessary. Note that fetchtext_full
		 * should return the entire body without seperating parts
		 */
		pilotMail.body = strip_cr(mail_fetchtext_full(stream, number,
														&len, NIL));
		if (len > gPrefs.truncate) {
			/* write over the last few chars of the mail */
			char *trunc_warning = "\n\n[pilot-mailsync truncated]";
			strcpy(pilotMail.body+gPrefs.truncate-strlen(trunc_warning),
					trunc_warning);
			printf("/");
			fflush(stdout);
			if (gVerbose)
				printf("\n  truncating mail #%li; new len is %i\n", number,
						strlen(pilotMail.body));
#ifdef DEBUG
			printMail(&pilotMail);
#endif
		}

		len = pack_Mail(&pilotMail, record, sizeof(record));

		/* write the mail record to the Palm now */
		if (0 < dlp_WriteRecord(gSocket, gDB, 0, 0, 0, record, len, 0)) {
			/* yay, wrote mail to Palm */
			nToPalm++;
			printf("."); /* some feedback - one '.' per delivered mail */
			fflush(stdout);
		} else {
			mm_log("  failure writing to Palm\n", ERROR);
		}
	}
}


/* markDeleted
 *
 * if we're deleting mail, mark it deleted here.
 */
void markDeleted(MAILSTREAM *stream, unsigned long number)
{
	char seq[5]; /* won't handle >9,999 msg mailboxes (including ',') */

	sprintf(seq, "%li", number);

#ifdef DEBUG
	printf("  marking mail #%s for deletion\n", seq);
#endif

	mail_setflag(stream, seq, "\\DELETED");
}

/* mm_searched: called on every message that matches the conditions set in
 *              mail_search_full (above).
 * 
 */
void mm_searched (MAILSTREAM *stream, unsigned long number)
{
	push(&gSearchResults, (void *)number);
}

void mm_login(NETMBX *mb, char *user, char *pwd, long trial)
{
	char *nl=NULL;

#ifdef DEBUG
	printf("acquiring login info\n");
#endif
	if (trial || *gUser == '\0') {
		printf("login: ");
		fgets(gUser, NETMAXUSER, stdin);
		/* get rid of the trailing \n */
		if ((nl=strchr(gUser, '\n')))
			*nl = '\0';
	}

	if (trial || *gPwd == '\0')
		strncpy(gPwd, getpass("Password: "), MAILTMPLEN);

	strncpy(user, gUser, NETMAXUSER);
	strncpy(pwd, gPwd, MAILTMPLEN);
}

void mm_log (char *string,long errflg)
{
	switch ((short) errflg) {
		case NIL:
			if (gVerbose)
				printf ("-- %s\n",string);
			break;
		case PARSE:
		case WARN:
			printf ("== %s\n",string);
			break;
		case ERROR:
			printf ("** %s\n",string);
			break;
	}
}

#ifdef DEBUG
void printMail(const struct Mail *mail)
{
	if (mail->subject) printf("  subject: %s", mail->subject);
	if (mail->from) printf("  from: %s", mail->from);
	if (mail->to) printf("  to: %s", mail->to);
	if (mail->cc) printf("  cc: %s", mail->cc);
	if (mail->bcc) printf("  bcc: %s", mail->bcc);
	if (mail->replyTo) printf("  replyTo: %s", mail->replyTo);
	if (mail->sentTo) printf("  sentTo: %s", mail->sentTo);
}
#endif

char *getHeader(MAILSTREAM *stream, unsigned long msg_no, const char *headerstr)
{
	STRINGLIST *header = mail_newstringlist();
	char *hdr=NULL;

	header->text.data = (unsigned char *)strdup(headerstr);
	header->text.size = strlen((char *)header->text.data);
	/* skip 'headerstr: ' */
	hdr = strchr(mail_fetchheader_full(stream,msg_no,header,NIL,NIL), ' ');
	if (hdr) {
		hdr++; /* skip leading space */
		hdr[strlen(hdr)-2] = '\0'; /* hdr ends with \r\n\r\n */
	}

	free(header->text.data);
	return hdr;
}


/***** nothing to see below this point *****/

void mm_notify (MAILSTREAM *stream,char *string,long errflg)
{
	mm_log(string, errflg);
}

long mm_diskerror (MAILSTREAM *stream,long errcode,long serious)
{
	abort();
	return NIL;
}

void mm_fatal (char *string)
{
	mm_log(string, ERROR);
}

void mm_dlog (char *string)
{
	puts(string);
}

void mm_critical (MAILSTREAM *stream)
{
}

void mm_nocritical (MAILSTREAM *stream)
{
}

void smtptest (long debug)
{
}

void mm_exists (MAILSTREAM *stream,unsigned long number)
{
}

void mm_expunged (MAILSTREAM *stream,unsigned long number)
{
}

void mm_flags (MAILSTREAM *stream,unsigned long number)
{
}

void mm_list (MAILSTREAM *stream,int delimiter,char *mailbox,long attributes)
{
}

void mm_lsub (MAILSTREAM *stream,int delimiter,char *mailbox,long attributes)
{
}

void mm_status (MAILSTREAM *stream,char *mailbox,MAILSTATUS *status)
{
}

