/*
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License. 
 * 
 * The Original Code is pilot-mailsync.
 * 
 * The Initial Developer of the Original Code is Hollis Blanchard
 * <hollis@andrew.cmu.edu>. Portions created by Hollis Blanchard are Copyright
 * (C) 2000 Hollis Blanchard. All Rights Reserved.
 * 
 * Contributor(s):
 * Frederic Devernay <Frederic.Devernay@sophia.inria.fr>
 * 
 */


#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

#include "imap/c-client/c-client.h"

#include "options.h"
#include "pilot-mailsync.h"
#include "util.h"

#define FILEMAX 128

char gConfigFile[MAILTMPLEN];
char gDevice[MAILTMPLEN];
char gSendmail[MAILTMPLEN];
char gMailbox[MAILTMPLEN];
char gPwd[MAILTMPLEN];
char gFrom[MAILTMPLEN];
char gUser[NETMAXUSER];
int gDelete;
int gVerbose;

/* private */
static void load_defaults(void);
static void load_configfile(void);
static void load_env(void);
static void load_cmdline(int argc, char *argv[]);
static int check_opts(void);
static void Help(char *prog_name, int longhelp);



void load_options(int argc, char *argv[])
{
	load_defaults();
	load_env();
	load_configfile();
	load_cmdline(argc, argv);
	if (check_opts()) {
		fprintf(stderr, "fatal config error; exiting\n");
		exit(1);
	}
}

void load_defaults(void)
{
	strcpy(gConfigFile, ".pilotmailrc");
	strcpy(gDevice, "/dev/pilot");
	strcpy(gSendmail, "/usr/sbin/sendmail -t -i");
	strcpy(gMailbox, "Mailbox");
	strcpy(gUser, "");
	strcpy(gPwd, "");

	/* set default From address to $USER@$HOSTNAME */
	strncpy(gFrom, getname(getuid()),
			MAILTMPLEN-1-strlen(getenv("HOSTNAME")));
	strcat(gFrom, "@");
	strcat(gFrom, getenv("HOSTNAME"));

	gDelete = 0;
	gVerbose = 0;
}

void load_env(void)
{
	char *envstr;

#ifdef DEBUG
	printf("loading environment variables\n");
#endif

	if ((envstr = getenv("PILOTPORT")))
		strncpy(gDevice, envstr, MAILTMPLEN);

	if ((envstr = getenv("MAIL")))
		strncpy(gMailbox, envstr, MAILTMPLEN);

	if ((envstr = getenv("PILOTFROM")))
		strncpy(gFrom, envstr, MAILTMPLEN);

	if ((envstr = getenv("LOGINNAME")))
		strncpy(gUser, envstr, NETMAXUSER);

	if ((envstr = getenv("PASSWD")))
		strncpy(gPwd, envstr, MAILTMPLEN);
}

/* XXX maybe load_configfile(char *file) later? */
void load_configfile(void)
{
	char *home = getenv("HOME");
	char filepath[FILEMAX]={0};
	FILE *cfgfile = 0;
	const int linesize=128;
	char line[linesize];

	strcat(filepath, home);
	strcat(filepath, "/");
	strcat(filepath, gConfigFile);

#ifdef DEBUG
	printf("loading config file %s\n", filepath);
#endif

	if (NULL == (cfgfile = fopen(filepath, "r"))) {
		if (gVerbose)
			printf("  could not open %s\n", filepath);
		return;
	}

	/*
	 * read a line:
	 * 'variable=value' or 'option'
	 */
	while (0 < fgets(line, linesize, cfgfile)) {
		char *value=line;
		char *var=line;
		int boundchk=0;

		line[strlen(line)-1] = '\0'; /* strip \n */

		if ((var[0] == '#') || (var[0] == '\0'))
			continue; /* skip comment and blank lines */

		while ((line[boundchk] != '=') && (boundchk < linesize))
			boundchk++; /* skip var */
		line[boundchk++] = '\0'; /* terminate var at the '=' */

		/* now boundchk points to the beginning of the 2nd word (if any) */
		value = &line[boundchk];

#ifdef DEBUG
		printf("  '%s' '%s'\n", var, value);
#endif
		if (!strcmp(var, "device")) {
			strncpy(gDevice, value, MAILTMPLEN);

		} else if (!strcmp(var, "sendmail")) {
			strncpy(gSendmail, value, MAILTMPLEN);

		} else if (!strcmp(var, "user")) {
			strncpy(gUser, value, NETMAXUSER);

		} else if (!strcmp(var, "from")) {
			strncpy(gFrom, value, MAILTMPLEN);

		} else if (!strcmp(var, "mailbox")) {
			strncpy(gMailbox, value, MAILTMPLEN);

		} else if (!strcmp(var, "passwd")) {
			/* bitch at user if permissions are uncool */
			struct stat fileinfo;
			if (-1 != stat(filepath, &fileinfo)) {
				if ((fileinfo.st_mode & 0777) ^ (S_IRUSR|S_IWUSR)) {
					fprintf(stderr,
							"%s contains \"passwd\" but is insecure!\n"
							"Please 'chmod 600 %s'.\n",
							filepath, filepath);
					exit(1);
				}
			}
			strncpy(gPwd, value, MAILTMPLEN);

		} else if (!strcmp(var, "delete")) {
			gDelete = 1;
		}
	}
	fclose(cfgfile);
}

void load_cmdline(int argc, char *argv[])
{
	int c, errflag=0, longhelp=0;

#ifdef DEBUG
	printf("reading cmdline\n");
#endif

	while ((c = getopt(argc, argv, "vdp:u:P:f:s:m:?")) != -1) {
		switch (c) {
			case 'v':
				gVerbose = 1;
				break;
			case 'd':
				gDelete = 1;
				break;
			case 'p':
				strncpy(gDevice, optarg, MAILTMPLEN);
				break;
			case 'u':
				strncpy(gUser, optarg, MAILTMPLEN);
				break;
			case 'P':
				strncpy(gPwd, optarg, MAILTMPLEN);
				break;
			case 'f':
				strncpy(gFrom, optarg, MAILTMPLEN);
				break;
			case 's':
				strncpy(gSendmail, optarg, MAILTMPLEN);
				break;
			case 'm':
				strncpy(gMailbox, optarg, MAILTMPLEN);
				break;
			case '?':
				errflag++;
				longhelp=1;
				break;
			default:
#ifdef DEBUG
				printf("  unknown switch %c\n", c);
#endif
				errflag++;
				break;
		}
	}

	/* too many arguments */
	if (optind != argc) {
#ifdef DEBUG
		printf("  too many arguments\n");
#endif
		errflag++;
	}

	if (errflag) {
		Help(argv[0], longhelp);
		exit(1);
	}
}

/* check for fatal errors */
int check_opts(void)
{
	int err=0;

	if (gDevice[0] == '\0') {
		fprintf(stderr, " sync with what device?");
		err++;
	} else if (gSendmail == '\0') {
		fprintf(stderr, " where is sendmail?");
		err++;
	} else if (gMailbox == '\0') {
		fprintf(stderr, " read from what mailbox?");
		err++;
	} else if (gFrom == '\0') {
		fprintf(stderr, " use what from address?");
		err++;
	}

	/* user name and passwd can be entered in mm_login (if necessary), and so
	 * aren't checked here.
	 */

#ifdef DEBUG
	printf("gDevice:   '%s'\n", gDevice);
	printf("gSendmail: '%s'\n", gSendmail);
	printf("gUser:     '%s'\n", gUser);
	printf("gPwd:      '%s'\n", gPwd);
	printf("gMailbox:  '%s'\n", gMailbox);
	printf("gFrom:     '%s'\n", gFrom);
	printf("gDelete:   '%s'\n", gDelete ? "on" : "off");
	printf("gVerbose:  '%s'\n", gDelete ? "on" : "off");
#endif

	return err;
}

void Help(char *prog_name, int longhelp)
{
	printf( "\npilot-mailsync " VERSION "\n");
	printf( "Please read docs/config for a more complete description.\n");
	printf( "usage: %s [options]\n", prog_name);
	printf( " -d                 delete\n"
			" -v                 verbose\n"
			" -p <device>        serial device connected to handheld\n"
			" -u <user>          user name for remote (POP3, IMAP) logins\n"
			" -P <passwd>        password for remote (POP3, IMAP) logins\n"
			" -f <from address>  from address for mail sent from handheld\n"
			" -s <sendmail cmd>  command used to send outgoing mail\n"
			" -m <mailbox>       c-client mailbox string (see docs/mailbox)\n"
			);
	
	if (longhelp) {
		printf( "\nrelevent environment variables:\n"
				"LOGINNAME      login name (POP and IMAP)\n"
				"MAIL           mailbox to read from (see docs/MAILBOX.txt)\n"
				"PASSWD         login password (POP and IMAP)\n"
				"PILOTFROM      'from' address\n"
				"PILOTPORT      tty device connected to PDA\n"
				);
	}

	exit(0);
}
