/*
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License. 
 * 
 * The Original Code is pilot-mailsync.
 * 
 * The Initial Developer of the Original Code is Hollis Blanchard
 * <hollis@andrew.cmu.edu>. Portions created by Hollis Blanchard are Copyright
 * (C) 2000 Hollis Blanchard. All Rights Reserved.
 * 
 * Contributor(s):
 * Frederic Devernay <Frederic.Devernay@sophia.inria.fr>
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <pi-source.h>
#include <pi-socket.h>
#include <time.h> /* XXX pilot-link bug: required before pi-mail.h */
#include <pi-mail.h>
#include <pi-dlp.h>

#include "imap/c-client/c-client.h"

#include "cclientcalls.h"
#include "options.h"
#include "pilot-mailsync.h"
#include "util.h"

/* Some things (gSocket, gDB) HAD to be global so they could be used in
 * the mm_searched callback... and then it made life easier to have the rest
 * global too.
 *
 * Luckily we never have to worry about multiple DB's, sigs, sockets, etc.
 */
struct MailSyncPref gPrefs;
struct MailSignaturePref gSig;
int gSocket=0;
int gDB=0;
int nFromPalm=0;
int nToPalm=0;

/* private */
static void OpenMailDB();
static void CloseMailDB();
static void PalmToMail(const char *from);
static int SendMail(struct Mail *mail, const char *from);


void OpenMailDB(void)
{
	struct PilotUser pi_user;
	int bufsize = 0xffff; /* XXX could probably be smaller */
	unsigned char buffer[bufsize];
	int version;
	int len, size, attr;
	recordid_t id;

	memset(&gPrefs, '\0', sizeof(struct MailSyncPref));
	memset(&gSig, '\0', sizeof(struct MailSignaturePref));

	/* ID the Palm (XXX unused but maybe necessary?) */
	dlp_ReadUserInfo(gSocket,&pi_user);

	/* display message on Palm */
	dlp_OpenConduit(gSocket);

	/* Open the Mail database, store access handle in db */
	if (0 > dlp_OpenDB(gSocket, 0, dlpOpenReadWrite, "MailDB", &gDB)) {
		fprintf(stderr, "Unable to open MailDB\n");
		dlp_AddSyncLogEntry(gSocket, "Unable to open MailDB.\n");
		exit(1);
	}

	if (pi_version(gSocket) <= 0x0100) {
		fprintf(stderr, "pilot-mailsync requires a higher version of"
						"the Palm OS than what you've got.\n");
		CloseMailDB();
		exit(1);
	}

	dlp_ReadAppBlock(gSocket, gDB, 0, buffer, 0xffff);

	/* read in Mail prefs (for signature, filter settings, etc) */
	if (0 < dlp_ReadAppPreference(gSocket, makelong("mail"), mailPrefLocal, 1,
									bufsize, buffer, &size, &version) ) {
		unpack_MailSyncPref(&gPrefs, buffer, bufsize);

#ifdef DEBUG
		printf("gPrefs.syncType = %i\n", gPrefs.syncType);
		printf("gPrefs.getHigh = %i\n", gPrefs.getHigh);
		printf("gPrefs.getContaining = %i\n", gPrefs.getContaining);
		printf("gPrefs.truncate = %i\n", gPrefs.truncate);
		printf("gPrefs.filterTo = %s\n", gPrefs.filterTo);
		printf("gPrefs.filterFrom = %s\n", gPrefs.filterFrom);
		printf("gPrefs.filterSubject = %s\n", gPrefs.filterSubject);
#endif

	} else {
		fprintf(stderr, "Couldn't read Mail preferences\n");
		CloseMailDB();
	}

	/* read in Mail's signature info */
	if (0 < dlp_ReadAppPreference(gSocket, makelong("mail"), mailPrefSig, 1,
									bufsize, buffer, 0, 0) ) {
		unpack_MailSignaturePref(&gSig, buffer, bufsize);
	}

	/* delete emails in the "Deleted" category on the Pilot */
	while (0 <= (len = dlp_ReadNextRecInCategory(gSocket, gDB, mailCtgDeleted,
			buffer, &id, 0, &size, &attr))) {
		dlp_DeleteRecord(gSocket, gDB, 0, id);
	}

	/* delete archived/deleted records */
	dlp_CleanUpDatabase(gSocket, gDB); 
}

void CloseMailDB(void)
{
	char status[1024];

	dlp_ResetLastSyncPC(gSocket);

	/* Close the database */
	dlp_CloseDB(gSocket, gDB);

	sprintf(status,	"pilot-mailsync " VERSION ":\n"
					"%d message%s sent from handheld;\n"
					"%d message%s delivered to handheld.\n",
					nFromPalm, (nFromPalm == 1) ? "" : "s",
					nToPalm, (nToPalm == 1) ? "" : "s");

	/* print out status info */
	fprintf(stderr, status);	
	/* append to hotsync log on Pilot */
	dlp_AddSyncLogEntry(gSocket, status);

	free_MailSyncPref(&gPrefs);
	free_MailSignaturePref(&gSig);
}

/* PalmToMail()
 *
 * loop over all msgs in Outbox and pass them to sendmail (via SendMail)
 */
void PalmToMail(const char *from)
{
	unsigned char buffer[0xffff]; /* XXX could probably be (a lot) smaller */
	int len, size, attr, sendmail_ret;
	recordid_t id;

	/* XXX pilot-link bug: does dlp_ReadNextRecInCategory not require bufsize
	 * anywhere?
	 */
	while (0 <= (len = dlp_ReadNextRecInCategory(gSocket, gDB, mailCtgOutbox,
												 buffer, &id, 0, &size,
												 &attr))) {
		struct Mail mail;

		/* Skip deleted or archived records */
		if ((attr & dlpRecAttrDeleted) || (attr & dlpRecAttrArchived))
			continue;

		unpack_Mail(&mail, buffer, len);

		sendmail_ret = SendMail(&mail, from);
		if (0 != sendmail_ret) {
			/* something bad happened w/ sendmail */
			fprintf(stderr, "ERROR: %s returned %i\n", gSendmail,
					WEXITSTATUS(sendmail_ret));
		} else {
			/* no error, so delete the just-sent mail */
			nFromPalm++;
			dlp_DeleteRecord(gSocket, gDB, 0, id);
		}

		free_Mail(&mail);
	}
}

/* SendMail()
 *
 * does the actual mail transmission.
 *
 * called (only) by PalmToMail.
 */
int SendMail(struct Mail *mail, const char *from)
{
	FILE *sendmail_fd;

	/* open up the sendmail command for writing */
	sendmail_fd = popen(gSendmail, "w");
	if (NULL == sendmail_fd) {
		/* this will NOT catch a bad gSendmail cmd. that can only be caught at
		 * pclose(). read the popen man page carefully. :(
		 */
		perror(gSendmail);
		return -1;
	}

	/* write out the standard email headers */
	fprintf(sendmail_fd, "From: %s\n", from);
	fprintf(sendmail_fd, "To: %s\n", mail->to);
	if (mail->cc)
		fprintf(sendmail_fd, "Cc: %s\n", mail->cc);
	if (mail->bcc)
		fprintf(sendmail_fd, "Bcc: %s\n", mail->bcc);
	if (mail->replyTo)
		fprintf(sendmail_fd, "Reply-To: %s\n", mail->replyTo);
	fprintf(sendmail_fd, "Subject: %s\n",
			mail->subject ? mail->subject : "<none>");
	fprintf(sendmail_fd, "X-Mailer: pilot-mailsync-" VERSION),
	fprintf(sendmail_fd, "\n");
	/* end header */

	fprintf(sendmail_fd, "%s\n", mail->body);

	/* append the signature */
	if (gSig.signature) {
		fprintf(sendmail_fd, "\n%s\n", gSig.signature);
	}

	return pclose(sendmail_fd);
}


void cancel(int sig)
{
	char status[1024];

	dlp_CleanUpDatabase(gSocket, gDB); /* delete archived/deleted records */
	dlp_CloseDB(gSocket, gDB); /* close the database */

	sprintf(status,	"\npilot-mailsync " VERSION ":\nuser aborted!\n"
					"%d message%s sent from handheld;\n"
					"%d message%s delivered to handheld.\n",
					nFromPalm, (nFromPalm == 1) ? "" : "s",
					nToPalm, (nToPalm == 1) ? "" : "s");

	fprintf(stderr, status);
	dlp_AddSyncLogEntry(gSocket, status);

	dlp_EndOfSync(gSocket, dlpEndCodeUserCan);
	exit(1);
}


int main(int argc, char *argv[])
{
	load_options(argc, argv); /* set up globals */
	signal(SIGINT, cancel);

	/* connect to Palm, open Mail db's */
	gSocket = Connect(gDevice);
	OpenMailDB();

	PalmToMail(gFrom); /* send all mail from Palm */

	if (gPrefs.syncType != mailSyncSend) {
		MailToPalm(gMailbox); /* deliver mail to Palm */
	}

	/* disconnect */
	CloseMailDB();
	Disconnect(gSocket);
	return 0;
}

