#!/usr/bin/perl
#
# POCKETMONEY TO QIF v0.1
#
# Exports a PocketMoney database to a Quicken .QIF file
#
# Tested with PocketMoney v1.1 to v1.21
#
# (c) Gareth Watts, May 2000
# gareth@omnipotent.net
# http://pm2qif.sourceforge.net/
#
# PocketMoney (c) Catamount Software
# http://www.catamount.com/
#
#
# This script requires the p5-palm perl modules which can be found at
# http://www.ooblick.com/software/coldsync/
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


#### CONFIGURATION ######

# Path to your home directory (I guess you'd want to change this under Windows)
my $home=$ENV{HOME};

# Where to place the resulting .qif file
my $qiffile="$home/PocketMoney.qif";

# Path to your hotsync directory
my $hotsync_dir="$home/Sync";

# Set this to 1 if you really want to include non-cleared transactions all the time
my $include_noncleared=0;

# File used to keep track of which transfers have already been seen
my $track_file="$home/.pm2qif";

# pdb filenames (shouldn't need to configure this)
my $accounts_pdb="$hotsync_dir/PM-Accounts.pdb";
my $transactions_pdb="$hotsync_dir/PM-Transactions.pdb";



my $debug=0;


#### NON CONFIGURABLE STUFF STARTS HERE ###

use Palm::Raw;
use Digest::MD5 qw(md5_hex);
use strict;
use POSIX qw(strftime);


my $VERSION = "0.1";
my $EPOCH_1904 = 2082844800;
my @AccountFeeType = ('feeFixed','feePercent');
my @AccountTypes = ('typeBank', 'typeCash', 'typeCreditCard');
my @BalanceType = ('balanceAbsolute', 'balanceCleared');
my @CustomColumnType = ('ccAccountCategory', 'ccAccountPayee', 'ccCategory', 'ccCategoryPayee', 'ccIdPayee', 'ccMemo', 'ccPayee');
my @TransactionTypes = ('withdrawal', 'deposit', 'transfer');

print "PocketMoney to QIF v$VERSION\n";

my $Accounts_PDB = new Palm::Raw;
my $Transactions_PDB = new Palm::Raw;

# Parse command line parameters
my $read_all_records=0;
foreach my $arg (@ARGV) {
	$arg =~ /^--help$/i && (usage_exit());
	$arg =~ /^--all$/i && ($read_all_records=1);
	$arg =~ /^--noncleared/i && ($include_noncleared=1);
	$arg =~ /^--clearedonly/i && ($include_noncleared=0);
}

$Accounts_PDB->Load($accounts_pdb) || die "Error opening $accounts_pdb: $!";
$Transactions_PDB->Load($transactions_pdb) || die "Error opening $transactions_pdb: $!";

# Load seen records
my %seen_records=read_track_file($track_file);

open(QIF, ">$qiffile") || die "Error opening QIF file: $!";

# Load accounts
my %Accounts;
foreach my $item (@{$Accounts_PDB->{records}}) {
	my %entry = parse_account_entry($item->{data});
	$Accounts{$entry{account}}=\%entry;
}

# Process transactions
my ($pdb_entries, $cleared_entries, $deleted_entries, $qif_entries, $seen_entries) = (0,0,0,0,0); # tallies
foreach my $item (@{$Transactions_PDB->{records}}) {

	if (!length($item->{data})) {
		($debug>0) && (print "Empty item: skipping\n");
		$deleted_entries++;
		next;
	}
	my @splits;
	my ($recordversion, $date, $transactiontype, $amount, $xrate, $runningbalance,
	 $cleared, $TheRest)
	 =unpack("c x N C x H16 H16 H16 C a*", $item->{data});

	next unless $recordversion; # empty record

	$pdb_entries++;
	next unless ($cleared || $include_noncleared);
	$cleared_entries++ if $cleared;
	
	my ($account,$category,$payee,$chknum,$description,$splitdata) = split( "\0", $TheRest,6);

	my ($numsplits, $splitarr) = unpack("n a*", $splitdata);

	$date=$date - $EPOCH_1904;
	
	for(my $i=0;$i<$numsplits;$i++) {
		my ($splitamount, $splitcategory, $splitdescription);
		($splitamount,$splitarr) = unpack("H16 a*", $splitarr);
		($splitcategory,$splitdescription,$splitarr) = split("\0",$splitarr,3);
		$splitdescription =~ s/[^\S ]/ /sg;
		$splits[$i]={
			'category' => $splitcategory,
			'memo'	=> $splitdescription,
			'amount' => format_hexdouble($splitamount,$Accounts{$account}->{currencydecimalplace})
		};
	}

	($debug>0) && (print "Found transaction:\n");
	($debug>0) && (print "    version: $recordversion\n");
	($debug>0) && (print "    account: $account\n");
	($debug>0) && (print "     amount: ".format_hexdouble($amount,2)."\n");
	($debug>0) && (print "  numsplits: $numsplits\n");
	($debug>0) && (print "   category: $category\n");
	($debug>0) && (print "      payee: $payee\n\n");

	$amount=format_hexdouble($amount,$Accounts{$account}->{currencydecimalplace});
	$xrate=format_hexdouble($xrate);
	$runningbalance=format_hexdouble($runningbalance,$Accounts{$account}->{currencydecimalplace});
	$transactiontype=$TransactionTypes[$transactiontype];
	$description =~ s/[^\S ]/ /sg;

	my $md5=md5_hex($date,$transactiontype,$amount,$category,$payee,$numsplits);
	((!$read_all_records) && $seen_records{$md5}) && (++$seen_entries) && next;
	$seen_records{$md5}=localtime($date)." $category $amount";

	if (!length($account)) {
		die("Fatal Error: One of your transactions has no account associated with it!");
	}
	if (!defined($Accounts{$account})) {
		die("Fatal Error: Account name '$account' wasn't found in PM-Accounts.pdb !");
	}
	# Output QIF entry
	$qif_entries++;
	my $qiftype;
	if ($Accounts{$account}->{type} eq 'typeBank') {
		$qiftype="Bank";
	} elsif ($Accounts{$account}->{type} eq 'typeCash') {
		$qiftype="CCard";
	} elsif ($Accounts{$account}->{type} eq 'typeCreditCard') {
		$qiftype="Cash";
	} else {
		die("Unknown account type found");
	}
	print QIF "!Account\n";
	print QIF "T$qiftype\n";
	print QIF "N$account\n";
	print QIF "^\n";
	print QIF "!Type:$qiftype\n";
	print QIF "D".strftime("%m/%d/%Y",localtime($date))."\n"; # Date
	print QIF "T$amount\n";
	if ($cleared) { print QIF "C*\n"; }
	if (length($chknum)) { print QIF "N$chknum\n"; }
	if (length($description)) { print QIF "M$description\n"; }
	if ($transactiontype eq 'transfer') {
		print QIF "L[$payee]\n";
	} else {
		if (length($payee)) { print QIF "P$payee\n"; }
		if (length($category)) { print QIF "L$category\n"; }
	}
	for(my $i=0;$i<$numsplits;$i++) {
		print QIF "S".$splits[$i]->{category}."\n";
		if (length($splits[$i]->{memo})) { print QIF "E".$splits[$i]->{memo}."\n"; }
		print QIF "\$".$splits[$i]->{amount}."\n";
	}
	print QIF "^\n";
}

close(QIF);

write_track_file($track_file, %seen_records);
print "Valid transactions in .pdb file: $pdb_entries\n";
print "Deleted transactions in .pdb file: $deleted_entries\n";
print "Cleared transactions in .pdb file: $cleared_entries\n";
print "Rejected entries due to duplication: $seen_entries\n";
print "Transactions written to .qif file: $qif_entries\n";
print "Export completed.\n\n";

# done.

############

# Convert a double from a .pdb file into a formatted float - input should be in hex
sub format_hexdouble {
	my ($num, $precision) = @_;
	$precision = 18 unless (defined($precision));
	return sprintf "%.${precision}f", unpack "d", reverse (pack ('H16', $num));
}


sub parse_account_entry {
	my $data=shift;

	my ($recordversion, $balance, $cleared, $type, $limit, $nolimit, $fee,
	$fixedpercent, $totalworth, $absoluteorclearedbalance, $customcolumn, $xrate,
	$currencysuffix, $currencyprefix, $currencydecimalplace, $TheRest)
	= unpack("cx H16 H16 Cx H16 Cx H16 C C C C H16 a4 a4 n a*", $data);

	my ($account, $expires, $number, $phone, $institution) = split ("\0", $TheRest);

	my %result = (
		'recordversion' => $recordversion,
		'balance' => format_hexdouble($balance,$currencydecimalplace),
		'cleared' => $cleared,
		'type'	=> $AccountTypes[$type],
		'limit' => format_hexdouble($limit,$currencydecimalplace),
		'nolimit' => $nolimit,
		'fee' => format_hexdouble($fee,$currencydecimalplace),
		'fixedpercent' => $AccountFeeType[$fixedpercent],
		'totalworth' => $totalworth,
		'absoluteorclearedbalance' => $BalanceType[$absoluteorclearedbalance],
		'customcolumn' => $CustomColumnType[$customcolumn],
		'xrate' => format_hexdouble($xrate,$currencydecimalplace),
		'currencysuffix' => $currencysuffix,
		'currencyprefix' => $currencyprefix,
		'currencydecimalplace' => $currencydecimalplace,
		'account' => $account,
		'expires' => $expires,
		'number' => $number,
		'phone' => $phone,
		'institution' => $institution
	);
	($debug>0) && (print "Found account:\n");
	($debug>0) && (print "  version: $recordversion\n");
	($debug>0) && (print "  account: $account\n");
	($debug>0) && (print "     type: $type (".$AccountTypes[$type].")\n\n");
	return(%result);
}

sub read_track_file {
	my $filename=shift;
	my %seen_records;

	if (!-f $filename) {
		return( () );
	}
	open(TRACK_FILE, $filename) || die "Error opening $track_file: $!";
	while(<TRACK_FILE>) {
		(/^seen: (\w+)(.*)/) && ($seen_records{$1}=$2);
	}
	close(TRACK_FILE);
	return(%seen_records);
}

sub write_track_file {
	my ($filename, %seen_records)=@_;

	open(TRACK_FILE, ">$filename") || die "Error opening $track_file: $!";
	foreach my $md5 (keys %seen_records) {
		print TRACK_FILE "seen: $md5 $seen_records{$md5}\n";
	}
	close(TRACK_FILE);
}

sub usage_exit {
	print "Usage: pm2qif.pl [OPTION]\n";
	print "  --help        - Display this help page\n";
	print "  --all         - Export all records; not just new ones\n";
	print "  --noncleared  - Include records that you haven't yet marked as 'cleared'\n";
	print "  --clearedonly - Include only records that you have marked as 'cleared'\n";
	print "\n";
	print "(c) 2000 Gareth Watts <gareth\@omnipotent.net>\n";
	print "\n";
	exit;
}

