/*
 * syncal.c
 * syncal Version 0.8.6
 *
 * Copyright (c) 1999-2000, John Franks
 *
 * Japanese character set modifications contributed by Hiroshi Kawashima
 *
 * Contributions by Bradley C. Kuszmaul.
 * Contributions by Michael D. Beynon.
 * This is free software, licensed under the GNU Public License V2.
 * See the file COPYING for details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include "syncal.h"


int	cat_type[MAX_PILOT_CATS];

struct A_list
	archive_list,
	ical_list,
	pilot_list,
	new_list;

int
	use_datebk3 = FALSE,
	pi_date_db,
	pilot_sd,
	no_action,
	quiet,
	lnk_verbose,
	verbose,
	hide_private = FALSE,
	use_private_text = FALSE;

char
	tmp_file[SMALLLEN],
	cal_file[SMALLLEN],
	archive_file[SMALLLEN],
	private_text[SMALLLEN],
	*extra_stuff;

enum CalType { MAIN_CAL, SUB_CAL, ARCHIVE };

static unsigned ahash( char *);

static int
	linecount;

static void
	usage( char *),
	initialize( int, char **),
	init_pilot( void),
	syncal_timeout( int),
	add_extra( char *),
	parse_error( char *),
	appt2contents( char *, struct Appointment *, unsigned *),
	dates2ical( char *, struct Appointment *),
	fill_ical_ent( char *, struct A_entry *, recordid_t, unsigned),
	create_uid( char *, recordid_t ),
	archive2mem( void),
	icalfiles2mem( void),
	ical2mem( struct A_list *, FILE *, int, enum CalType),
	pilot2mem( void),
	set_category( struct A_entry *),
	backups( void),
	ilist2pilot( void);


static	FILE
 	*calfp,
	*archfp,
	*newfp;

static struct PilotUser
	pi_user;

static struct A_entry
	*make_new_entry( char *, recordid_t, unsigned, char *);


static char
	*filename,
	linebuf[MIDLEN],
	my_username[128],
	my_hostname[SMALLLEN] = "",
	rec_id_str[SMALLLEN],
	pilot_device[SMALLLEN];


static void
usage( char *name )
{
	fprintf( stderr, 
 "%s\nUsage: %s [-a archive_file] [-c calendar] [-d pilot_device]\n"
 "[-p private_text] [-u user] [-hinsv]\n"
 "\t-a use 'archive_file' as archive -- default is ($HOME/.calendar.archive\n"
 "\t-c use 'calendar' as calendar file -- default is ($HOME/.calendar\n"
 "\t-d use 'pilot_device' as target -- default is /dev/pilot\n"
 "\t-h print this message\n"
 "\t-l link verbose -- display info about linked appointments\n"
 "\t-n only display actions to be taken, don't do them\n"
 "\t-q quiet -- no information (including errors) is printed\n"
 "\t-u install name 'user' on Pilot as the username field of PilotUser\n"
 "\t-3 using with datebk3/4 (syncal will remember this for future invocations)\n"
 "\t-s hide private datebook entries\n"
 "\t-p use 'private_text' as private datebook entry in ical\n"
 "\t-v verbose -- display info about appointments as they are transferred\n\n",
	VERSION,  name);
}


int
main(int argc, char *argv[])
{
	initialize( argc, argv);

	/* Read in the ical calendar and ical archive files */
	archive2mem( );
	icalfiles2mem( );

	/* Upload the entire DateBookDB to pilot_list */
	init_pilot( );
	pilot2mem( );

	/* Write the new pilot entries to new_list */
	pilotlist2newlist( );

	/* Write the new ical calendar entries to the pilot and new_list */
	ilist2pilot( );

	/* Copy old items to new_list reconciling changes */
	reconcile( );
	sort_list( &new_list);

	/* Print it to tmp_file */
	mem2ical( &new_list, newfp);

	if ( newfp)
		fclose( newfp);

	free_Alist( &ical_list);
	free_Alist( &pilot_list);
	free_Alist( &archive_list);

	cleanup( dlpEndCodeNormal);
	backups( );
	exit( 0);
}


static void
initialize( int argc, char *argv[])
{
	struct stat
		stat_buf;

	int	a_opt,
		statresult,
		i;

	char
		*cp,
		user_home_dir[SMALLLEN],
		home_dir[SMALLLEN];

	int 	c;

	cat_type[0] = CAT_REG;
	for ( i = 1; i < MAX_PILOT_CATS; i++) {
		cat_type[i] = CAT_UNKNOWN;
	}

	if ( (cp = getenv( "PILOTPORT")) != NULL )
		mystrncpy( pilot_device, cp, SMALLLEN);
	else
		mystrncpy( pilot_device, "/dev/pilot", SMALLLEN);

	if ( (cp = getenv( "HOME")) != NULL ) {
		mystrncpy( user_home_dir, cp, SMALLLEN);
		mystrncat( user_home_dir, "/", SMALLLEN);
	}
	else
		user_home_dir[0] = '\0';

	if ( SYNCAL_HOME_DIR[0] == '/' )
		mystrncpy( home_dir, SYNCAL_HOME_DIR, SMALLLEN);
	else if ( user_home_dir[0] ) {
		mystrncpy( home_dir, user_home_dir, SMALLLEN);
		mystrncat( home_dir, SYNCAL_HOME_DIR, SMALLLEN);
	}
	else {
		mystrncat( home_dir, "./", SMALLLEN);
		mystrncat( home_dir, SYNCAL_HOME_DIR, SMALLLEN);
	}

	if ( user_home_dir[0] ) {
		mystrncpy( cal_file, user_home_dir, SMALLLEN);
		mystrncat( cal_file, MAIN_CALENDAR, SMALLLEN);
	}
	else {
		mystrncpy( cal_file, "./", SMALLLEN);
		mystrncat( cal_file, MAIN_CALENDAR, SMALLLEN);
	}
	/* Follow symlinks properly. */

	statresult = lstat(cal_file, &stat_buf);
	if ((statresult >= 0) && S_ISLNK(stat_buf.st_mode)) {
		char symlink_contents[SMALLLEN];
		int n_chars = readlink(cal_file, symlink_contents, SMALLLEN);

		if (n_chars<=0) {
			if ( !quiet )
				fprintf(stderr, ERRMSG3, cal_file);
			exit(1);
		}
		else if ((unsigned) n_chars >= SMALLLEN) {
			if ( !quiet ) 
				fprintf(stderr, ERRMSG4, cal_file);
			exit(1);
		}

		symlink_contents [n_chars] = 0;
		if (symlink_contents[0] != '/') {
			/* it is a relative path. */
			char *slashptr = strrchr(cal_file, '/');
			/* find the last /, and append the symlink */

			if (slashptr==NULL) {
				/* there is no / at the end, so just use the 
				   symlink_contents as the path. */

				mystrncpy (cal_file,
					   symlink_contents, SMALLLEN);
			} 
			else {
				int ret;
				/* there's a / at the end */
				if ( (*++slashptr == '\0') && !quiet ) {
					fprintf(stderr, ERRMSG5);
					exit(1);
				}
				*slashptr = '\0';
				ret = mystrncat(cal_file,
						symlink_contents, SMALLLEN);
				if ( (ret < 0) && !quiet  ) {
					fprintf(stderr, ERRMSG6, cal_file,
						SMALLLEN);
					exit(1);
				}
			}
		}
		else {
			mystrncpy (cal_file, symlink_contents, SMALLLEN);
		}
	} 

	mystrncpy( archive_file, home_dir, SMALLLEN);
	mystrncat( archive_file, "/", SMALLLEN);
	mystrncat( archive_file, ARCHIVE_FILE, SMALLLEN);
	mystrncpy( tmp_file, home_dir, SMALLLEN);
	mystrncat( tmp_file, "/", SMALLLEN);
	mystrncat( tmp_file, SYNCAL_TMP_FILE, SMALLLEN);

	my_username[0] = '\0';
	ical_list.first = ical_list.last = NULL;
	ical_list.num = 0;
	ical_list.default_alarm = ALARM_NONE;
	ical_list.defalrm_units = ALARM_MINUTES;
  	pilot_list.first = pilot_list.last = NULL;
  	pilot_list.num = 0;
	pilot_list.default_alarm = ALARM_NONE;
	pilot_list.defalrm_units = ALARM_MINUTES;
	archive_list.first = archive_list.last = NULL;
	archive_list.num = 0;
	archive_list.default_alarm = ALARM_NONE;
	archive_list.defalrm_units = ALARM_MINUTES;
	new_list.first = new_list.last = NULL;
	new_list.num = 0;
	new_list.default_alarm = ALARM_NONE;
	new_list.defalrm_units = ALARM_MINUTES;
  
	quiet = verbose = lnk_verbose= FALSE;
	no_action = FALSE;
	a_opt = FALSE;

	while ((c = getopt(argc, argv, 	"a:c:d:hlnqp:su:v3")) != -1) {
		switch ((char) c) {
		case 'a':
			mystrncpy( archive_file, optarg, SMALLLEN);
			a_opt = TRUE;
			break;
		case 'c':
			mystrncpy( cal_file, optarg, SMALLLEN);
			break;
		case 'd':
			mystrncpy( pilot_device, optarg, SMALLLEN);
			break;
		case 'h':
			usage( argv[0]);
			exit( 0);
		case 'l':
			lnk_verbose = TRUE;
			break;
		case 'n':
			no_action = TRUE;
			verbose = TRUE;
			break;
		case 'q':
			quiet = TRUE;
			break;
		case 'p':
		        mystrncpy( private_text, optarg, SMALLLEN);
			use_private_text = TRUE;
			break;
		case 'v':
			verbose = TRUE;
			break;
		case 's':
		        hide_private = TRUE;
			break;
		case 'u':
			mystrncpy( my_username, optarg, 128);
			break;
		case '3':
			use_datebk3 = TRUE;
			break;
			break;
		default:
			usage( argv[0]);
			exit( 1);
		}
	}

	if ( (!a_opt) && (stat( home_dir, &stat_buf) == -1)) {
		if ( (errno != ENOENT) || (mkdir( home_dir, 0777) != 0) ) {
			fprintf( stderr, "Can't create directory %s\n", home_dir);
			exit( 1);
		}
		else {
			int
				fd;

			if ( (fd = creat( archive_file, 0666)) < 0) {
				fprintf( stderr, "Can't create %s\n", archive_file);
				exit( 1);
			}
			close( fd);
		}
	}

	if ( (calfp = fopen( cal_file, "r")) == NULL ) {
		fprintf( stderr, "Can't open calendar file: %s\n", cal_file);
		exit( 1);
	}

	if ( (archfp = fopen( archive_file, "r")) == NULL ) {
		fprintf( stderr, "Can't open archive file: %s\n", archive_file);
		exit( 1);
	}

	if ( no_action)
		newfp = NULL;
	else
		if ( (newfp = fopen( tmp_file, "w")) == NULL ) {
			fprintf( stderr, "Can't open tmp file: %s\n", tmp_file);
			exit( 1);
		}

	extra_stuff = (char *) malloc( SMALLLEN); 
	if ( extra_stuff == NULL) {
		fprintf( stderr, "Out of memory\n");
		exit( 1);
	}
}

static void
init_pilot( void )
{
	char
		buffer[0xffff];
	int
		flag,
		err;

	struct pi_sockaddr
		addr;

	struct AppointmentAppInfo
		apt_ainfo;

	signal( SIGALRM, syncal_timeout );
	alarm( HOTSYNC_TIMEOUT);

	if ( !quiet ) {
		printf( "Please press the HotSync button now...  ");
		fflush( stdout);
	}

	if (!(pilot_sd = pi_socket(PI_AF_SLP, PI_SOCK_STREAM, PI_PF_PADP))){
		if ( !quiet )
			fprintf(stderr, "pi_socket failed\n");
		exit(1);
	}
    
	addr.pi_family = PI_AF_SLP;
	strcpy(addr.pi_device, pilot_device);

	if ( pi_bind(pilot_sd, (struct sockaddr*)&addr, sizeof(addr)) <0 ) {
		if ( !quiet )
			fprintf(stderr, "pi_bind failed\n");
		exit(1);
	}

	if ( pi_listen(pilot_sd, 1) < 0 ) {
		if ( !quiet )
			fprintf(stderr, "pi_listen failed\n");
		exit(1);
	}

	if ( (pilot_sd = pi_accept(pilot_sd, 0, 0)) < 0 ) {
		if ( !quiet )
			fprintf(stderr, "pi_accept failed\n");
		exit(1);
	}
	
	signal( SIGALRM, SIG_DFL);  /* Turn off timeout alarm */
	alarm( 0);

	if ( !quiet )
		printf("Connected to Palm device.\n");

	/* Ask the pilot who it is. */
	if ( dlp_ReadUserInfo( pilot_sd, &pi_user) < 0 ) {
		if ( !quiet )
			fprintf( stderr, "dlp_ReadUserInfo failed\n");
		exit( 1);
	}

	if ( my_username[0] != '\0') {
		mystrncpy( pi_user.username, my_username, 128);
		if ( !quiet)
			printf( "Installing '%s' as username on Pilot\n",
						my_username);
		pi_user.userID = ahash( pi_user.username);
	}
	if ( pi_user.userID == 0)
		pi_user.userID = ahash( pi_user.username);
	if ( ++(pi_user.lastSyncPC) == 0 )
		++(pi_user.lastSyncPC);

	/* Tell user (via Pilot) that we are starting things up */
	if ( dlp_OpenConduit( pilot_sd) < 0 ) {
		if ( !quiet )
			fprintf( stderr, "dlp_OpenConduit failed\n");
		exit( 1);
	}

	/* Open the Datebook's database, 
	   store access handle in pi_date_db */

	if ( no_action )
		flag = 0x80;
	else
		flag = 0x80 | 0x40;

	if( (err = dlp_OpenDB( pilot_sd, 0, flag, "DatebookDB", &pi_date_db)) < 0) {
		if ( !quiet )
			fprintf(stderr, "Unable to open DatebookDB: Err=%d\n", err);
		dlp_AddSyncLogEntry( pilot_sd, "Unable to open DatebookDB.\n");
		pi_close(pilot_sd);
		exit(1);
	}
	dlp_ReadAppBlock(pilot_sd, pi_date_db, 0, buffer, 0xffff);
	unpack_AppointmentAppInfo(&apt_ainfo, buffer, 0xffff);
	get_pcats( &apt_ainfo);
}

/*
 * static void ilist2pilot ( struct A_list *lptr)
 * Install all entries with pilot id 0 from the list pointed to
 * by lptr into the Pilot.  Also add them to new_list.
 * Then delete each of those entries from the list *lptr.
 * Continue through all entries in the list and set the 
 * appropriate datebk_magic strings and ical_match in the archive_list.
 */


static void
ilist2pilot ( void)
{
	struct A_entry 	*ep,
			*aptr,
			*next;

	ep = ical_list.first;

	while ( ep != NULL ) {
		char	ent_buf[MIDLEN];

		recordid_t	new_id = 0;

		if ( ep->pi_rec_id != 0 ) {
			aptr = ismatch( ep->pi_rec_id, &archive_list);
			if ( aptr) {
			        ep->datebk_magic = str_dup(aptr->datebk_magic);
				aptr->ical_match = ep;
			}
			set_category( ep);
			ep = ep->next;
			continue;
		}
		ep->type |= FOR_PILOT;

		if ( (!no_action) && (ep->type | (IS_NEW + IS_REVISION ))) {
			write2pilot( &(ep->appt), 0, &new_id, ep->category);
		}
			

		next = ep->next;
		if (ep->type | (IS_NEW + IS_REVISION )) {
			fill_ical_ent( ent_buf, ep, new_id, ep->type);
			ep->pi_rec_id = new_id;
			add2list( ent_buf,  ep, &new_list);
		}

		ep = next;
	}
}

static void
set_category( struct A_entry *ep)
{
	if ( (ep->datebk_magic == NULL) || ep->datebk_magic[0] == '\0')
		ep->datebk_magic = strdup( "##@@@@@@@\n");

	ep->datebk_magic[4] = '@' + ep->category;

	if ( (strcmp( ep->datebk_magic, "##@@@@@@@\n") == 0 ) ||
				(strcmp( ep->datebk_magic, "##@@@@@@@@@@@@@@@@@\n") == 0 )) {
		free( ep->datebk_magic);
		ep->datebk_magic = NULL;
	}
}


/*
 * static void pilot2mem ( void)
 * Read all the DateBookDB entries from the pilot and install them
 * in memory in the A_list structure pointed to by lptr
 */


static void
pilot2mem ( void)
{
	unsigned char
		buffer[0xffff];

	int	attr,
		i,
		cat,
		len,
	        len_magic;

	struct A_entry
		*newp;
	struct A_list *lptr;
	recordid_t
		pilot_id;

	char
		*cp,
		ent_buf[MIDLEN];


	lptr = &pilot_list;
	lptr->first = lptr->last = NULL;
	lptr->num = 0;
	cat = 0;

	for ( i = 0; ; i++) {
		len = dlp_ReadRecordByIndex( pilot_sd, pi_date_db,
					i, buffer, &pilot_id, 0, &attr, &cat);
		if ( len < 0)
			break;
		/* Skip deleted/archived records */
		if (attr & (dlpRecAttrDeleted + dlpRecAttrArchived))
			continue;

		/* Skip private entries if '-s' flag has been used */
		if (hide_private && (attr & (dlpRecAttrSecret)))
		        continue;
		
		
		/* create empty new entry */
		newp = make_new_entry( "<empty>", pilot_id, 0, pi_user.username);

		/* Fill in newp->appt */
		unpack_Appointment(&(newp->appt), buffer, len);

		/* Set 'private_text' to Desc_field if '-p' has been used */

		if(use_private_text && (attr & (dlpRecAttrSecret))){
		  mystrncpy(newp->appt.description, private_text, SMALLLEN);
		}
		
		/* Set the category */
		newp->category = cat;

		/* Put ical_ent in ent_buf */
		fill_ical_ent( ent_buf, newp, pilot_id, 0);

		/* Put the ical entry in newp->ical_ent */
		if ( newp->ical_ent) free( newp->ical_ent);
		newp->ical_ent = strdup( ent_buf);


		/* Set the type correctly */
		newp->type = (newp->appt.event ? IS_EVENT : 0 );

		cp = newp->appt.note;
		len_magic = is_datebk_magic(cp, &(newp->datebk_magic_ver));
		if (len_magic > 0) {
			use_datebk3 = TRUE;
			if ( tolower(cp[2]) == 'f' )
				newp->type |= IS_TODO;
			if ( tolower(cp[2]) == 'c' )
				newp->type |= (IS_TODO + IS_DONE);
			if ( tolower(cp[2]) == 'd' )
				newp->type |= IS_DONE;

			newp->datebk_magic = str_ndup(cp, len_magic);
		}
		else
		        newp->datebk_magic = NULL;
		
		/* Add the new entry to the list */
		if ( lptr->last == NULL) {
			lptr->first = lptr->last = newp;
		}
		else {
			lptr->last->next = newp;
			lptr->last = newp ;
		}		
		lptr->last->next = NULL;
		lptr->num++;
	}
	sort_list( lptr );
}

/*
 * static
 * fill_ical_ent( char *buf, A_entry *ep, recordid_t pilot_id, unsigned type)
 * 
 * Fill buf with ical_ent appropriate for *ap. Also use correct owner and
 * hilite and category
 */

static void
fill_ical_ent( char *buf, struct A_entry *ep, recordid_t pilot_id, unsigned type)
{
	int
		start,
		length;

	struct Appointment
		*ap;

	char
		*own_p,
		lbuf[MIDLEN + SMALLLEN];

	ap = &(ep->appt);
	own_p = ep->owner;

	if (ap->event) {
		mystrncpy( buf, "Note [\nLength [30]\n", MIDLEN);
	}
	else {
		mystrncpy( buf, "Appt [\n", MIDLEN);
		start = ap->begin.tm_hour * 60 + ap->begin.tm_min;
		if ( (ap->end.tm_hour == 0) && (ap->end.tm_min == 0))
			length = 24*60 - start;
		else
			length = ap->end.tm_hour * 60 + ap->end.tm_min - start;

		if ( is_datebk_magic( ap->note, NULL) &&
		     ( tolower( ap->note[6]) == 's' ) &&
		     ( ap->repeatFrequency == 1)) {
			length += 24*60;
			ap->repeatType = repeatNone;
			ap->repeatForever = FALSE;
			ap->repeatFrequency = 0;
		}
		else if (  (ap->end.tm_mday != ap->begin.tm_mday) && length < 0 )
			length += 24*60;


		sprintf(buf + strlen( buf), "Start [%d]\n", start);
		sprintf(buf+strlen( buf), "Length [%d]\n", length);
	}

	if (ap->alarm && !ap->event && (ap->advanceUnits != 2)) {
		int	val;
		switch (ap->advanceUnits) {
		case 0:
			val = ap->advance;
			break;
		case 1:
			val = 60 * ap->advance;
			break;
		}
		sprintf( lbuf, "Alarms [%d]\n", val);
		mystrncat( buf, lbuf, MIDLEN);
	}
	if (verbose && ap->alarm && ap->event && (ap->advanceUnits != 2)) {
		if ( !quiet )
			fprintf( stderr,
			   "Minute or hour alarm on untimed event ignored: %s\n",
				 ap->description);
	}
	
	mystrncat( buf, "Uid [", MIDLEN);
	create_uid( lbuf, pilot_id);
	mystrncat( buf, lbuf, MIDLEN);
	mystrncat( buf, "]\n", MIDLEN);
	
	if ( own_p && *own_p) {
		mystrncat( buf, "Owner [", MIDLEN);
		mystrncat( buf, own_p, MIDLEN);
		mystrncat( buf, "]\n", MIDLEN);
	}

	appt2contents( lbuf, ap, &type );
	mystrncat( buf, lbuf, MIDLEN);

	if (ap->alarm && (ap->advanceUnits == 2)) {
		sprintf( lbuf, "Remind [%d]\n", ap->advance);
		mystrncat( buf, lbuf, MIDLEN);
	}
	else
		mystrncat( buf, "Remind [1]\n", MIDLEN);

	mystrncat( buf, "Hilite [", MIDLEN);
	switch ( ep->hilite) {
	case hi_always:
		mystrncat( buf, "always]\n", MIDLEN);
		break;
	case hi_never:
		mystrncat( buf, "never]\n", MIDLEN);
		break;
	case hi_expire:
		mystrncat( buf, "expire]\n", MIDLEN);
		break;
	case hi_holiday:
		mystrncat( buf, "holiday]\n", MIDLEN);
		break;
	default:
		mystrncat( buf, "always]\n", MIDLEN);
	}

	if ( type & IS_TODO)
		mystrncat( buf, "Todo []\n", MIDLEN);
	if ( type & IS_DONE)
		mystrncat( buf, "Done []\n", MIDLEN);
	
	dates2ical( lbuf, ap );
	mystrncat( buf, lbuf, MIDLEN);

	sprintf( lbuf, "PilotRecordId [%lx]\n", pilot_id);
	mystrncat( buf, lbuf, MIDLEN);

	mystrncat( buf, "]\n", MIDLEN);  /* END of the entry */
}

#if defined(Japanese)
#define isSjis1stByte(c) \
    (((c) >= 0x81 && (c) <= 0x9f) || ((c) >= 0xe0))

static unsigned int CharSjis2Jis(unsigned int code)
{
    unsigned char sh,sl,jh,jl;

    sh = (code >> 8) & 0xff;
    sl = code & 0xff;

    if (sl <= 0x9e){
	if (sh <= 0x9f){
	    jh = (sh - 0x71) * 2 + 1;
	} else {
	    jh = (sh - 0xb1) * 2 + 1;
	}
	jl = sl - 0x1f;
	if (sl >= 0x80)
	    jl--;
    } else {
	if (sh <= 0x9f){
	    jh = (sh - 0x70) * 2;
	} else {
	    jh = (sh - 0xB0) * 2;
	}
	jl = sl - 0x7e;
    }
    return (jh << 8) + jl;
}

static void Sjis2Euc(unsigned char *buf)
{
    unsigned char *p, *q;
    int half_byte = 0;
    int word16, first_byte;

    if (buf == NULL)
	return;
    for (p = buf, q = buf; *p; p++) {
	if (half_byte) {
	    word16 = CharSjis2Jis((first_byte << 8) | *p) | 0x8080; /* to EUC */
	    half_byte = 0;
	    *q++ = word16 >> 8;
	    *q++ = word16 & 0xff;
	} else {
	    if (isSjis1stByte(*p)) {
		/* shift JIS kanji first byte */
		half_byte = 1;
		first_byte = *p;
	    } else {
		*q++ = *p;
	    }
	}
    }
    *q++ = '\0';
}
#endif

/*
 * appt2contents( char* lbuf, struct Appointment *ap, unsigned *type_p)
 * Take the contents of a Pilot appointment struct's note and
 * description fields and place them in an ical Contents  format.
 * In addition look for magic strings ##f@@@@@@ and ##c@@@@@@ at beginning
 * of note.  We don't put them in lbuf, but leave them in ap->note,
 * and set *type_p bits IS_TODO and IS_DONE appropriately.  This type
 * is only used in fill_ical_ent().
 */

static void
appt2contents( char* lbuf, struct Appointment *ap, unsigned *type_p)
{
	int n = MIDLEN - 6,
	    len;
	char	*cp,
		*cp2;
#if defined(Japanese)
	if (ap->description != NULL)
	    Sjis2Euc(ap->description);
	if (ap->note != NULL)
	    Sjis2Euc(ap->note);
#endif

	cp = ap->description;
	mystrncpy( lbuf, "Contents [", MIDLEN);  /* start of contents */

	cp2 = lbuf + strlen( lbuf);

	if (ap->note && *(ap->note)) {
		while ( cp && *cp && n > 0 && (cp2 < lbuf + MIDLEN) ) {
			n--;
			if ( (*cp == ']') || (*cp == '[') )
				*cp2++ = '\\';
			*cp2++ = *cp++;
		}
		cp = ap->note;
		len = is_datebk_magic( cp, NULL);
		if ( len > 0) {
			if ( tolower(cp[2]) == 'f' )
				*type_p |= IS_TODO;
			if ( tolower(cp[2]) == 'c' )
				*type_p |= (IS_TODO + IS_DONE);
			cp += len;
		}

		if ( cp && *cp ) {    /* There is still a note */
			*cp2++ = '\n';
			*cp2++ = '\n';
		}
		
		while ( cp && *cp && n > 0 && (cp2 < lbuf + MIDLEN)) {
			n--;
			if ( (*cp == ']') || (*cp == '[') )
				*cp2++ = '\\';
			*cp2++ = *cp++;
		}
		*cp2 = '\0';
	}
	else {
		while ( cp && *cp && n > 0 && (cp2 < lbuf + MIDLEN)) {
			n--;
			if ( (*cp == ']') || (*cp == '[') )
				*cp2++ = '\\';
			*cp2++ = *cp++;
		}
		*cp2 = '\0';
	}
	mystrncat( lbuf, "]\n", MIDLEN);  /* end of contents */
}

static void
dates2ical( char* lbuf, struct Appointment *ap)
{
	
	char	buf[MIDLEN];
	int	i;

	mystrncpy( lbuf, "Dates [", MIDLEN);

	if ( ap->repeatType == repeatNone ) {
		sprintf( buf, "Single %d/%d/%d",
			 ap->begin.tm_mday, ap->begin.tm_mon + 1,
			 ap->begin.tm_year + 1900);
		mystrncat( lbuf, buf, MIDLEN);
		mystrncat( lbuf, " End\n]\n", MIDLEN);
		return;
	}
	else if ( ap->repeatType == repeatDaily ) {
		sprintf( buf, "Days %d/%d/%d %d",
			 ap->begin.tm_mday, ap->begin.tm_mon + 1,
			 ap->begin.tm_year + 1900, ap->repeatFrequency);
		mystrncat( lbuf, buf, MIDLEN);
	}
	else if ( ap->repeatType == repeatYearly ) {
		sprintf( buf, "Months %d/%d/%d %d",
			 ap->begin.tm_mday, ap->begin.tm_mon + 1,
			 ap->begin.tm_year + 1900, 12 * ap->repeatFrequency);
		mystrncat( lbuf, buf, MIDLEN);
	}
	else if ( ap->repeatType == repeatMonthlyByDate ) {
		sprintf( buf, "Months %d/%d/%d %d",
			 ap->begin.tm_mday, ap->begin.tm_mon + 1,
			 ap->begin.tm_year + 1900, ap->repeatFrequency);
		mystrncat( lbuf, buf, MIDLEN);
	}
	else if ( ap->repeatType == repeatMonthlyByDay ) {
		if ( ap->repeatDay > domLastSun ) {
			sprintf( buf, "ComplexMonths %d 1", ap->repeatFrequency);
			mystrncat( lbuf, buf, MIDLEN);
			sprintf( buf, " %d/%d/%d Backward ByWeek %d",
				ap->begin.tm_mday, ap->begin.tm_mon + 1,
				ap->begin.tm_year + 1900, (ap->repeatDay % 7) + 1);
			mystrncat( lbuf, buf, MIDLEN);
		}
		else {
			sprintf( buf, "ComplexMonths %d %d", ap->repeatFrequency,
				 (ap->repeatDay/7) + 1 );
			mystrncat( lbuf, buf, MIDLEN);
			sprintf( buf, " %d/%d/%d Forward ByWeek %d",
				ap->begin.tm_mday, ap->begin.tm_mon + 1,
				ap->begin.tm_year + 1900, (ap->repeatDay % 7) + 1);
			mystrncat( lbuf, buf, MIDLEN);
		}
	}
	else if ( ap->repeatType == repeatWeekly ) {
		int	j,
			found;

		found = 0;
		for (j = 0; j < 7; j++)
			if (ap->repeatDays[j]) {
				found++;
			}
		if ((found > 1) && (ap->repeatFrequency == 1) ) {
			mystrncat( lbuf, "WeekDays  ", MIDLEN);
			for (j = 0; j < 7; j++)
				if (ap->repeatDays[j]) {
					sprintf(buf, "%d ", j + 1);
					mystrncat( lbuf, buf, MIDLEN);
				}
			mystrncat( lbuf,
				   "Months  1 2 3 4 5 6 7 8 9 10 11 12", MIDLEN);
		}
		else {
			sprintf( buf, "Days %d/%d/%d %d",
				 ap->begin.tm_mday, ap->begin.tm_mon + 1,
				 ap->begin.tm_year + 1900, 7*ap->repeatFrequency);
			mystrncat( lbuf, buf, MIDLEN);
		}
		if ((found > 1) && (ap->repeatFrequency > 1) )
			if ( !quiet )
				fprintf(stderr, "Incomplete translation of %s\n",
					ap->description);
	}

	sprintf( buf, "\nStart %d/%d/%d",
			 ap->begin.tm_mday, ap->begin.tm_mon + 1,
			 ap->begin.tm_year + 1900);
	mystrncat( lbuf, buf, MIDLEN);

	if ( !ap->repeatForever ) {
		sprintf( buf, "\nFinish %d/%d/%d",
			 ap->repeatEnd.tm_mday, ap->repeatEnd.tm_mon + 1,
			 ap->repeatEnd.tm_year + 1900);
		mystrncat( lbuf, buf, MIDLEN);
	}
	if ( ap->exceptions > 0 ) {
		for ( i = 0; i < ap->exceptions; i++) {
			sprintf( buf, "\nDeleted %d/%d/%d",
					ap->exception[i].tm_mday,
					ap->exception[i].tm_mon + 1,
					ap->exception[i].tm_year + 1900);
			mystrncat( lbuf, buf, MIDLEN);
		}
	}

	mystrncat( lbuf, " End\n]\n", MIDLEN);
	return;
}




static void
archive2mem( void )
{
	struct A_list *lptr;

	lptr = &archive_list;
	lptr->first = lptr->last = NULL;
	lptr->num = 0;

	filename = archive_file;
	ical2mem( lptr, archfp, 0, ARCHIVE );
	fclose( archfp);

	sort_list( lptr);
}

static void
icalfiles2mem( void)
{

	struct stat
		stat_buf;

	int i;

	char	*cp,
		fname[SMALLLEN];

	struct A_entry
		*lep,
		*nep;

	struct A_list *lptr;

	lptr = &ical_list;
	lptr->first = lptr->last = NULL;
	lptr->num = 0;

	filename = cal_file;
	ical2mem( lptr, calfp, 0, MAIN_CAL );
	fclose( calfp);

	filename = fname;

	for ( i = 1; i < MAX_PILOT_CATS; i++) {
		FILE	*cfp;

		if ( archv_cats[i][0] == '\0')
			continue;

		mystrncpy( fname, archive_file, SMALLLEN);
		if ( (cp = strrchr( filename, '/')) == NULL)
			fname[0] = '\0';
		else
			*(++cp) = '\0';

		mystrncat( filename, archv_cats[i], SMALLLEN - 9);

		if ( (stat( filename, &stat_buf) == 0 ) &&
					     ( S_ISREG(stat_buf.st_mode))) {
				cat_type[i] = CAT_REG;
		} 

		if ( cat_type[i] == CAT_UNKNOWN) {
			mystrncat( filename, ICAL_SUFFIX, SMALLLEN - 4);
			if ( (stat( filename, &stat_buf) == 0 ) &&
					     ( S_ISREG(stat_buf.st_mode))) {
				cat_type[i] = CAT_ICAL;
			}
		}

		if ( cat_type[i] == CAT_UNKNOWN) {
			char *sp;
			sp = strrchr( filename, '.');
			*sp = '\0';
			mystrncat( filename, PILOT_SUFFIX, SMALLLEN - 4);
			if ( (stat( filename, &stat_buf) == 0 ) &&
					     ( S_ISREG(stat_buf.st_mode))) {
				cat_type[i] = CAT_PILOT;
			}
		}

		if ( cat_type[i] == CAT_UNKNOWN) {
			char *sp;
			sp = strrchr( filename, '.');
			*sp = '\0';
		}
 
 		if ( (cfp = fopen( fname, "r")) == NULL ) {
			if ( !quiet )
				fprintf( stderr,
			"Warning: can't read calendar file: %s, skipping it.\n",
				 	fname);
			continue;
		}
		ical2mem( lptr, cfp, i, SUB_CAL );
		fclose( cfp);
	}

	sort_list( lptr);
	/* 
	 * Now go back and look for duplicate PilotRecordId's and 
	 * change them all to 0 so that they get loaded to Pilot.
	 * Then resort.
	 */

	lep = lptr->first;
	while( lep ) {

		if ( lep->pi_rec_id == 0) {
			lep = lep->next;
			continue;
		}

		nep = lep->next;
		while ( nep && (nep->pi_rec_id == lep->pi_rec_id)) {
			nep->pi_rec_id = 0;
			nep->type &= ~IS_NEW;
			nep->type |= IS_REVISION;
			nep = nep->next;
		}

		if ( nep > lep->next) {
			lep->pi_rec_id = 0;
			lep->type &= ~IS_NEW;
			lep->type |= IS_REVISION;
		}
		lep = nep;
	}
	sort_list( lptr);
}


static void
ical2mem( struct A_list *lptr, FILE *fp, int cat, enum CalType file_type )
{
        int	brack_count, flag_esc;

	struct A_entry
		tmp_ae; /* don't ever free this */

	char	*cp,
		*cp2,
		ownbuf[SMALLLEN],
		tmpbuf[MIDLEN];

	lptr->default_alarm = ALARM_NONE;
	lptr->defalrm_units = ALARM_MINUTES;

	cp2 = tmpbuf;
	linecount = 0;
	brack_count = 0;
	flag_esc = 0;
	while ( fgets( linebuf, MIDLEN - 4, fp) != NULL) {

	        /* 07/16/2000 : mdb - Changed parser.
		 * The parser was slightly rewritten to fix the problem
		 * of accidentally counting escaped \[ and \] which are a
		 * legal part of the ical file format.  I am now using
		 * \[ and \] within the DateBk3Info [] field instead of
		 * hard coding [->(, ]->) replacements as was done before,
		 * which is how I found this parsing problem.
		 *
		 * There are likely still many bugs in syncal where escaped
		 * chars are not handled when reading the fields, potentially
		 * causing mismatches with the pilot records.  It also looks
		 * like existing code in syncal escapes [,] but not the escape
		 * itself, so "\[" in the actual text of an appt would cause
		 * problems.
		 */

		cp = linebuf;
		linecount++;
		while ( *cp ) {
			switch ( *cp) {
			case '\\':
			        if (flag_esc == 0) flag_esc = 1;
                                break;
			case '[':
				if (flag_esc == 0) brack_count++;
				break;

			case ']':
				if (flag_esc == 0) brack_count--;
				break;
			}
			*cp2++ = *cp++;

			if (flag_esc == 2) flag_esc = 0;
			if (flag_esc) ++flag_esc;
		}
		*cp2 = '\0';

		if ( brack_count < 0 )
			parse_error( "Unbalanced brackets [ ]" );

		else if ( brack_count > 0 )
			continue;

		/* balanced brackets at end of line */

		/* initialize tmp_ae */
		tmp_ae.datebk_magic = NULL;
		tmp_ae.owner = NULL;
		tmp_ae.category = cat;

		cp2 = tmpbuf;

		if ( strncasecmp( tmpbuf, "Appt", 4) == 0 )
			tmp_ae.type = 0;  /* it's an appt */
		else if ( strncasecmp( tmpbuf, "Note", 4) == 0 )
			tmp_ae.type = IS_EVENT;
		else if ( strncasecmp( tmpbuf, "PilotCategories", 15) == 0 ) {
			use_datebk3 = TRUE;
			if ( file_type == ARCHIVE)
				get_arcats( tmpbuf);
			continue;
		}
       		else if ( strncasecmp( tmpbuf, "Calendar [", 10) == 0 ) {
			if ( tmpbuf[11] -'0' < 2) {
				fprintf( stderr, ERRMSG1, tmpbuf);
				fprintf( stderr, ERRMSG2);
				exit( 1);
			}
			continue;
		}
       		else if ( strncasecmp( tmpbuf, "DefaultAlarms [", 15) == 0 ) {
			char *cpx;

			cpx = &tmpbuf[15];
			while ( *cpx && (*cpx != ']'))
				cpx++;

			cpx--;
			while ( isspace( *cpx) && ( cpx > tmpbuf))
				cpx--;

			if ( *cpx == '[') { 
				add_extra( tmpbuf);
				continue;
			}

			while ( !isspace( *cpx) && ( cpx > tmpbuf))
				cpx--;

			lptr->default_alarm = atoi( cpx);

			if ( (lptr->default_alarm % 60) == 0 ) {
				lptr->default_alarm = lptr->default_alarm/60;
				lptr->defalrm_units = ALARM_HOURS;
			}
			else
				lptr->defalrm_units = ALARM_MINUTES;

			add_extra( tmpbuf);
			continue;
		}
		else {
			if ( file_type == MAIN_CAL)
				add_extra( tmpbuf);
			/* unknown but copy it to new calendar */
			continue;
		}


		if ( (cp = strstr( tmpbuf, "Todo [")) != NULL)
			tmp_ae.type |= IS_TODO;
		if ( (cp = strstr( tmpbuf, "Done [")) != NULL)
			tmp_ae.type |= IS_DONE;

		tmp_ae.hilite = hi_always;
		if ( (cp = strstr( tmpbuf, "Hilite [")) != NULL) {
			cp += 8;
			if ( strncasecmp( cp, "always", 6) == 0 )
				tmp_ae.hilite = hi_always;
			if ( strncasecmp( cp, "never", 5) == 0 )
				tmp_ae.hilite = hi_never;
			if ( strncasecmp( cp, "expire", 6) == 0 )
				tmp_ae.hilite = hi_expire;
			if ( strncasecmp( cp, "holiday", 7) == 0 )
				tmp_ae.hilite = hi_holiday;
		}
		if ( (cp = strstr( tmpbuf, "Owner [")) != NULL) {
			char	*cpx;

			cp += 7;
			if ( (cpx = strchr( cp, ']')) == NULL)
				parse_error( "Bad 'Owner' field" );
			
			*cpx = '\0';
			mystrncpy( ownbuf, cp, SMALLLEN);
			*cpx = ']';
			tmp_ae.owner = ownbuf;
		}

		if ( (cat_type[cat] != CAT_ICAL) &&
		     		((cp = strstr( tmpbuf, "PilotRecordId [")) != NULL)) {
			char	*xcp;

			cp += 15;
			xcp = rec_id_str;
			while ( *cp && ( *cp != ']' ))
				*xcp++ = *cp++;
			*xcp = '\0';
			tmp_ae.pi_rec_id = strtol( rec_id_str, 0, 16);
		}
		else {
			/* It has no PilotRecordId or it's a CAT_ICAL category */
			tmp_ae.pi_rec_id = 0;
			rec_id_str[0] = '\0';
			if ( file_type == ARCHIVE)
				parse_error("Archive record without PilotRecordID");
			else
				tmp_ae.type |= IS_NEW;
		}

		/* Set category and db3magic (only for ARCHIVE) */

		if ( file_type == ARCHIVE) {
			if ( (cp = strstr( tmpbuf, "DateBk3Info [")) != NULL) {
				int len;
				char *unstuffed;
				cp += 13;

				len = str_findenddelim(cp, ']', '\\');
				tmp_ae.datebk_magic = str_ndup(cp, len);
				if ((unstuffed = str_byteunstuff(tmp_ae.datebk_magic, '\\'))) {
				        if (tmp_ae.datebk_magic) free(tmp_ae.datebk_magic);
					tmp_ae.datebk_magic = unstuffed;
				}
				
				cp += len;
				tmp_ae.category = (int) (tmp_ae.datebk_magic[4] - '@');
			}
		}
		add2list( tmpbuf, &tmp_ae, lptr);
	}
}

/*
 * void cleanup( int status )
 * Cleanup the connection to the pilot and close down connection.
 */

void
cleanup( int status )
{
	char 	buf[2*SMALLLEN];

	if ( no_action) {
		dlp_CloseDB_All( pilot_sd); 	/* Close any (all?) databases */
		pi_close( pilot_sd);
	}
	else {
		if ( my_hostname[0] == '\0') {
			mystrncpy( buf, 
			"Synchronized Ical Calendar with Pilot ",  SMALLLEN);
		}
		else
			sprintf( buf, 
			"Synchronized Ical Calendar (%s) with Pilot ", my_hostname);

		if (use_datebk3 )
			mystrncat( buf, "DateBk3/4 database\n", SMALLLEN);
		else
			mystrncat( buf, "DateBookDB\n", SMALLLEN);

		if ( pi_user.userID != 0 ) {
			char buf2[SMALLLEN];

			sprintf( buf2, "User: %s, User Id: %ld\n",
				 pi_user.username, pi_user.userID);
			mystrncat( buf, buf2, 2*SMALLLEN);
		}
		mystrncat( buf, VERSION, 2*SMALLLEN);
		mystrncat( buf, "\n", 2*SMALLLEN);
		  
 		/* If normal end, clear out deleted/archived DB entries */
		if ( status == dlpEndCodeNormal)
			dlp_CleanUpDatabase( pilot_sd, pi_date_db);

		/* Tell the user who it is, with a different PC id. */
		pi_user.successfulSyncDate = time(NULL);
		pi_user.lastSyncDate = pi_user.successfulSyncDate;
		if ( dlp_WriteUserInfo( pilot_sd, &pi_user) < 0 )
			if ( !quiet )
				fprintf( stderr, "dlp_WriteUserInfo failed\n");

  

		if ( dlp_AddSyncLogEntry( pilot_sd, buf) < 0 )
			if ( !quiet )
				fprintf( stderr, "dlp_WriteUserInfo failed\n");
  
		dlp_EndOfSync( pilot_sd, status);
		pi_close( pilot_sd);
	}
}


static void
parse_error( char * msg)
{
	if ( !quiet )
		fprintf( stderr, "Parse error in file %s, line %d\n",
			 filename, linecount);
	if ( *msg && !quiet)
		fprintf( stderr, "\t%s\n", msg);

	cleanup( dlpEndCodeOther);
	exit( 1);
}


static void
create_uid( char * buf, recordid_t pil_id)
{
#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN	(64)
#endif
	static unsigned long	hostid = 0;
	static unsigned long	count = 0;
	static pid_t	pid = 0;

	int hostnamelen;

	if ( MAXHOSTNAMELEN > (SMALLLEN/4))
		hostnamelen = (SMALLLEN/4);
	else
		hostnamelen = MAXHOSTNAMELEN;
	  
	if ( !my_hostname[0]) {
		if ( gethostname( my_hostname, hostnamelen) != 0 ) {
			if ( !quiet )
				fprintf( stderr, "Can't get hostname\n");
			strcpy( my_hostname, "Unknown.Host");
		}
		hostid = gethostid( );
		pid = getpid( );
	}

	sprintf( buf, "%s_%lx_%x_%lx", my_hostname, hostid, pid, count);
	count++;
}


void
be_verbose( struct A_entry *ep, unsigned type)
{
	struct Appointment
		*apt;

	char  	*cp,
		*cp2,
		*prefix,
		*weekdays[] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"},
		desc[MIDLEN];

	int	i;

	if ( cat_type[ep->category] != CAT_REG ) {
		if ( !lnk_verbose)
			return;
	}
	else {
		if ( !verbose)
			return;
	}

	apt = &(ep->appt);

	switch( type & (IS_EVENT +IS_NEW +IS_DELETION +IS_REVISION +FOR_PILOT)) {
	case (IS_NEW):
	case (IS_NEW + IS_EVENT):
		prefix = "Adding entry to ical calendar:\n\t";
		break;
	case (IS_NEW + FOR_PILOT):
	case (IS_NEW + IS_EVENT + FOR_PILOT):
		prefix = "Adding entry to Pilot DateBook:\n\t";
		break;
	case (IS_REVISION + FOR_PILOT):
	case (IS_REVISION + IS_EVENT + FOR_PILOT):
		prefix = "Revising entry on Pilot:\n\t";
		break;
	case (IS_REVISION):
	case (IS_REVISION + IS_EVENT):
	  	if ( type & IS_TODO)
			return;
		prefix = "Revising ical calendar entry:\n\t";
		break;
	case (IS_DELETION):
	case (IS_DELETION + IS_EVENT):
		prefix = "Deleting ical calendar entry:\n\t";
		break;
	case (IS_DELETION + FOR_PILOT):
	case (IS_DELETION + IS_EVENT + FOR_PILOT):
		prefix = "Deleting Pilot calendar entry:\n\t";
		break;
	default:
		prefix = "";
	}

	cp = apt->description;
	cp2 = desc;
	while  ( cp && *cp && ((cp2 - desc) < (MIDLEN - 1))) {
		*cp2++ = *cp++;
		if ( *(cp -1) == '\n')
			*cp2++ = '\t';
	}
	*cp2 = '\0';

	printf("%s%s\t%s\n", prefix, asctime( &(apt->begin)), desc);

	if ( apt->repeatType == repeatDaily ) {
		if ( apt->repeatFrequency == 1 )
			printf("\tRepeat every day");
		else
			printf("\tRepeat every %d days", apt->repeatFrequency);

		if ( apt->repeatForever == TRUE )
			printf( "\n");
		else
			printf( " until %s", asctime( &(apt->repeatEnd)));
	}
	if ( apt->repeatType == repeatWeekly ) {
		printf("\tRepeat every ");
		for ( i = 0; i < 7; i++ ) {
			if ( apt->repeatDays[i] == TRUE )
				printf("%s ", weekdays[i]);
		}
		if ( apt->repeatForever == TRUE )
			printf( "\n");
		else
			printf( "until %s", asctime( &(apt->repeatEnd)));
	}
	if ( apt->repeatType == repeatMonthlyByDate ) {
		if ( apt->repeatFrequency == 1 )
			printf("\tRepeat every month");
		else
			printf("\tRepeat every %d months", apt->repeatFrequency);

		if ( apt->repeatForever == TRUE )
			printf( "\n");
		else
			printf( " until %s", asctime( &(apt->repeatEnd)));
	}
	if ( apt->repeatType == repeatMonthlyByDay ) {
		int	rday;
		char	*weekstr;

		rday = apt->repeatDay;
		switch ( rday/7) {
		case 0:
			weekstr = "first";
			break;
		case 1:
			weekstr = "second";
			break;
		case 2:
			weekstr = "third";
			break;
		case 3:
			weekstr = "fourth";
			break;
		case 4:
			weekstr = "last";
			break;
		default:
			weekstr = "?";
		}

		if ( apt->repeatFrequency == 1 )
			printf("\tRepeat every month, ");
		else
			printf("\tRepeat every %d months, ", apt->repeatFrequency);

		printf( "the %s %s of the month", weekstr, weekdays[rday % 7]);

		if ( apt->repeatForever == TRUE )
			printf( "\n");
		else
			printf( " until %s", asctime( &(apt->repeatEnd)));
	}
	if ( apt->repeatType == repeatYearly ) {
		printf("\tRepeat annually");
		if ( apt->repeatFrequency > 1)
			printf( " every %d years", apt->repeatFrequency);
		if ( apt->repeatForever == TRUE )
			printf( "\n");
		else
			printf( " until %s", asctime( &(apt->repeatEnd)));
	}
	if ( (i = apt->exceptions) > 0 ) {
		while ( i > 0 ) {
			i--;
			printf( "\tomitting %s", asctime( &(apt->exception[i])));
		}
	}
}


/*
 * struct A_entry * make_new_entry( char *buf, recordid_t pi_id, int type, char *)
 * mallocs space for a new struct A_entry and its associated 
 * data and copies the contents of  buf to the proper place.  The 
 * pi_id field is properly set if pi_id is not 0.  The parameter 'type'
 * is the type field for the struct A_entry which will be created.
 * If own_p is non-NULL then strdup this string and put in the owner field.
 */

static struct A_entry *
make_new_entry( char *buf, recordid_t pi_id, unsigned type, char *own_p )
{
	struct A_entry *ep;

	ep = (void *) malloc( sizeof( struct A_entry));

	if( ep) {
		ep->ical_ent = strdup( buf);
		if ( own_p)
			ep->owner = strdup( own_p);
		else
			ep->owner = NULL;
	}
	if ( (ep == NULL) || (ep->ical_ent == NULL) || 
	     			( own_p && (ep->owner == NULL))) {
		if ( !quiet )
			fprintf( stderr, "Out of memory\n");
		cleanup( dlpEndCodeOutOfMemory);
		exit( 1);
	}
	mystrncpy( ep->ical_ent, buf, strlen( buf) + 1);
	ep->pi_rec_id = pi_id;
	ep->type = type;
	ep->next = NULL;
	return ep;
}

void
add2list( char *buf, struct A_entry *ae_p, struct A_list *ll )
{
	struct A_entry * newp;

	newp = make_new_entry( buf, ae_p->pi_rec_id, ae_p->type, ae_p->owner);
	newp->hilite = ae_p->hilite;
	newp->datebk_magic = str_dup(ae_p->datebk_magic);
	newp->category = ae_p->category;
	newp->ical_match = NULL;
	newp->pilot_match = NULL;
	fill_appt( newp, ll);

	if ( ll->last == NULL)
		ll->first = ll->last = newp;
	else {
		ll->last->next = newp;
		ll->last = newp ;
	}		
	ll->last->next = NULL;
	ll->num++;
}

static void
add_extra( char *buf)
{
	int	len;
	char	*cp;

	cp = buf;
	while ( *cp && (*cp != '\n'))
		cp++;
	if ( *cp)
		cp++;
	*cp = '\0';

	len = strlen( extra_stuff) + strlen( buf) + 4;
	extra_stuff = realloc( extra_stuff, len);
	if ( extra_stuff == NULL) {
		if ( !quiet )
			fprintf( stderr, "Out of memory\n");
		cleanup( dlpEndCodeOutOfMemory);
		exit( 1);
	}
	mystrncat( extra_stuff, buf, len);
}

static void
syncal_timeout( int sig)
{
	signal( SIGALRM, SIG_DFL);
	alarm( 0);

	if ( !quiet)
		fprintf( stderr, "\007\nTimed out waiting for HotSync.\n");
	exit( 2);
}

static void
backups( void)
{
	char
		backname[2*SMALLLEN];

	if ( no_action)
		return;


	mystrncpy( backname, archive_file, SMALLLEN);
	mystrncat( backname, ".bak", 2*SMALLLEN);
	if ( rename( archive_file, backname) < 0 ) {
		if ( !quiet )
			fprintf( stderr,
				 "Renaming %s failed: aborting\n", archive_file);
		cleanup( dlpEndCodeOther);
		exit( 1);
	}

	if ( rename( tmp_file, archive_file) < 0 ) {
		if ( !quiet )
			fprintf( stderr,"Renaming %s failed: aborting\n", tmp_file);
		cleanup( dlpEndCodeOther);
		exit( 1);
	}
}

/* returns length of the magic DateBk header (if any is found)
 *  for DateBk3, this is 0,10
 *  for DateBk4, this is 0,21,21+len any link tags
 *
 *  if ver != NULL : *ver = 4 DateBk4, 3 DateBk3, 0 otherwise
 */
int
is_datebk_magic(char *ptr, int *ver)
{
        int len;
	char *next_nl;  /* next newline in note */

	if (!ptr || ((next_nl = strchr( ptr, '\n')) == NULL) || strncmp(ptr, "##", 2) != 0) {
		if (ver) *ver = 0;
		return 0;
	}
	len = strlen(ptr);
	if (len >= 20 && (next_nl == &ptr[19])) { /* DateBk4 */
		char	*cp;

		if (ver) *ver = 4;
	        if (ptr[9] == '@') return 20;  /* no Links present */

	        /* parse the Link entries */
		/* ASSUME: They all appear immediately after the ## header
		 * and have the form "[Link ...]" separated by '\n'.
		 */
		
		cp = next_nl + 1;
		while ( strncmp( cp, "[Link ", 6) == 0 ) {
			if ( (next_nl = strchr( cp, '\n')) == NULL) {
				if ( cp[strlen(cp)-1] == ']')
					cp += strlen(cp);
				break;
			}
			if ( *(next_nl - 1) != ']')
				break;
			cp = next_nl + 1;
		}

		return (cp - ptr) ;
	} else if (len >= 10 && (next_nl == &ptr[9])) { /* DateBk3 */
	        if (ver) *ver = 3;
	        return 10;
	} else if (ver)
		*ver = 0;
	return 0;
}

#define BIGPRIME	(1000193)

static unsigned 
ahash( char *str)
{

	unsigned i,
		tmp,
		val;

	val = 0;
	tmp = 0;
	i = 0;
	while ( *str ) {
		if ( i == 4) {
			i = 0;
			val += tmp;
			val = val % BIGPRIME;
			tmp = 0;
		}
		tmp += (*str++)<<(4*i);
		i++;
	}
	val += tmp;
	val = val % BIGPRIME;
	return val;
}


/* copy the given string into a newly allocated string */
char *str_dup(char *str) {
	int len;
	char *str_new;
	if (!str) return NULL;
	len = strlen(str);
	str_new = (char*)malloc(len+1);
	memcpy(str_new, str, len);
	str_new[len] = '\0';
	return str_new;
}

/* copy up to n of the given string into a newly allocated string */
char *str_ndup(char *str, int n) {
	int len;
	char *str_new;
	if (!str) return NULL;
	len = strlen(str);
	if (len > n) len = n;
	str_new = (char*)malloc(len+1);
	memcpy(str_new, str, len);
	str_new[len] = '\0';
	return str_new;
}


/* These fcns are intended to be used to replace the [,] -> () stuffing
 * in the DateBk3Info field, and to clean up all other Contents field
 * text escaping.
 */

/* rtn: NULL=no changes, !NULL=allocated buffer with changes */
char *str_bytestuff(char *str, char *str_toesc, char esc) {
	int len, len_new;
	char *p, *q, *str_new;

	if (!str) return NULL;

	/* pass 1 - determine size */
	len_new = len = strlen(str);
	for (p=str; *p; ++p) {
		if (strchr(str_toesc, *p) || *p == esc) ++len_new;
	}
	if (len_new == len) return NULL;

	/* pass 2 - do byte stuffing */
	q = str_new = (char*)malloc(len_new+1);
	for (p=str; *p; ++p) {
		if (strchr(str_toesc, *p) || *p == esc) {
			*q++ = esc;
			*q++ = *p;
		} else {
			*q++ = *p;
		}
	}
	*q = '\0';
	return str_new;
}

/* rtn: NULL=no changes, !NULL=allocated buffer with changes */
char *str_byteunstuff(char *str, char esc) {
	int len, len_new;
	char *p, *q, *str_new;

	if (!str) return NULL;

	/* pass 1 - determine size */
	len_new = len = strlen(str);
	for (p=str; *p; ++p) {
		if (*p == esc) {
			++p;
			--len_new;
		}
	}
	if (len_new == len) return NULL;

	/* pass 2 - do byte unstuffing */
	q = str_new = (char*)malloc(len_new+1);
	for (p=str; *p; ++p) {
		if (*p == esc) {
			++p;
			*q++ = *p;
		} else {
			*q++ = *p;
		}
	}
	*q = '\0';
	return str_new;
}

int str_findenddelim(char *str, char delim, char esc) {
	int rtn;
	for (rtn=0; *str; ++str,++rtn) {
		if (*str == delim) break;
		if (*str == esc) {
			++str; ++rtn;
		}
		if (*str == '\0') {  /* in case str ends with 'esc' */
			break;
		}
	}
	return rtn;
}
