/*
 * syncal.h:
 * syncal Version 0.8.6
 *
 * Copyright (c) 1999-2000, John Franks
 *
 * This is free software, licensed under the GNU Public License V2.
 * See the file COPYING for details.
 */


#include "pi-source.h"
#include "pi-socket.h"
#include "pi-todo.h"
#include "pi-datebook.h"
#include "pi-dlp.h"

/* This is the default time in minutes for alarms to go off */

#define DEFAULT_ALARM	(0)
 

/* 
 * These are the calendar file suffixes which indicate
 * an ical dominant file and a Pilot dominant file respectively.
 */

#define	ICAL_SUFFIX	".Ical"
#define	PILOT_SUFFIX	".Pilot"


/*
 * The main calendar (default $HOME/.calendar) and the 
 * syncal home directory (default $HOME/.syncal) are obtained
 * by evaluating the environment variable $HOME and appending
 * the values below unless there is no $HOME in which case the
 * current directory is used.  If values below are
 * defined with a leading '/' then they are taken as an absolute
 * path and $HOME is not used.
 */

#define MAIN_CALENDAR	".calendar"
#define SYNCAL_HOME_DIR	".syncal"

/* 
 * The following will have the  SYNCAL_HOME_DIR prepended before use.
 */

#define SYNCAL_TMP_FILE	"syncal.tmp"
#define ARCHIVE_FILE	"calendar.archive"


#define ERRMSG1 "Error: Your ical calendar version is: %s\n"
#define ERRMSG2 "Ical version 2.0 or later is required by syncal\n"
#define ERRMSG3 "Couldn't read symlink %s\n"
#define ERRMSG4 "Symlink for %s too long.\nFix syncal to use larger filename lengths.\n"
#define ERRMSG5 "A symlink shouldn't have a / at the end.\nSomething is wrong here\n"
#define ERRMSG6 "The symlink for %s produces a filename that is longer than %d characters.  Fix the link or fix syncal.\n"


/* Not all systems have timegm( ), but at least Linux does. */
#ifdef  __linux
#define gmt_mktime	timegm
#define HAVE_TIMEGM
#endif


#define VERSION		"Syncal Version 0.8.6"
#define HOTSYNC_TIMEOUT	(60)	/* Wait at most 60 seconds for HotSync */
#define TRUE		(1)
#define FALSE		(0)
#define MIDLEN		(6*1024)
#define SMALLLEN	(512)
#define TINYLEN		(32)
#define PI_DESC_LEN	(256)
#define PI_NOTE_LEN	(4*1024)

/* Type for ical hilite[] field */
enum hilite_types { hi_always, hi_never, hi_expire, hi_holiday};


/*
 * struct A_entry 
 * This is the basic data type used by syncal to sychronize an
 * an ical calendar and a Pilot DatebookDB.  It contains the 
 * relevant data for an appointment or "event" in two formats:
 * one used by ical and the other used by the Pilot.  The struct
 * appt is declared in dlp-datebook.h and contains all the information
 * for a Pilot DateBook entry.  The pointer ical_ent points to a string
 * consisting of several lines representing an "Appt" or "Note" as they
 * occur in an ical calendar file.
 *
 * The function fill_appt( struct A_entry *) translates the data
 * pointed to by ical_ent into data for an Appointment and fills in
 * the appt struct. The function
 * fill_ical_ent( char *buf, struct Appointment *ap, recordid_t pilot_id)
 * goes the other way taking a pointer to an Appointment and the pilot
 * record id and filling buf with a corresponding ical calendar entry.
 */

struct A_entry {
	struct Appointment
		appt;		/* Pilot Appointment */

	char			
		*ical_ent,	/* the ical entry data */
		*owner,		/* ical Owner field entry */
	        *datebk_magic;  /* can be variable sized */
        int     datebk_magic_ver; /* 0 = none, 3 = DateBk3, 4 = DateBk4 */

	recordid_t
		pi_rec_id;	/* PilotRecordId */
	enum hilite_types
		hilite;		/* hi_always, hi_never, hi_expire, hi_holiday */
	unsigned
		type,
		category;
	struct A_entry
		*pilot_match,
		*ical_match,
		*next;
};

/* bitmasks for the 'type' field of struct A_entry */
#define IS_EVENT	(1<<0)
#define IS_NEW		(1<<1)
#define IS_REVISION	(1<<2)
#define IS_DELETION	(1<<3)
#define FOR_PILOT	(1<<4)
#define IS_TODO		(1<<5)
#define IS_DONE		(1<<6)

#ifdef HAVE_TIMEGM
#define gmt_mktime	timegm
#endif

#define MAX_PILOT_CATS	(16)

/*
 * struct A_list
 * This a linked list of A_entries.  This is the data type used to 
 * store in memory all the appointments from an ical calendar or from
 * a Pilot DateBookDB.
 */
 
struct A_list {
	int		num;

	struct A_entry	*first,
			*last;

	int
			default_alarm,
			defalrm_units;
};

extern int	cat_type[MAX_PILOT_CATS];

/* Values for cat_type */
#define CAT_UNKNOWN	(0)
#define CAT_REG		(1)
#define CAT_ICAL	(2)
#define CAT_PILOT	(3)

#define ALARM_NONE	(-1)
#define ALARM_MINUTES	(0)
#define ALARM_HOURS	(1)


extern struct AppointmentAppInfo
	datebk_info;

extern time_t
	gmt_mktime( struct tm *);

extern struct A_entry	
	*ismatch( recordid_t, struct A_list *);

extern struct A_list
	archive_list,
	ical_list,
	pilot_list,
	new_list;

extern int
	use_datebk3,
	pi_date_db,
	pilot_sd,
	no_action,
	quiet,
	verbose,
        is_datebk_magic( char *, int *),
	get_int( void ),
	mystrncpy( char *, char *, int),
        mystrncat( char *, char *, int),
        str_findenddelim(char *, char, char);

extern FILE
 	*fp;

extern void
	get_pcats( struct AppointmentAppInfo *),
	get_arcats( char *),
	format_error( char *),
	reconcile( void),
	fill_appt( struct A_entry *, struct A_list *),
	get_token( void ),
	get_lb( void ),
	get_rb( void ),
	be_verbose( struct A_entry *, unsigned),
	cleanup( int),
	add2list( char *, struct A_entry *, struct A_list *ll ),
	delete_entry( struct A_entry * , struct A_list *),
	free_Alist( struct A_list *),
	pilotlist2newlist( void ),
	write2pilot( struct Appointment *, recordid_t, recordid_t *, int),
	mem2ical( struct A_list *, FILE *),
	sort_list( struct A_list * ),
	sort_array( struct A_list *);

extern char
	pilot_cats[MAX_PILOT_CATS][16],
	archv_cats[MAX_PILOT_CATS][16],
	tmp_file[SMALLLEN],
	cal_file[SMALLLEN],
	archive_file[SMALLLEN],
	*extra_stuff,
	*skip( char *),
        *str_dup( char *),
        *str_ndup( char *, int),
        *str_bytestuff( char *, char *, char),
        *str_byteunstuff( char *, char);



