/*
 * syncal_reconcile.c
 * syncal Version 0.8.6
 *
 * Copyright (c) 1999-2000, John Franks
 *
 * Japanese character set modifications contributed by Hiroshi Kawashima
 *
 * This is free software, licensed under the GNU Public License V2.
 * See the file COPYING for details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "syncal.h"


static int
	are_equal( struct Appointment *, struct Appointment *);

struct A_entry	*ismatch( recordid_t, struct A_list *);


/*
 * static struct A_entry *ismatch( recordid_t id, struct A_list *lptr)
 * Searches the linked list *lptr for the first entry with Pilot record
 * id equal to "id".  It returns a pointer to the matching entry or
 * NULL if none is found.
 */

struct A_entry *
ismatch( recordid_t id, struct A_list *lptr)
{
	struct A_entry	*lep;

	lep = lptr->first;
	while ( lep ) {
		if ( lep->pi_rec_id == id)
			return (lep);
		lep = lep->next;
	}
	return NULL;
}


/*
 * reconcile( void) goes through the archive_list and compares all 
 * items to see if they are in ical_list or pilot_list or both.
 */

void
reconcile( void)
{
	int	i_eq_a,
		p_eq_a;

	unsigned
		oldtype;

	struct A_entry
		*alp,
		*imatch_p,
		*pmatch_p;

	alp = archive_list.first;

	while ( alp != NULL ) {

		imatch_p = alp->ical_match;
		pmatch_p = alp->pilot_match;

		/* 
		 * We need to canonicalize the Appointment struct in 
		 * the pilot list entry.  We do this by converting its
		 * ical_ent field with fill_appt( ).
		 */

		if ( pmatch_p) {
			fill_appt( pmatch_p, NULL);
		}
		if ( imatch_p) {
			fill_appt( imatch_p, NULL);
		}

		if ( pmatch_p && imatch_p ) {
			/* 
			 * It occurs in the archive list , the pilot list
			 * and the ical list.   Check for changes.
			 */
			i_eq_a = are_equal( &(imatch_p->appt), &(alp->appt));
			p_eq_a = are_equal( &(pmatch_p->appt), &(alp->appt));

			/* check todo/done status */
			i_eq_a = i_eq_a && ((alp->type & (IS_TODO + IS_DONE)) ==
					(imatch_p->type & (IS_TODO + IS_DONE)));
			/* check hilite value */
			i_eq_a = i_eq_a && (alp->hilite == imatch_p->hilite);

			/* check category */
			i_eq_a = i_eq_a && (alp->category == imatch_p->category);
			p_eq_a = p_eq_a && (alp->category == pmatch_p->category);

			if ( i_eq_a && p_eq_a ) {  /* both unchanged */
				add2list( imatch_p->ical_ent, imatch_p, &new_list);
				alp = alp->next;
				continue;
			}
			else if ( p_eq_a ) {  /* only ical changed */
				recordid_t	new_id = 0;

				if ( !no_action)
					write2pilot( &(imatch_p->appt),
					     imatch_p->pi_rec_id, &new_id,
						     imatch_p->category);
				   
				imatch_p->pi_rec_id = new_id;
				oldtype = imatch_p->type;
				imatch_p->type = (oldtype & (IS_TODO + IS_DONE))
						+ IS_REVISION + FOR_PILOT;

				add2list( imatch_p->ical_ent, imatch_p, &new_list);
				imatch_p->type = oldtype;

				alp = alp->next;
				continue;
			}
			else {  /* only pilot changed or both changed  */
				oldtype = pmatch_p->type;
				pmatch_p->type = oldtype | ((imatch_p->type & 
					(IS_TODO + IS_DONE)) + IS_REVISION);
				pmatch_p->hilite = imatch_p->hilite;
				if (imatch_p->owner)
					pmatch_p->owner = strdup( imatch_p->owner);
				else
					pmatch_p->owner = NULL;

				add2list( pmatch_p->ical_ent, pmatch_p, &new_list);
				pmatch_p->type = oldtype;

				alp = alp->next;
				continue;
			}

		}

		if ( !pmatch_p ) {
			/* 
			 * It's been deleted in the pilot so
			 * don't add it to new_list.
			 */
			if (cat_type[alp->category] == CAT_REG)
				be_verbose( alp, IS_DELETION);
			alp = alp->next;
			continue;
		}
		if ( !imatch_p ) {
			int	res;

			/* 
			 * It's been deleted in the ical calendar
			 * so delete it in the pilot and don't add to 
			 * new_list.  If it is in a read_only category
			 * it's already been deleted from pilot so 
			 * don't do it again.
			 */
			if ( (!no_action) && (cat_type[alp->category] == CAT_REG)) {
				res = 0;
				res = dlp_DeleteRecord( pilot_sd,
						pi_date_db, 0, alp->pi_rec_id );
				if ( (res != 0) && !quiet ) {
					fprintf( stderr, "Deletion failed!\n");
					fprintf( stderr, "\t%s\n",
						 alp->appt.description);
				}
			}
			if ( res == 0)
				be_verbose( alp, (IS_DELETION + FOR_PILOT));
			alp = alp->next;
			continue;
		}
		alp = alp->next;
	}
}

/*
 * void pilotlist2newlist( void )
 * Run through pilot_list and delete any which are for CAT_ICAL
 * categories and find any new entries which are then added to new_list.
 */


void
pilotlist2newlist( void )
{
	struct A_entry  *ptr,
			*amatch;

	ptr = pilot_list.first;
	while ( ptr != NULL ) {
		if ( (!no_action) && (cat_type[ptr->category] == CAT_ICAL)) {
			int	res;

			/* It should be deleted from the pilot */
			res = dlp_DeleteRecord( pilot_sd,
					pi_date_db, 0, ptr->pi_rec_id );
			if ( (res != 0) && !quiet ) {
				fprintf( stderr, "Deletion failed!\n");
				fprintf( stderr, "\t%s\n",
					 ptr->appt.description);
			}
			else
				be_verbose( ptr, (IS_DELETION + FOR_PILOT));
			ptr = ptr->next;
			continue;
		}
		if ( ((amatch = ismatch( ptr->pi_rec_id, &archive_list)) == NULL) ||
		     			cat_type[ptr->category] == CAT_PILOT) {
			/* it's a new pilot entry or CAT_PILOT category */
			ptr->type |= IS_NEW;
			ptr->hilite = hi_always;
			add2list( ptr->ical_ent, ptr, &new_list);
		}
		else 
			/* remember the match */
			amatch->pilot_match = ptr;

		ptr = ptr->next;
	}
}

/*
 * Check equality of two appointments.  Equals means same 
 * description, note and exception list and repeat data.
 */

static int
are_equal( struct Appointment *p1, struct Appointment *arch_p)
{
	char
		*cp,
		*anote_p;

	int i;
	
	if ( strcmp( p1->description, arch_p->description) != 0)
		return FALSE;

	cp = p1->note;
	if ( cp && !*cp )
		cp = NULL;
 
	anote_p = arch_p->note;
	if ( anote_p && !*anote_p )
		anote_p = NULL;

	if ( (anote_p && !cp) || ( cp && !anote_p ))
		return FALSE;
	if ( cp && anote_p && (strcmp( cp, anote_p) != 0))
		return FALSE;
	if ( gmt_mktime( &(p1->begin)) != gmt_mktime( &(arch_p->begin)))
		return FALSE;
	if ( gmt_mktime( &(p1->end)) != gmt_mktime( &(arch_p->end)))
		return FALSE;

	if ( p1->event !=  arch_p->event)
		return FALSE;

	if ( p1->alarm != arch_p->alarm )
		return FALSE;

	if ( p1->alarm) {
		if ( p1->advance != arch_p->advance )
			return FALSE;
		if ( p1->advanceUnits != arch_p->advanceUnits )
			return FALSE;
	}

	if ( p1->repeatType != arch_p->repeatType )
		return FALSE;

	if ( p1->repeatType == repeatNone)   /* no repeats, so we're done */
		 return TRUE;

	if ( (p1->repeatType == repeatMonthlyByDay) &&
				    ( p1->repeatDay != arch_p->repeatDay ))
		 return FALSE;

	if ( p1->repeatType == repeatWeekly) {
		for ( i = 0; i < 7; i++ ) {
			if ( p1->repeatDays[i] != arch_p->repeatDays[i])
				return FALSE;
		}
	}

	if ( p1->repeatFrequency != arch_p->repeatFrequency )
		return FALSE;

	if ( p1->repeatForever !=  arch_p->repeatForever )
		 return FALSE;

	if ( (!p1->repeatForever) &&
	     ( gmt_mktime( &(p1->repeatEnd)) != gmt_mktime( &(arch_p->repeatEnd))))
		return FALSE;

	if ( p1->exceptions != arch_p->exceptions )
		return FALSE;
	for ( i = 0; i < p1->exceptions; i++ ) {
		if ( p1->exception[i].tm_mday != arch_p->exception[i].tm_mday)
			return FALSE;
		if ( p1->exception[i].tm_mon != arch_p->exception[i].tm_mon)
			return FALSE;
		if ( p1->exception[i].tm_year != arch_p->exception[i].tm_year)
			return FALSE;
	}
	return TRUE;
}

#if defined(Japanese)
unsigned int CharEuc2Sjis(unsigned int code)
{
    register unsigned char jh, jl, sh, sl;

    jh = (code >> 8) & 0x7f;
    jl = code & 0x7f;
    if (jh <= 0x5e) {
	sh = (jh - 1) / 2 + 0x71;
    } else {
	sh = (jh - 1) / 2 + 0xb1;
    }
    if (jh & 1) {
	if ((sl = jl + 0x1f) >= 0x7f)
	    sl++;
    } else {
	sl = jl + 0x7e;
    }
    return (sh << 8) + sl;
}

static void Euc2Sjis(unsigned char *buf)
{
    unsigned char *p = buf, *q = buf;
    unsigned int code, cl, ch;
    while (p != NULL && *p) {
	if ((ch = *p++) & 0x80) {
	    cl = *p++;
	    code = (ch << 8) + cl;
	    code = CharEuc2Sjis(code);
	    *q++ = (code >> 8) & 0xff;
	    *q++ = code & 0xff;
	} else {
	    *q++ = ch;
	}
    }
}
#endif

void
write2pilot( struct Appointment * ap, recordid_t old_id,
	     					recordid_t *new_id_p, int cat )
{
	char	*cp;
	unsigned char 
		Appointment_buf[0xffff];
	int
		result,
		aptsize;
#if defined(Japanese)
	Euc2Sjis(ap->description);
	Euc2Sjis(ap->note);
#endif
	cp = ap->note;
	aptsize = pack_Appointment( ap, Appointment_buf, sizeof(Appointment_buf));
	result = dlp_WriteRecord( pilot_sd, pi_date_db, 0, old_id, cat,
					  Appointment_buf, aptsize, new_id_p );
	if ( (result < 0) && !quiet ) {
		fprintf( stderr, "dlp_WriteRecord failed\n");
		fprintf( stderr, "\t%s\n", ap->description);
	}
}


